/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/

/*
 * Abstract:	This file contains the histogram buffer
 *		kernel routines based on the IPROBE API
 *              kernel rationalization (reorganization).
 *
 * Author:		Bradley M. Waters
 * Creation Date:	20-Jul-1994
 *
 * Modified by:
 *
 * X-6  BMW             Bradley M. Waters       14-Nov-1994
 *      Fixed bug in HbCreate where histogram being created was 1
 *      element too small when mult. subhistograms were created.
 *      Fixed bug in HbRead where the status wasn't being checked
 *      to verify that the index was in the histogram.  Added call
 *      to OS specific HbValidRange to insure that the PC range being
 *      created was valid.  If not, an error is returned - this is
 *      necessary to handle sign extension in NT and VMS.
 *
 * X-5  BMW             Bradley M. Waters       26-Oct-1994
 *      Moved KIprHbMaxElements from OS specific module to
 *      this module based on Anil's suggestion (it was the only
 *      routine in histogram_buffer_nt and there was no reason for
 *      it to be OS specific).
 *
 * X-4  BMW             Bradley M. Waters       25-Oct-1994
 *      Fixed bug in KIprHbFind - it was returning IPR__FAILURE rather than
 *      IPR_HBNOTFND when the histogram buffer list was empty.
 *
 * X-3  BMW		Bradley M. Waters	10-Oct-1994
 *	Cleanup of code
 *
 * X-2	BMW		Bradley M. Waters	17-Aug-1994
 *	Updated based on code review input.
 *
 * X-1	BMW		Bradley M. Waters	20-Jul-1994
 *	Original Version.
 *
*/

#ifdef __LINUX__
#include <asm/uaccess.h>
#include <linux/kernel.h>
#endif

#ifdef __NT
#  include "ntddk.h"
#endif

#include "ipr_kernel_int_defs.h"
#include "hb_library.h"

#ifdef __NT
#  include "ipr_driver_nt.h"
#endif




#ifdef   TRCI
#include "TraceDrv.h"
#endif

#ifdef DBGDPC
extern int skipHist;
#endif

IPR_BUFHDR histogram_template = {
	IPR_PROD_IDENT,
	0,    // 2nd field (size) is filled in by allocation routine
	IPR_HBUFFER_TYPE,
	IPR_HBUFFER_SUBTYPE,
	IPR_HBUFFER_KIND
	};

extern IPR_STATUS KIprHbCreate (IN IPR_CPUNUM cpu, OUT IPR_HDESC *descrip,
                                OUT IPR_HBUFFER **retptr)
{

   IPR_BUFDSC
      *BDptr;
   IPR_HBUFFER
      *Hptr, *HOptr;
   unsigned int hsize,maxelem,allocSize,cnt;
   IPR_STATUS Status;

#ifdef TRC0
   TrcPrint("KIprHbCreate enter, cpu = %d\n",
     cpu);
#endif
   if (!descrip)
   {
#ifdef TRCE
      TrcPrint("KIprHbCreate Error: descrip ptr is null\n");
#endif
      return IPR__NULLPTR;
   }
   if (!retptr)
   {
#ifdef TRCE
      TrcPrint("KIprHbCreate Error: retptr is null\n");
#endif
      return IPR__NULLPTR;
   }
   *retptr = NULL;
   Status = get_bufdsc_address (cpu, &BDptr);
   if (!IPR_SUCCESS(Status))
      return Status;
   if (!BDptr)
      return IPR__INVCPU;

   if (BDptr -> numcounters == 0)
      return IPR__NOCOUNT;

   Status = KIprHbValidRange(descrip);
   if (!IPR_SUCCESS(Status))
      return Status;

   Status = IprHbElements(BDptr -> numcounters,
   			  descrip -> numvalues,
   			  descrip -> gran,
   			  &descrip -> elemcnt);
   if (!IPR_SUCCESS(Status))
      return Status;

   // compute number of bytes based on the size histogram elements
   hsize = descrip -> elemcnt * sizeof(IPR_HISTELEM);

   Status = KIprHbMaxElements(&maxelem);
   if (!IPR_SUCCESS(Status))
      return IPR__FAILURE;

   allocSize = sizeof (IPR_HBUFFER) + hsize;
#ifdef DBGH
   DbgPrint("Attempting to allocate histogram of size %d (total size %d) with %d elements, max. elements allowed is %d, minval %Lx\n",
         hsize,allocSize,descrip->elemcnt,maxelem,descrip->minval);
//   DbgBreakPoint();
#endif
#ifdef TRCH
   TrcPrint("Attempting to allocate histogram of size %d (total size %d) with %d elements, max. elements allowed is %d, minval %Lx\n",
         hsize,allocSize,descrip->elemcnt,maxelem,descrip->minval);
#endif
//   if (descrip -> elemcnt <= maxelem)
   if (descrip -> elemcnt > maxelem)
      // histogram too large
      return IPR__BIGHIST;
   else
   {
      // OK, histogram isn't too big - create it
      Status = allocate_memory (allocSize,&histogram_template,&Hptr);
      if (!IPR_SUCCESS(Status))
      {
#ifdef TRCE
         TrcPrint("KIprHbCreate Error return from allocate_memory, Status %x\n",
            Status);
#endif
#ifdef DBGH2
         DbgPrint("KIprHbCreate Error return from allocate_memory, Status %x\n",
            Status);
#endif
         return Status;
      }

      if (!Hptr)
         return IPR__ALLOCFAIL;
      all_hb_mem_cnt++;

      Hptr -> elemcnt = descrip -> elemcnt;
      Hptr -> item = descrip -> item;
      Hptr -> cpuid = cpu;
      Hptr -> numvals = descrip -> numvalues;
      Hptr -> pcgran = descrip -> gran;
      Hptr -> index = descrip -> index;
      Hptr -> minval = descrip -> minval;

      if (!BDptr -> pHBuf)
      {
         // No histogram buffers exist yet, create first
         Hptr -> pNext = NULL;
         BDptr -> pHBuf = Hptr;
#ifdef TRCH
         TrcPrint("Creating 1st histogram buffer at %x\n",
            Hptr);
#endif
      }
      else
      {
         // histogram buffers already exist, add one to the end of the list
         HOptr = BDptr -> pHBuf;
	 cnt = 0;
         while (HOptr->pNext)
         {
            HOptr = HOptr->pNext;
#ifdef TRCH
            TrcPrint("Scan for end, HOptr %x\n",
               HOptr);
#endif
            cnt++;
	    if (cnt > IPR_MAX_LOOP_CNT)
	    {
#ifdef TRCE
               TrcPrint("KIprHbCreate Error: Went through loop %d times before getting a match, exiting\n",
                  cnt);
#endif
	       return IPR__PTRLOOPLIM;
	    }
         }
         HOptr->pNext = Hptr;
         Hptr->pNext = NULL;
#ifdef TRCH
         TrcPrint("Creating non 1st histogram buffer at %x, old last %x\n",
            Hptr,HOptr);
#endif
      }
   }

   *retptr = Hptr;
   return IPR__SUCCESS;

}

static void delete_histo_buffer (IN IPR_BUFDSC *pBD, IN IPR_HBUFFER *pHB)
{
   IPR_PBUFFER
      *pPB;
   IPR_PC_RANGE
      *pPCR;
   unsigned int
      i;

   // Kill off PC range references

   pPB = pBD->pPBuf;
   if (pPB)
   {
      pPCR = (IPR_PC_RANGE *)(pPB + 1);
      for (i = 0; i < pPB->rangecnt; i++)
         if (pPCR->pHBuf == pHB)
         {
            pPCR->pHBuf = NULL;
            pPCR->histogram = FALSE;
         }
   }

   // Get rid of the buffer

   deallocate_memory (pHB);
   deall_hb_mem_cnt++;
}

extern IPR_STATUS KIprHbDelete (IN IPR_CPUNUM cpu, IN IPR_HBUFFER *bufptr)
{

   IPR_STATUS Status;
   IPR_BUFDSC
      *BDptr;
   IPR_HBUFFER
      *Hptr, *HOptr;
   int cnt;

#ifdef TRC0
   TrcPrint("KIprHbDelete enter, cpu = %d, bufptr = %x\n",
     cpu,bufptr);
#endif
   if (!bufptr)
   {
#ifdef TRCE
      TrcPrint("KIprHbDelete Error: bufptr is null\n");
#endif
      return IPR__NULLPTR;
   }
   if ((bufptr->header.prod_ident != IPR_PROD_IDENT) ||
      (bufptr->header.bufkind != IPR_HBUFFER_KIND))
   {
#ifdef TRCE
      TrcPrint("KIprHbDelete Error: found buffer prod %Lx kind %x when looking for %Lx %x\n",
         bufptr->header.prod_ident,bufptr->header.bufkind,
         IPR_PROD_IDENT,IPR_HBUFFER_KIND);
#endif
      return IPR__INVBUFKIND;
   }
   Status = get_bufdsc_address (cpu, &BDptr);
   if (!IPR_SUCCESS(Status))
      return Status;
   if (!BDptr)
      return IPR__INVCPU;

   if (!BDptr -> pHBuf)
      return IPR__HBNOTFND;  // No histogram buffers to delete
   else
   {
      // Look for the specified buffer
      if (BDptr -> pHBuf == bufptr)
      {
         // 1st buffer in list to be deleted
         BDptr -> pHBuf = bufptr ->pNext;
#ifdef TRCH
         TrcPrint("Deleting 1st buffer in the list at %x, new BD Hptr is %x, next %x \n",
            bufptr,BDptr->pHBuf,bufptr->pNext);
#endif
#ifdef DBG2
         DbgPrint("HbDelete, deleting 1st histogram buffer at %x\n\n",
         	bufptr);
         DbgBreakPoint();
#endif
         delete_histo_buffer (BDptr, bufptr);
         return IPR__SUCCESS;
      }
      else
      {
         // check the list for the specified buffer
         HOptr = BDptr -> pHBuf;
         Hptr = HOptr -> pNext;
         cnt = 0;
         while (Hptr != NULL)
         {
            if (Hptr == bufptr)
            {
               //  found it
#ifdef TRCH
               TrcPrint("Deleting non-1st buffer in the list at %x, prev. %x, next %x \n",
                  bufptr,HOptr,Hptr);
#endif
#ifdef DBG2
               DbgPrint("HbDelete, deleting non-1st histogram buffer at %x\n\n",
                  bufptr);
               DbgBreakPoint();
#endif
               HOptr -> pNext = Hptr -> pNext;
               delete_histo_buffer (BDptr, bufptr);
               return IPR__SUCCESS;
            }
            else
            {
               //  keep-a-lookin
               HOptr = Hptr;
               Hptr = Hptr -> pNext;
#ifdef TRCH
               TrcPrint("Checking next buffer %x, prev %x\n",
                  Hptr,HOptr);
#endif
            }
            cnt++;
	    if (cnt > IPR_MAX_LOOP_CNT)
	    {
#ifdef TRCE
               TrcPrint("KIprHbDelete Error: Went through loop %d times before getting a match, exiting\n",
                  cnt);
#endif
	       return IPR__PTRLOOPLIM;
	    }
         }  // end - while
      }
   }
   // histogram buffer to delete not found
   return IPR__FAILURE;
}

extern IPR_STATUS KIprHbFind (IN IPR_CPUNUM cpu,
			IN IPR_HBUFFER *previous_buffer,
			OUT IPR_HDESC *descrip,
			OUT IPR_HBUFFER **retptr)
{
   IPR_BUFDSC
      *BDptr;
   IPR_HBUFFER
      *Hptr, *HOptr, *HFptr;
   IPR_STATUS Status;
   int cnt;

#ifdef TRC0
   TrcPrint("KIprHbFind enter, cpu=%d, prevbuf=%x, descrip=%x, retptr=%x\n",
     cpu,previous_buffer,descrip,retptr);
#endif
   if (!descrip)
   {
#ifdef TRCE
      TrcPrint("KIprHbFind Error: descrip is null\n");
#endif
      return IPR__NULLPTR;
   }
   if (!retptr)
   {
#ifdef TRCE
      TrcPrint("KIprHbFind Error: retptr is null\n");
#endif
      return IPR__NULLPTR;
   }
   *retptr = NULL;
   if (previous_buffer)
   {
      if ((previous_buffer->header.prod_ident != IPR_PROD_IDENT) ||
         (previous_buffer->header.bufkind != IPR_HBUFFER_KIND))
      {
#ifdef TRCE
         TrcPrint("KIprHbFind Error: found buffer prod %Lx kind %x when looking for %Lx %x\n",
            previous_buffer->header.prod_ident,previous_buffer->header.bufkind,
            IPR_PROD_IDENT,IPR_HBUFFER_KIND);
#endif
         return IPR__INVBUFKIND;
      }
   }
   Status = get_bufdsc_address (cpu, &BDptr);
   if (!IPR_SUCCESS(Status))
      return Status;
   if (!BDptr)
      return IPR__INVCPU;
   if (!BDptr -> pHBuf)
      // No histogram buffers to search
      return IPR__HBNOTFND;  	// No histogram buffers to search
   HOptr = NULL;
   Hptr = BDptr -> pHBuf;
   // If previous_buffer is NULL start with first buffer, otherwise,
   // find previous buffer in the buffer ring and start there
   if (previous_buffer != NULL)
   {
      // Look for previous buffer
      HFptr = (IPR_HBUFFER *) NULL;
      cnt = 0;
      while (Hptr != NULL)
      {
	 if (HOptr == previous_buffer)
	 {
	    HFptr = Hptr;
#ifdef TRCH
            TrcPrint("Found buffer at %x\n",
               HFptr);
#endif
	    break;
	 }
	 else
	 {
	    HOptr = Hptr;
	    Hptr = Hptr -> pNext;
#ifdef TRCH
            TrcPrint("Checking next buffer %x, prev %x\n",
                  Hptr,HOptr);
#endif
	 }
         cnt++;
         if (cnt > IPR_MAX_LOOP_CNT)
	 {
#ifdef TRC0
            TrcPrint("KIprHbFind Error: Went through loop %d times before getting a match, exiting\n",
               cnt);
#endif
	    return IPR__PTRLOOPLIM;
	 }
      }  // end - while
      if (HFptr == NULL)
	 // Didn't find previous buffer
	 return IPR__HBNOTFND;
      else
	 Hptr = HFptr;
   }
   //  OK, found starting point, now check descriptions
   cnt = 0;
   while (Hptr != NULL)
   {
      if ((descrip -> itemfilter . ipr_hmask_l_bits ==
           BIprAll1 (descrip -> itemfilter . ipr_hmask_l_bits))
          || ((Hptr -> item . ipr_hmask_l_bits
               & descrip -> itemfilter . ipr_hmask_l_bits)
              && Hptr -> index == descrip -> index))
      {
         // Fill in rest of descrip block
         Status = IprHbElements(BDptr->numcounters,
         				Hptr -> numvals,
         				Hptr -> pcgran,
         				&descrip -> elemcnt);
   	 if (!IPR_SUCCESS(Status))
      	     return Status;
         descrip -> minval = Hptr -> minval;
         descrip -> numvalues = Hptr -> numvals;
         descrip -> gran = Hptr -> pcgran;
         descrip -> index = Hptr -> index;
         if ((Hptr->header.prod_ident != IPR_PROD_IDENT) ||
             (Hptr->header.bufkind != IPR_HBUFFER_KIND))
         {
#ifdef TRCE
            TrcPrint("KIprHbFind Error: found buffer prod %Lx kind %x when looking for %Lx %x\n",
               Hptr->header.prod_ident,Hptr->header.bufkind,
               IPR_PROD_IDENT,IPR_HBUFFER_KIND);
#endif
            return IPR__INVBUFKIND;
         }

#ifdef TRCH
         TrcPrint("Histogram match at %x\n",
            Hptr);
#endif
         *retptr = Hptr;
         return IPR__SUCCESS;
      }
      else
      {
         // itemfilter mismatch, check next
         Hptr = Hptr -> pNext;
#ifdef TRCH
         TrcPrint("Checking for match at next Histogram %x\n",
            Hptr);
#endif
      }
      cnt++;
      if (cnt > IPR_MAX_LOOP_CNT)
      {
#ifdef TRC0
         TrcPrint("KIprHbFind Error: Went through loop %d times before getting a match, exiting\n",
            cnt);
#endif
	 return IPR__PTRLOOPLIM;
      }
   }  // end - while
   // Histogram not found
   return IPR__HBNOTFND;
}

static IPR_STATUS FindHistogram (IN IPR_CPUNUM cpu,
		      IN IPR_HBUFFER *bufptr,
		      OUT IPR_BUFDSC **BDptr)
{
   IPR_STATUS Status;
   IPR_HBUFFER
      *Hptr, *HOptr, *HFptr;
   IPR_BUFDSC *tBDptr;
   int cnt;

#ifdef TRC0
   TrcPrint("FindHistogram enter, cpu=%d, bufptr=%x. BDptr=%x\n",
     cpu,bufptr,BDptr);
#endif
   if (!bufptr)
   {
#ifdef TRCE
      TrcPrint("FindHistogram Error: bufptr is null\n");
#endif
      return IPR__NULLPTR;
   }
   if ((bufptr->header.prod_ident != IPR_PROD_IDENT) ||
      (bufptr->header.bufkind != IPR_HBUFFER_KIND))
   {
#ifdef TRCE
      TrcPrint("FindHistogram Error: found buffer prod %Lx kind %x when looking for %Lx %x\n",
         bufptr->header.prod_ident,bufptr->header.bufkind,
         IPR_PROD_IDENT,IPR_HBUFFER_KIND);
#endif
      return IPR__INVBUFKIND;
   }
   Status = get_bufdsc_address (cpu, BDptr);
   if (!IPR_SUCCESS(Status))
      return Status;
   if (!*BDptr)
      return IPR__INVCPU;

   tBDptr = *BDptr;
   if (!tBDptr -> pHBuf)
      return IPR__FAILURE;  	// No histogram buffers

   HOptr = NULL;
   Hptr = tBDptr -> pHBuf;
   HFptr = NULL;
   cnt = 0;
   while (Hptr != NULL)
   {
      if (Hptr == bufptr)
      {
         HFptr = Hptr;
#ifdef TRCH
         TrcPrint("Found matching Histogram at %x\n",
            HFptr);
#endif
         break;
      }
      else
      {
         HOptr = Hptr;
         Hptr = Hptr -> pNext;
#ifdef TRCH
         TrcPrint("Looking for next Histogram at %x, prev %x\n",
            Hptr,HOptr);
#endif
      }
      cnt++;
      if (cnt > IPR_MAX_LOOP_CNT)
      {
#ifdef TRC0
         TrcPrint("FindHistogram Error: Went through loop %d times before getting a match, exiting\n",
            cnt);
#endif
	 return IPR__PTRLOOPLIM;
      }
   }
   if (HFptr == NULL)
      return IPR__FAILURE;	// HFptr points to the buffer that was found

   return IPR__SUCCESS;

}

extern IPR_STATUS KIprHbRead (IN IPR_CPUNUM cpu,
		      IN IPR_HBUFFER *bufptr,
		      IN int start,
		      IN IPR_COUNT cnt,
		      OUT IPR_HISTELEM *retptr)
{
   IPR_STATUS error;
   unsigned int elemcnt;
   IPR_HISTELEM
      *HEptr;
   IPR_BUFDSC
      *BDptr;

#ifdef TRC0
   TrcPrint("KIprHbRead enter, cpu=%d, bufptr=%x, start=%x, cnt=%d, retptr=%x\n",
     cpu,bufptr,start,cnt,retptr);
#endif

   // bufptr prod. and type are checked in FindHistogram
   error = FindHistogram(cpu, bufptr, &BDptr);
   if (error != IPR__SUCCESS)
      return error;
   if (!retptr)
   {
#ifdef TRCE
      TrcPrint("KIprHbRead Error: retptr is null\n");
#endif
      return IPR__NULLPTR;
   }

   error = IprHbElements(BDptr->numcounters,
         		bufptr -> numvals,
         		bufptr -> pcgran,
         		&elemcnt);
   if (!IPR_SUCCESS(error))
      return error;

   if ((start + cnt) > elemcnt)
      return IPR__HISTOVFLW;	// not enough elements to satisfy entire request

   HEptr = (IPR_HISTELEM *) (bufptr + 1);
#ifdef TRCH
   TrcPrint("Moving %d elements from %x\n",
      elemcnt,HEptr);
#endif
#ifdef DBGH
   DbgPrint("KIprHbRead, cpu=%d, bufptr=%x, start=%x, cnt=%d, elem=%d, dest %x, src %x, bytes=%d\n",
     cpu,bufptr,start,cnt,elemcnt,retptr,&HEptr[start],cnt*sizeof(IPR_HISTELEM));
#endif


#ifdef __LINUX__
   copy_to_user(retptr,&HEptr[start],cnt*sizeof(IPR_HISTELEM));
#else 
   MoveMem(retptr,&HEptr[start],cnt*sizeof(IPR_HISTELEM));
#endif


   return IPR__SUCCESS;
}

extern IPR_STATUS KIprHbClear (IN IPR_CPUNUM cpu,
			IN IPR_HBUFFER *bufptr,
			IN int start,
			IN IPR_COUNT cnt)
{
   IPR_STATUS error;
   unsigned int elemcnt;
   unsigned int ecnt;
   IPR_HISTELEM
      *HEptr;
   IPR_BUFDSC
      *BDptr;

#ifdef TRC0
   TrcPrint("KIprHbClear enter, cpu=%d, bufptr=%x, start=%x, cnt=%d\n",
     cpu,bufptr,start,cnt);
#endif
#ifdef DBGH
   DbgPrint("KIprHbClear enter, cpu=%d, bufptr=%x, start=%x, cnt=%d\n",
     cpu,bufptr,start,cnt);
#endif
   // bufptr prod. and type are checked in FindHistogram
   error = FindHistogram(cpu, bufptr, &BDptr);
   if (error != IPR__SUCCESS)
      return error;

   error = IprHbElements(BDptr->numcounters,
        		bufptr -> numvals,
         		bufptr -> pcgran,
         		&elemcnt);
   if (!IPR_SUCCESS(error))
   {
      return error;
   }
#ifdef DBGH
   DbgPrint("start %x, cnt %x, elemcnt %x, bufptr %x, bdptr %x\n",
     start,cnt,elemcnt,bufptr,BDptr);
   DbgBreakPoint();
#endif
   if ((start + cnt) > elemcnt)
      return IPR__FAILURE;	// not enough elements to return

   HEptr = (IPR_HISTELEM *) (bufptr + 1);
#ifdef TRCH
   TrcPrint("Clearing %d elements from %x\n",
      elemcnt,HEptr);
#endif

   for (ecnt = 0; ecnt < cnt; ecnt++)
      // it may be better to do a block move here
      HEptr[ecnt+start] = 0;

   return IPR__SUCCESS;
}

extern IPR_STATUS KIprHbMaxElements(OUT IPR_COUNT *elem)
{
   if (!elem)
      return IPR__NULLPTR;

   *elem = ((IPR_MAX_BUFFER_SIZE-sizeof(IPR_HBUFFER))
      /sizeof(IPR_HISTELEM));
   return IPR__SUCCESS;
}
