/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 *  Module:     IPREDUCE_OS_OSF1.C
 * 
 *  Version:    NONE
 *
 *
 *  Abstract:   This module contains all OSF/1 specific routines for IPROBE
 *              data reduction (IPREDUCE).
 *
 *  Author:         ??
 *  Creation Date:  07-Sep-1994
 *
 *  Modified by:
 *
 *  NONE          Nelson Montalvo
 *
 *  X-2           Michael Dempsey         01-Feb-1995
 *      Reimplemented with headers from the "rationalized kernel" version of 
 *      the IPROBE API. 
 *
 *  X-1           Michael Dempsey         07-Sep-1994
 *      Original, extracted from ipreduce.c
 *
 */

/*** #includes ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include "ipreduce.h"


/*** #defines ***/

#define OSF_IPL_BITMASK		0x7
#define OSF_MODE_BITMASK	0x10
#define OSF_MODE_BITPOS		3

/*** Global Constants (initialized in ipreduce.c) ***/

extern const char *DayOfWeek[];
extern const char *Month[];


/*** Functions ***/

/****************************************************************************/
/*  Name: DateToString                                                      */
/*                                                                          */
/*  Purpose: Converts the passed unsigned int date into a date string.      */
/*                                                                          */
/*  Arguments: Date: the date to be converted to a string.                  */
/*                                                                          */
/*  Outputs: None.                                                          */
/*                                                                          */
/*  Function Value: The passed date, as a string.                           */
/*                                                                          */
/*  Notes: o Converts OSF/1 specific binary time.                           */
/*                                                                          */
/****************************************************************************/
char *DateToString(uint64 Date)
{
  char *pEndOfString;
  static int BufferNumber = 0;
  static char DateBuffer[2][64];

  /* ?? Some OSF/1 time specific stuff... ?? */
  
  BufferNumber = (BufferNumber == 1) ? 0 : 1;
  strcpy(DateBuffer[BufferNumber], ctime((time_t *) &Date));

  /* Place a null terminator at the end of the string. */

  if (pEndOfString = strchr(DateBuffer[BufferNumber], '\n'))
  {
    *pEndOfString = '\0';
  }

  return DateBuffer[BufferNumber];
}


/****************************************************************************/
/*  Name: GetModeSelection                                                  */
/*                                                                          */
/*  Purpose: Given a string of mode requests, this function will parse out  */
/*           the proper modes and set the mode mask appropriately.          */
/*                                                                          */
/*  Arguments: ModeRequest: string assumed to contain the mode requests.    */
/*                                                                          */
/*  Outputs: None.                                                          */
/*                                                                          */
/*  Function Value: an unsigned int mask, where each set bit represents a   */
/*                  a mode.                                                 */
/*                                                                          */
/*  Notes: o #defined modes are in ipr_api_base_defs_nt.h                   */
/*                                                                          */
/****************************************************************************/
uint64 GetModeSelection(char *pModeRequest)
{
  uint64 BadModeRequest;
  uint64 ModeSelection = 0;
  
  /* Set mode mask bits as requested. */
  
  /* Use the #defined constants to determine where to shift the 1. */
  /* These are #defined in ipr_api_base_defs_nt.h                  */

  /* Kernel mode */
  
  if (strchr(pModeRequest, 'k') != NULL)
  {
    ModeSelection |= 1 << IPR_MODE_KERNEL;
  }
  
  /* User mode */
  
  if (strchr(pModeRequest, 'u') != NULL)
  {
    ModeSelection |= 1 << IPR_MODE_USER;
  }
  
  /* Idle mode */
  
  if (strchr(pModeRequest, 'n') != NULL)
  {
    ModeSelection |= 1 << IPR_MODE_IDLE;
  }
  
  /* Complain about any illegal options. */
  
  while ((BadModeRequest = strspn(pModeRequest, "kun")) < strlen(pModeRequest))
  {
    fprintf(stderr, "Unknown mode \"%c\"\n", pModeRequest[BadModeRequest]);
    pModeRequest += BadModeRequest + 1;
  }
  
  return ModeSelection;
}
 

/****************************************************************************/
/*  Name: ParsePS                                                           */
/*                                                                          */
/*  Purpose: Parse the processor status word.                               */
/*                                                                          */
/*  Arguments: pRecord: sample data pulled from file.                       */
/*             pFileHeader: not used.                                       */
/*                                                                          */
/*  Outputs: None.                                                          */
/*                                                                          */
/*  Function Value: None                                                    */
/*                                                                          */
/*  Notes: o Idle mode is inferred if PID is zero.                          */
/*                                                                          */
/****************************************************************************/
void ParsePS (datarecord *pRecord, fileheader *pFileHeader)
{
  pRecord->IPL = (pRecord->PSval & OSF_IPL_BITMASK);
  pRecord->Mode = (pRecord->PSval & OSF_MODE_BITMASK) >> OSF_MODE_BITPOS;

  /* Idle mode inferred? */

	if (pRecord->PID == 0)
  {
    pRecord->Mode = IPR_MODE_IDLE;
  }
}
