/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program Name: iprobe_args.c
 *
 * Abstract:  Command line argument processing for the IPROBE command.
 *            This module should work proprely for VMS, OSF/1 and Windows NT.
 * 
 * Author: Greg Tarsa
 * 
 * Creation Date: Dec-93
 *
 * Modification History
 *
 * J. Goddard			11-Sep-1997
 *	Added the current date/time and version/rev to the help line and
 *	reformatted it slightly.
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni:  Oct-94
 *   Modified the code by using getopt for parsing.
 *
 * Ravindra Kulkarni: Aug-94
 *   Included conditional compilation for unixio.h in case of VMS
 *   unistd.h in case of OSF1 and none in case of NT.
 *   Defined the macro F_OK for use by NT only.
 *
 * Ravindra Kulkarni: Jun-94 
 *  Added new command line options -collect, -nocollect, -add for
 *  specifying the required sampling data.
 *  Added help for specifying above options.
 *
 *  Ravindra Kulkarni: Jun-94, 
 *   Added new command line option buffer_size to read the sample buffer
 *   size.
 *
 *  Tarsa
 *  Added support for pc_ranges.
 *   Moved option table manipulation code to ipr_keylib.c.
 *   Moved option table definitions to ipr_keylib.h.
 *   Added support for multiple methods.
 *   Added support for mode specification.
 *
 *  Greg Tarsa: Apr-94
 *   Merged Ravindra's help changes into main source.
 *
 *  Ravindra Kulkarni:  Mar-94
 *   Added quiet keyword and enhanced help.
 *
 *  Greg Tarsa: Mar-94
 *   Added secret DEBUG qualifier.
 *   Added frequency skipping parameter.
 *   Tabularized the cmd line arg help.
 *   Changed to use spaces rather than equal signs to separate
 *   values from options.
 *   Changed finish time to be a delta time, not an absolute time.
 *   Changed keywords for start and end time.
 *
 *  Ravindra Kulkarni: Feb-94,
 *   Fixed lower bound check error in interval assignment.
 *
 *  Greg Tarsa: Dec-93
 *   Wrote original code.
 *
*/
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include "stdtypes.h"

#ifdef vms
#include <unixio.h>
#endif

#ifdef unix
#include <unistd.h>
#endif

#include "iprobe_struct.h"
#include "iprobe_hdr.h"
#include "iprobe_eventlib.h"
#include "getopt1.def"
#include "getopt.def"
#include "iprobe_global.h"


#ifdef nt
#undef errno
#define errno (GetLastError())
#define F_OK 0
#endif

static char *GetNextOption(char *option_string, char *the_option, int32 reset_flag, int32 *bit_width);

/************************************************************************************/
/* Name: Usage                                                                      */
/*                                                                                  */
/* Purpose:  This prints out the usage for the iprobe command, complete with default*/
/*           values.                                                                */
/*                                                                                  */
/* Arguments: None.                                                                 */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: This seems a little excessive to simply print out a help screen.          */
/*                                                                                  */
/************************************************************************************/


static void Usage(void)
{
  char Field[BUFSIZ];
  char *ArguementType;
  int32 CurrentSampleDataType;
  struct keywords *CurrentOptionPointer;
  struct keywords *OptionTable;
  struct keywords *MethodTable;
  struct keywords *ModeTable;
  struct keywords *SampleDataTable;


  MethodTable = GetMethodTable();
  OptionTable = GetOptionTable();
  ModeTable = OSGetModeTable();
  SampleDataTable = OSGetSampleDataTable();


  printf("IPROBE V%d.%02d compiled on " __DATE__ " at " __TIME__
	"\n\tDigital Equipment Corporation,"
	"\n\tCopyright (c) 1993, 1994, 1997, 1998\n\n", IPROBE_MAJOR_VER, IPROBE_MINOR_VER);


  /* Help column is specified to left justify the text */
#define HELP_COL0_LEN -18
#define HELP_COL1_LEN -45

  (void) printf("Usage: iprobe [%cflag [%cflag...]] [eventspec [eventspec...]]\n flags, one of:\n\n",
		FLAG_CHAR1, FLAG_CHAR1);
  (void) fflush(stdout);

  /* Move through the table and print out each option and its description */
  for (CurrentOptionPointer = OptionTable; CurrentOptionPointer->name; CurrentOptionPointer++)
  {


    /* Skip invisible optons */

    if (CurrentOptionPointer->attr == A_INVISIBLE && !(*IsDebug()))
      continue;

    /* put the flag char and keyword into a Field */

    (void) sprintf(Field, "        %c%s", FLAG_CHAR1, CurrentOptionPointer->name);

    /*
       Put the Field and the default on the same line, unless the Field
       is longer then the Help Column, then put the default on the next line.
       */

    if ((int)strlen(Field) <= abs(HELP_COL0_LEN))
      (void) printf("%*s %s\n", HELP_COL0_LEN, Field, CurrentOptionPointer->desc);
    else
      (void) printf("%s\n%*s\n",
		    Field, abs(HELP_COL0_LEN)+strlen(CurrentOptionPointer->desc)+1, CurrentOptionPointer->desc);
  }   /* end for "each option in the table" */

  (void) putchar('\n');
  (void) fflush(stdout);

  /* Print default values */
  (void) printf("\n%-*s %s\n%-*s %s\n",
		HELP_COL1_LEN, "        Flag name", "Default",
		HELP_COL1_LEN, "        ~~~~~~~~~", "~~~~~~~");

  /* Move through the table and print out each option and its default */
  for (CurrentOptionPointer = OptionTable; CurrentOptionPointer->name; CurrentOptionPointer++)
  {

    /* Skip invisible optons */
    if (CurrentOptionPointer->attr == A_INVISIBLE && !(*IsDebug()))
      continue;

    /* argument descriptions are based on the type */

    switch ((uint32)CurrentOptionPointer->type)
    {
    case T_REQ_NUM:
      ArguementType = " integer";
      break;
    case T_REQ_FILE:
      ArguementType = " filename";
      break;
    case T_NONE:
      ArguementType = "";
      break;
    case T_REQ_METH:
      ArguementType = " measuring-type";
      break;
    case T_REQ_MODE:
      ArguementType = " mode-type";
      break;
    case T_REQ_RSPEC:
      ArguementType = " range-spec";
      break;
    case T_REQ_STRING:
      ArguementType = " quoted string";
      break;
    default:
      ArguementType = "";
      break;
    }

    /* put the flag char, keyword, and value into a Field */
    (void) sprintf(Field, "        %c%s%s%s",
		   FLAG_CHAR1, CurrentOptionPointer->name, ArguementType,
		   (CurrentOptionPointer->attr&A_REPEATING) ? " (R)" : "");

    /*
       Put the Field and the default on the same line, unless the Field
       is longer then the Help Column, then put the default on the next line.
       */

    if ((int)strlen(Field) <= abs(HELP_COL1_LEN))
      (void) printf("%*s %s\n", HELP_COL1_LEN, Field, CurrentOptionPointer->def_val);
    else
      (void) printf("%s\n%*s\n",
		    Field, abs(HELP_COL1_LEN)+strlen(CurrentOptionPointer->def_val)+1, CurrentOptionPointer->def_val);
  }   /* end for "each option in the table" */

  /* Add extra, explanatory texts. */

  (void) printf(
		"\n\tflags should begin with either '%c'\n\n",
		FLAG_CHAR1);
  (void) printf(
		"\tflags marked (R) can be specified multiple times\n\n");
  (void) printf(
		"\teventspec:\n\
        event[.freq][.skip]\n\n");
  (void) printf(
		"\tevent is the name of the event to be measures, and must be one of\n\
        those appearing in the list that follows.\n\n");
  (void) printf(
		"\tfreq is the frequency at which interrupts are to be generated\n\
        and can be HIGH, LOW, or a value.\n\n");
  (void) printf(
		"\tskip indicates that all but the skip-th interrupt is to be\n\
        ignored and must be a value. (ignored for all but sampling)\n\n");

  (void) printf(
		"\tThe valid measuring-types are:\n\n");

  /* Move through the method table and print out each method and its descr */

  for (CurrentOptionPointer = MethodTable; CurrentOptionPointer->name; CurrentOptionPointer++)
  {

    /* Skip invisible methods */

    if (CurrentOptionPointer->attr == A_INVISIBLE && !(*IsDebug()))
      continue;

    /* put the keyword into a Field */
    (void) sprintf(Field, "        %s", CurrentOptionPointer->name);

    /*
       Put the Field and the default on the same line, unless the Field
       is longer then the Help Column, then put the default on the next line.
       */

    if ((int)strlen(Field) <= abs(HELP_COL0_LEN))
    {
      (void) printf("%*s %s\n", HELP_COL0_LEN, Field, CurrentOptionPointer->desc);
    }
    else
    {
      (void) printf("%s\n%*s\n",
		    Field, abs(HELP_COL0_LEN)+strlen(CurrentOptionPointer->desc)+1, CurrentOptionPointer->desc);
    }
  }   /* end for "each method in the table" */

  (void) putchar('\n');

  (void) printf(
		"\tFinally, the valid mode-types are:\n\n");

#undef HELP_COL0_LEN
#define HELP_COL0_LEN -20

  /* Move through the mode list and print out each mode and its descr */
  for (CurrentOptionPointer = ModeTable; CurrentOptionPointer->name; CurrentOptionPointer++)
  {


    /* Skip invisible methods */

    if (CurrentOptionPointer->attr == A_INVISIBLE && !(*IsDebug()))
      continue;

    /* put the keyword into a Field */
    (void) sprintf(Field, "        %s", CurrentOptionPointer->name);

    /*
       Put the Field and the default on the same line, unless the Field
       is longer then the Help Column, then put the default on the next line.
       */

    if ((int)strlen(Field) <= abs(HELP_COL0_LEN))
    {
      (void) printf("%*s %s\n", HELP_COL0_LEN, Field, CurrentOptionPointer->desc);
    }
    else
    {
      (void) printf("%s\n%*s\n",
		    Field, abs(HELP_COL0_LEN)+strlen(CurrentOptionPointer->desc)+1, CurrentOptionPointer->desc);
    }  
}   /* end for "each method in the table" */

  (void) putchar('\n');


  (void) printf("\tThe various data types that can be collected are as follows.\n");
  (void) printf("\tThese options can be used with following options\n");
  (void) printf("\t-collect -add and -nocollect.\n\n");

  for (CurrentSampleDataType = 0; SampleDataTable[CurrentSampleDataType].name; CurrentSampleDataType++)
  {

    /* put the keyword into a Field */

    (void) sprintf(Field, "        %s", SampleDataTable[CurrentSampleDataType].name);

    if ((int)strlen(Field) <= abs(HELP_COL0_LEN))
    {
      (void) printf("%*s %s\n", HELP_COL0_LEN, Field,
		    SampleDataTable[CurrentSampleDataType].desc);
    }
    else
    {
      (void) printf("%s\n%*s\n", Field, abs(HELP_COL0_LEN) +
		    strlen(SampleDataTable[CurrentSampleDataType].desc) + 1,
		    SampleDataTable[CurrentSampleDataType].desc);
    }
  }

  (void) putchar('\n');


  (void) fflush(stdout);

  (void) DisplayColumnHelp();

  (void) fflush(stdout);




}

/*
    ProcessArguments

    Process the command line arguments and return
    the input, output and error files derived from them
    as well as any options specified.

    Returns the number of errors encountered.
*/



/************************************************************************************/
/* Name: ProcessArguments                                                           */
/*                                                                                  */
/* Purpose: This takes a command line, and extracts all of the useful information   */
/*          from it.                                                                */
/*                                                                                  */
/* Arguments: argc: The number of command line arguments.                           */
/*            argv:  Pointers to each command line argument.                        */
/*            Flags: Return flags for which options set                             */
/*            TotalNumberOfBuffers: Number of buffers specified                     */
/*            BufferSize: The size of one sample buffer                             */
/*            StartTime: Returns number of secs till start                          */
/*            Duration:  Returns the number of secs to measure                      */
/*            Interval: Return length of interval                                   */
/*            OutputFileName: Return name of sampling file                          */
/*            Method: Return method of measuring                                    */
/*            ModeMask:  Returns Program modes to measure                           */
/*            DataMask: Data mask for sample data                                   */
/*            QuadDataMask:  Data mask for sample data                              */
/*            PCRange:  Pointer to lnked list of pc ranges                          */
/*            EventArray: Return array of events to measure                         */
/*            FrequencyArray: Return array of freqs to measure at                   */
/*            SkipArray: Return array of skip values                                */
/*            ArraySize: Return length of prev 2 arrays                             */
/*	      CommandLine: Pointer to command line buffer                           */
/*	      CommandLineLength: usable size of CommandLine                         */
/*                                                                                  */
/* Output: See Arguments.                                                           */
/*                                                                                  */
/* Function Value: Returns the number of errors encountered.                        */
/*                                                                                  */
/* Notes: Perhaps this could be broken down into smaller modules.                   */
/*                                                                                  */
/************************************************************************************/

int32 ProcessArguments(int32 argc, char **argv, /*  command line info   */
		       FLAGS *Flags,          /*  ret flags for which options set */
		       IPR_COUNT *TotalNumberOfBuffers, /*  ret # of buffers specified  */
		       IPR_COUNT *BufferSize,  /*  buffer size of sample buffers   */
		       time_t *StartTime,         /*  ret # of secs till start    */
		       time_t *Duration,        /*  ret # of secs to measure    */
		       IPR_COUNT *Interval,     /*  ret length of interval  */
		       char **OutputFileName,        /*  ret name of sampling file   */
		       FLAGS *Method,         /*  ret method of measuring     */
		       IPR_MMASK *ModeMask,    /*  ret prog modes to measure   */
		       IPR_DMASK *DataMask,    /*  data mask for sample data */
		       IPR_DMASK *QuadDataMask,    /*  data mask for sample data */
		       struct pc_ranges *PCRange,/*  ptr to lnked list of pc ranges  */
		       IPR_EVENT *EventArray,   /*  ret array of events to measure  */
		       IPR_FREQ *FrequencyArray,   /*  ret array of freqs to measure at    */
		       IPR_COUNT *SkipArray,  /*  ret array of skip values    */
		       IPR_COUNT *ArraySize, /* ret length of prev 2 arrays */
		       char *CommandLine,	/* pointer to command line buffer */
		       size_t CommandLineLength) /* usable size of CommandLine */
{
  char *ArgumentValue;
  char *EventName;     /* current event spec */
  char *IrqlString = "irql";
  char OptionBuffer[80];
  char *OptionName;
  char *ThisOption;

  int32 CurrentArgument;
  int32 CurrentOptionEntry;
  int32 CurrentTableEntry;
  int32 DataWidth = 0;
  int32 ErrorCount = 0;         /* count of errors*/
  int32 EventCount = 0;     /* count of event specifications */
  int32 MethodFlag;
  int32 ModeFlag;
  int32 OptionCharacter;
  int32 ResetFlag; 
  int32 Status;
  int32 TempArgc = argc;

  int OptionIndex; /* This is left as int because VMS won't let it compile otherwise... */


  system_info *SystemInformation;


  struct keywords *OptionPointer;
  struct keywords *OptionPointer2;
  struct pc_ranges *TempPCRangePointer;  /* ptr to last list element (0 = none) */
  struct option LongOptions[30];
  struct keywords *MethodTable;
  struct keywords *OptionTable;
  struct keywords *ModeTable;
  struct keywords *SampleQuadDataTable;
  struct keywords *SampleDataTable;


  time_t CurrentTime;         /* current time in absolute format */

  uint32 TempDataMask = DEFAULT_DATA;
  uint32 TempQuadDataMask = DEFAULT_QUADDATA;
  uint32 *TotalNumberOfCounters;

  ModeTable = OSGetModeTable();
  MethodTable = GetMethodTable();
  OptionTable = GetOptionTable();
  SampleQuadDataTable = OSGetSampleQuadDataTable();
  SampleDataTable = OSGetSampleDataTable();


  TotalNumberOfCounters = GetTotalNumberOfCounters();

#ifdef nt
  /*  The command line is converted to lower case in case of NT only  */
  while(TempArgc--)
    ConvertStringToLowerCase(argv[ TempArgc ]);
  /*	In case user enters method ipl instead of irql on nt it is
    	internally converted to irql.	*/

  for (CurrentArgument = 0; CurrentArgument < argc; CurrentArgument++)
  {
    if ((strcmp(argv[CurrentArgument], "ipl") == 0) ||
	(strcmp(argv[CurrentArgument], "ip") == 0))
      argv[CurrentArgument] = IrqlString;
  }
#endif


  SystemInformation = GetSystemInformation();

  for (CurrentOptionEntry = 0; OptionTable[CurrentOptionEntry].name != NULL; CurrentOptionEntry++)
  {
    LongOptions[CurrentOptionEntry].name = OptionTable[CurrentOptionEntry].name;
	
    LongOptions[CurrentOptionEntry].has_arg = OptionTable[CurrentOptionEntry].type == T_NONE ? no_argument : required_argument;
    LongOptions[CurrentOptionEntry].flag = NULL;
    LongOptions[CurrentOptionEntry].val = 0;
  }

  LongOptions[CurrentOptionEntry].name = NULL;
  LongOptions[CurrentOptionEntry].has_arg = no_argument;
  LongOptions[CurrentOptionEntry].flag = NULL;
  LongOptions[CurrentOptionEntry].val = 0;

  /* Init the list element pointer */
  TempPCRangePointer = PCRange;

  /* Establish the current time */
  CurrentTime = time(0);

  /* init the output variables to defaults */
  *TotalNumberOfBuffers = (IPR_COUNT)atoi(FindOptionByNumber(T_BUFFER_COUNT)->def_val);

  *BufferSize = SystemInformation->pagelength;    /*  default buffer size is one page length.    */

  *StartTime    = (IPR_COUNT) atoi(FindOptionByNumber(T_BEGIN_TIME)->def_val);
  *Duration   = (IPR_COUNT) atoi(FindOptionByNumber(T_END_TIME)->def_val);
  *Interval     = (IPR_COUNT) atoi(FindOptionByNumber(T_INTERVAL)->def_val);
  *OutputFileName      = FindOptionByNumber(T_OUTPUT)->def_val;
  *Method   = 0;  /* special default code at end */
  strcpy(CommandLine, FindOptionByNumber(T_COMMAND)->def_val);

  *Flags    = 0;
  DataMask->ipr_dmask_l_bits = TempDataMask;
  QuadDataMask->ipr_dmask_l_bits = TempQuadDataMask;
  
  /* Set mode mask to default of collecting data in all modes. */

  ModeMask->ipr_mmask_l_bits = (uint32) atoi(FindOptionByNumber(T_MODE)->def_val);

  /* Init event and frequency arrays */
  (void) memset(EventArray, 0, (*TotalNumberOfCounters) * sizeof(IPR_EVENT));
  (void) memset(FrequencyArray, 0, (*TotalNumberOfCounters) * sizeof(IPR_FREQ));
  (void) memset(SkipArray, 0, (*TotalNumberOfCounters) * sizeof(IPR_COUNT));

  /* skip past program name and process any flags */

  while(TRUE)
  {

    OptionCharacter = getopt_long_only(argc, argv, "", LongOptions, &OptionIndex);

    if (OptionCharacter == EOF)
    {
      break;
    }

    if (OptionCharacter == '?')
    {
      ThisOption = "help";
    }
    else
    {
      ThisOption = (char *)LongOptions[OptionIndex].name;
    }
    /* Look the option up in the table */
    OptionPointer2 = FindOptionByName(ThisOption);
    if (OptionPointer2->num <= 0)   /* Unknown option: error */
    {
      ErrorCount++;
      continue;
    }

    switch ((uint32)OptionPointer2->type)
    {
    case T_REQ_NUM:
    case T_REQ_FILE:
    case T_REQ_METH:
    case T_REQ_MODE:
    case T_REQ_RSPEC:
    case T_REQ_STRING:
      /*
	 An argument is required here; so shift to the next
	 argument, if there is one.

	 Error if there is not, or if the next argument is a
	 flag.  Error action is to flag the error, increment
	 the count, unshift, and iterate the parsing loop.
	 */
      if (optarg == NULL)
      {
	(void) fprintf(stderr, "?Required argument to %c%s missing",
		       FLAG_CHAR1, ThisOption);
	ErrorCount++;
	continue;
      }


      if (*optarg == FLAG_CHAR1)
      {
	(void) fprintf(stderr,
		       "?Found (%s) when looking for required argument to %c%s\n",
		       *argv, FLAG_CHAR1, ThisOption);
	ErrorCount++;
	continue;
      }

      ArgumentValue = optarg;
      break;

    case T_NONE:
      ArgumentValue = 0;
      break;

    default:
      (void) fprintf(stderr, "%% Unknown argument type specifier: %d",
		     OptionPointer2->type);
      break;

    }/* end switch(type) */

    if (OptionPointer2->num == T_HELP)
    {
      *Flags |= CL_HELP;
      continue;
    }

    if (OptionPointer2->num == T_DEBUG)
    {
      /* set the global debug flag */
      (*IsDebug()) = TRUE;
      continue;
    }
    if (OptionPointer2->num == T_QUIET)
    {
      *Flags |= CL_QUIET;
      (*IsQuiet()) = TRUE;
      continue;
    }

    if (OptionPointer2->num == T_NO_INTERMEDIATE_RESULTS)
    {
      (*IsNoIntermediate()) = TRUE;
      continue;
    }

    if (OptionPointer2->num == T_PERCPU)
    {
      *Flags &= ~(CL_PEREVENT);   /* clear the CL_PEREVENT flag */
      continue;
    }
    if (OptionPointer2->num == T_PEREVENT)
    {
      *Flags |= CL_PEREVENT;
      continue;
    }

    if (OptionPointer2->num == T_BUFSIZ)
    {
      *Flags |= CL_BUFSIZ;
      *BufferSize = (IPR_COUNT)atoi(ArgumentValue);
      continue;
    }


    if (OptionPointer2->num == T_BUFFER_COUNT)
    {
      *Flags |= CL_BUFFER_COUNT;
      *TotalNumberOfBuffers = atoi(ArgumentValue);
      if (*TotalNumberOfBuffers < MIN_BUFFER_COUNT
	  || *TotalNumberOfBuffers > MAX_BUFFER_COUNT)
      {
	(void) fprintf(stdout,
		       "? Error: BUFFER_COUNT must be between %d and %d\n",
		       MIN_BUFFER_COUNT, MAX_BUFFER_COUNT);
	ErrorCount++;
      }
      continue;
    }
    if (OptionPointer2->num == T_BEGIN_TIME)
    {
      *Flags |= CL_BEGIN_TIME;
      *StartTime = atoi(ArgumentValue);
      if (*StartTime >= MIN_TIME && *StartTime <= MAX_TIME)
      {
	/* convert start time into an absolute time */
	*StartTime += CurrentTime;
      }
      else
      {
	(void) fprintf(stdout,
		       "? Error: BEGINNING_TIME must be between %d and %u\n",
		       0, -1);
	ErrorCount++;
      }
      continue;
    }
    if (OptionPointer2->num == T_END_TIME)
    {
      *Flags |= CL_END_TIME;
      *Duration = atoi(ArgumentValue);
      if (*Duration >= MIN_TIME && *Duration <= MAX_TIME)
      {
	/* No action   */
      }
      else
      {
	(void) fprintf(stdout,
		       "? Error: ENDING_TIME must be between %d and %u\n",
		       0, -1);
	ErrorCount++;
      }
      continue;
    }
    if (OptionPointer2->num == T_INTERVAL)
    {
      *Flags |= CL_INTERVAL;
      *Interval = atoi(ArgumentValue);
      if (*Interval < MIN_INTERVAL || *Interval > MAX_INTERVAL)
      {
	(void) fprintf(stdout,
		       "? Error: INTERVAL must be between %d and %d\n",
		       MIN_INTERVAL, MAX_INTERVAL);
	ErrorCount++;
      }
      continue;
    }
    if (OptionPointer2->num == T_RANGE)
    {
      *Flags |= CL_RANGE;
      /* Allocate and initialize a list element */
      TempPCRangePointer->next=(struct pc_ranges *)malloc(sizeof(struct pc_ranges));
      (void) memset(TempPCRangePointer->next, 0, sizeof(struct pc_ranges));

      /* Parse the passed pc range */
      Status = ParsePCRange(ArgumentValue,  /* pc range specification */
			    &TempPCRangePointer->next->start,
			    &TempPCRangePointer->next->end,
			    &TempPCRangePointer->next->granularity);
      if (Status) /* If error, then deallocate and give error */
      {
	free(TempPCRangePointer->next);
	TempPCRangePointer->next == 0;
	ErrorCount++;
	continue;
      }

      TempPCRangePointer = TempPCRangePointer->next;    /* success: adjust ptr */
      continue;
    }
    if (OptionPointer2->num == T_OUTPUT)
    {
      *Flags |= CL_OUTFILE;

      /*
	 if file exists or file does not exist (and path does), then
	 all is well, else error.
	 */
      Status = access(ArgumentValue, F_OK);
      if ((Status == 0) || (Status && errno == ENOENT))
	*OutputFileName = ArgumentValue;
      else
      {
	perror("?Error: output file %s");
	ErrorCount++;
      }

      continue;
    }

    if (OptionPointer2->num == T_INPUT)
    {
      *Flags |= CL_INPUT;

      /*
	 if file exists or file does not exist (and path does), then
	 all is well, else error.
	 */
      Status = access(ArgumentValue, F_OK);
      if ((Status == 0) || (Status && errno == ENOENT))
	*OutputFileName = ArgumentValue;
      else
      {
	perror("?Error: output file %s");
	ErrorCount++;
      }

      continue;
    }

    if (OptionPointer2->num == T_COLLECT)
    {

      ResetFlag = 1; /*  Flag to clear internal static variable
			 belonging to GetNextOption.   */

      TempDataMask =0x00;
      TempQuadDataMask =0x00;


      *Flags |= CL_COLLECT;

      /*  Now read all the options    */
      /*  Here the options is a option list separated by commas. One by
	  one each option is extracted and the corresponding flag
	  is updated. */

      while ((OptionName = GetNextOption(ArgumentValue, OptionBuffer,
					 ResetFlag, &DataWidth)))
      {
	ResetFlag = 0;

	if ((OptionPointer = FindByName(SampleDataTable, OptionName))&&
	    DataWidth == 0 &&(strcmp(OptionPointer->name,"*UNKNOWN*")!=0))
	  TempDataMask |= OptionPointer->num;
	else if ((OptionPointer = FindByName(SampleQuadDataTable,
					     OptionName))&& DataWidth != 0 &&(strcmp(OptionPointer->name,
										     "*UNKNOWN*")!=0))
	  TempQuadDataMask |= OptionPointer->num;
	else
	{
	  (void) fprintf(stderr, "Invalid data specified.\n");
	  (void) fprintf(stderr, "It must be one among following\n");

	  for (CurrentTableEntry = 0; SampleDataTable[CurrentTableEntry].name; CurrentTableEntry++)
	  {
	    (void) fprintf(stderr, "%4s %s\n", SampleDataTable[CurrentTableEntry].name,
			   SampleDataTable[CurrentTableEntry].desc);
	  }
	  (void) fprintf(stderr, "The sample data can be optionally "
			 "followed by a '.' and either 32 (or low) or "
			 "64 (or high)\n");

	  ErrorCount++;
	  break;
	}
      }

      DataMask->ipr_dmask_l_bits = TempDataMask;
      QuadDataMask->ipr_dmask_l_bits = TempQuadDataMask;

      continue;
    }

    if (OptionPointer2->num == T_NOCOLLECT)
    {

      ResetFlag = 1; /*  Flag to clear internal static variable
			 belonging to GetNextOption.   */

      TempDataMask = DataMask->ipr_dmask_l_bits;
      TempQuadDataMask = QuadDataMask->ipr_dmask_l_bits;

      *Flags |= CL_NOCOLLECT;


      while ((OptionName = GetNextOption(ArgumentValue, OptionBuffer,
					 ResetFlag, &DataWidth)))
      {
	ResetFlag = 0;

	if ((OptionPointer = FindByName(SampleDataTable, OptionName))&&
	    DataWidth == 0)
	  TempDataMask &= ~OptionPointer->num;
	else if ((OptionPointer = FindByName(SampleDataTable,
					     OptionName))&& DataWidth != 0)
	  TempQuadDataMask &= ~OptionPointer->num;
	else
	{
	  (void) fprintf(stderr, "Invalid data specified.\n");
	  (void) fprintf(stderr, "It must be one among following\n");
	  
	  for (CurrentTableEntry = 0; SampleDataTable[CurrentTableEntry].name; CurrentTableEntry++)
	  {
	    (void) fprintf(stderr, "%s %s\n", SampleDataTable[CurrentTableEntry].name,
			   SampleDataTable[CurrentTableEntry].desc);
	  }
	  (void) fprintf(stderr, "The base data can be optionally \n"
			 "followed by a '.' and either 32 (or low) or \n"
			 "64 (or high)\n");

	  ErrorCount++;
	  break;
	}
      }

      DataMask->ipr_dmask_l_bits = TempDataMask;
      QuadDataMask->ipr_dmask_l_bits = TempQuadDataMask;

      continue;
    }


    if (OptionPointer2->num == T_ADD)
    {

      ResetFlag = 1; /*  Flag to clear internal static variable
			 belonging to GetNextOption.   */

      *Flags |= CL_ADD;

      /*  Now read all the options    */
      /*  Here the options is a option list separated by commas. One by
	  one each option is extracted and the corresponding flag
	  is updated. */

      while ((OptionName = GetNextOption(ArgumentValue, OptionBuffer, ResetFlag, &DataWidth)))
      {
	ResetFlag = 0;

	if ((OptionPointer = FindByName(SampleDataTable, OptionName)) && DataWidth == 0)
	{
	  TempDataMask |= OptionPointer->num;
	}
	else if ((OptionPointer = FindByName(SampleDataTable, OptionName))&& DataWidth != 0)
	{
	  TempQuadDataMask |= OptionPointer->num;
	}
	else
	{
	  (void) fprintf(stderr, "Invalid data specified.\n");
	  (void) fprintf(stderr, "It must be one among following\n");
	  for (CurrentTableEntry  = 0; SampleDataTable[CurrentTableEntry].name; CurrentTableEntry++)
	  {
	    (void) fprintf(stderr, "%s %s\n", SampleDataTable[CurrentTableEntry].name,
			   SampleDataTable[CurrentTableEntry].desc);
	  }
	  (void) fprintf(stderr, "The base data can be optionally \n"
			 "followed by a '.' and either 32 (or low) or \n"
			 "64 (or high)\n");

	  ErrorCount++;
	  break;
	}
      }

      DataMask->ipr_dmask_l_bits = TempDataMask;
      QuadDataMask->ipr_dmask_l_bits = TempQuadDataMask;

      continue;
    }


    if (OptionPointer2->num == T_COMMAND)
    {
      if (strlen(ArgumentValue) > 0 && strlen(ArgumentValue) <= CommandLineLength)
      {
	strcpy(CommandLine, ArgumentValue);
      }
      else
      {
	(void) fprintf(stdout,"? Command line null or longer than %d characters\n", CommandLineLength);
	ErrorCount++;
      }

      continue;
    }

    if (OptionPointer2->num == T_METHOD)
    {


      *Flags |= CL_METHOD;

      /* Lookup the method in the table */
      MethodFlag = FindMethodByName(ArgumentValue)->num;

      /* must be one of the valid methods, or else error */
      if (MethodFlag != 0)
	{	
	*Method |= MethodFlag;
	}
      else
      {
	(void) fprintf(stdout,"? Invalid METHOD keyword.  Must be one of:\n");
	/* Go thru table and print the name of each method. */
	for (OptionPointer2 = MethodTable; OptionPointer2->name; OptionPointer2++)
	{
	  /* skip invisible options */
	  if (OptionPointer2->attr == A_INVISIBLE && !(*IsDebug()))
	    continue;

	  (void) printf("\t%s\n", OptionPointer2->name);
	}/* end for "each method in the table" */

	ErrorCount++;
      }
      continue;
    }
    if (OptionPointer2->num == T_MODE)
    {
      *Flags |= CL_MODE;

      /* Lookup the method in the table */
      ModeFlag = FindModeByName(ArgumentValue)->num;

      /* must be one of the valid methods, or else error */
      if (ModeFlag != 0)
      {
	ModeMask->ipr_mmask_l_bits |= ModeFlag;
      }
      else
      {
	(void) fprintf(stdout, "? Invalid MODE keyword.  Must be one of:\n");

	/* Go thru table and print the name of each method. */
	for (OptionPointer2 = ModeTable; OptionPointer2->name; OptionPointer2++)
	{
	  /* skip invisible options */
	  if (OptionPointer2->attr == A_INVISIBLE && !(*IsDebug()))
	    continue;

	  (void) printf("\t%s\n", OptionPointer2->name);
	}/* end for "each mode in the table" */

	ErrorCount++;
      }
      continue;
    }
    /* final cases: error */
    else if (CaseInsensitiveStringCompare(ThisOption, "")) /* flag char w/o flag name */
    {
      if (ArgumentValue == 0)
      {
	(void) fprintf(stdout, "Iprobe: Value required for %cERROR flag\n", ThisOption);
	ErrorCount++;
	continue;
      }
      continue;
    }
    else            /* unrecognized flag */
    {
      /* Indicate any unknown flags */
      (void) fprintf(stdout, "?Ignoring flag: %s\n", *argv);
      ErrorCount++;
      continue;
    }

  }/* end "for each option" */

  /* Get the event specifications */
  for (EventCount = 0;                /* init event count */
       optind < argc && EventCount < (*TotalNumberOfCounters);   /* while args and legal evts */
       optind++, EventCount++)      /* next arg, incr evt count */
  {
    EventName = argv[optind];        /* current event spec */
printf("event: %s\n", EventName);

    /* Clear the current event and freq entries */
    EventArray[EventCount] = FrequencyArray[EventCount] = 0;

    /* parse the event specification */
    Status = ParseEventString(EventName,    /* evt spec */
			      &EventArray[EventCount],    /* event array */
			      &FrequencyArray[EventCount],   /* freq array */
			      &SkipArray[EventCount]);  /* skip array */
    if (Status) /* If error, incr count and loop */
    {
      ErrorCount++;
      continue;
    }
  }

  if (optind >= 0 && EventCount == 0 && (*Flags & CL_HELP) == 0)
  {
    if ((*IsDebug()))
      (void) fprintf(stdout, "%% Defaulting event spec to %s\n",
		     IprEventNumberToName (DEF_EVENT));
    EventArray[0] = DEF_EVENT;
    FrequencyArray[0] = FREQ_LOW;
    EventCount++;
  }

  if (EventCount > (*TotalNumberOfCounters))
  {
    (void) fprintf(stdout, "?ERROR: too many events specified: %d > %d\n",
		   EventCount, (*TotalNumberOfCounters));
    ErrorCount++;
  }

  /* Any extra arguments? */
  if (optind < argc)
  {
    (void) fprintf(stdout, "?Too many arguments: \"");
    while (optind < argc)
      (void) fprintf(stdout, " %s", argv[optind++]);
    (void) fprintf(stdout, "\"\n");
    ErrorCount++;
  }

  /* Now exit, if there were errors */
  if (ErrorCount)
  {
    (void) fprintf(stdout, "\nSpecify \"-help\" for usage information\n");
  }
  /* Default the method, if unspecified */
  if (*Method == 0)
  {   
    *Method = METH_COUNT;
  }

  /* Output the help, if requested */
  if (*Flags & CL_HELP)
  {
    /*
       If the help is to go to an output file, then open that
       file on standard output.
       */
    if (*Flags & CL_OUTFILE)
      (void) freopen(*OutputFileName, "w", stdout);

    Usage();
    exit(0);
  }

  /*
     At this point all argument parsing should be complete
     So create, on the heap, a stack version of the event and
     frequency arrays and return pointers to them.
     */
  *ArraySize = EventCount;

  return ErrorCount;
}/* end ProcessArguments() */

/* Here datawidth is set to 0 for a 32 bit value and to 1 for a 64 bit value */

/************************************************************************************/
/* Name: GetNextOption                                                              */
/*                                                                                  */
/* Purpose: Datawidth is set to 0 for a 32 bit value and to 1 for a 64 bit value.   */
/*                                                                                  */
/* Arguments:  OptionString: The unparsed option string, (Could have multiple       */
/*                           options on it.                                         */
/*             TheOption: The particular option that we are working on.             */
/*             ResetFlag: This resets the scanning to the begining.                 */
/*             DataWidth: This is 0 for a 32 bit value. 1 for a 64 bit value.       */
/*                                                                                  */
/* Output: DataWidth: This is the datawidth of the current                          */
/*                                                                                  */
/* Function Value: This is an individual option within the OptionString.            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

char *GetNextOption(char *OptionString, char *TheOption, int32 ResetFlag, int32 *DataWidth)
{
  struct keywords *OptionPointer;
  static int32 CurrentPosition = 0;
  int32 CurrentPositionInOption = 0;
  struct keywords *SampleDataTable;
  struct keywords *SampleQuadDataTable;
    
  char *TempCharacterPointer;

  SampleDataTable = OSGetSampleDataTable();

  SampleQuadDataTable = OSGetSampleQuadDataTable();





  if (ResetFlag)
  {
    CurrentPosition = 0;
  }

  if (OptionString[CurrentPosition] == 0)
  {
    return NULL;
  }

  while((OptionString[CurrentPosition] != 0) && (OptionString[CurrentPosition] != ','))
  {
    TheOption[CurrentPositionInOption++] = OptionString[CurrentPosition++];
  }

  TheOption[CurrentPositionInOption] = 0;

  if (OptionString[CurrentPosition] == ',')
  {
    CurrentPosition++;
  }

  TempCharacterPointer = strchr(TheOption, '.');
  if(TempCharacterPointer == NULL)
  {
    if ((OptionPointer = FindByName(SampleDataTable, TheOption)) && (strcmp(OptionPointer->def_val,"") != 0))
    {
      if (strcmp(OptionPointer->def_val, "32")== 0)
      {
	*DataWidth = 0;
      }
      else if (strcmp(OptionPointer->def_val, "64")== 0)
      {
	*DataWidth = 1;
      }
      else
      {
	(void) fprintf(stderr, "Unknown default value.\n");
	exit(-1);
      }
    }
    else if ((OptionPointer = FindByName(SampleQuadDataTable, TheOption))
	     && (strcmp(OptionPointer->def_val,"") != 0))
    {
      if (strcmp(OptionPointer->def_val, "32")== 0)
      {
	*DataWidth = 0;
      }
      else if (strcmp(OptionPointer->def_val, "64")== 0)
      {
	*DataWidth = 1;
      }
      else
      {
	(void) fprintf(stderr, "Unknown default value.\n");
	exit(-1);
      }
    }
    else
    {
      (void) fprintf(stderr, "Unknown data to be collected.\n");
      exit(-1);
    }
  }
  else
  {
    if((strcmp(TempCharacterPointer+1, "32")==0)||(strcmp(TempCharacterPointer+1, "low")==0))
    {
      *DataWidth = 0;
      *TempCharacterPointer = 0;
    }
    else if((strcmp(TempCharacterPointer+1, "64")==0)||(strcmp(TempCharacterPointer+1, "high")==0))
    {
      *DataWidth = 1;
      *TempCharacterPointer = 0;
    }
    else
    {
      (void) fprintf(stderr, "Unknown option entered.\n");
      exit(0);
    }
  }

  return TheOption;
}




