/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program:   iprobe_counting.c
 *
 * Abstract:   Module containing all the code to create a histogram consisting
 *             simply of the number of times the counter for each specified event
 *             interrupted the system, giving the ability to count the number of times
 *             specific events occurred.
 *
 * Modification History:
 *
 * Phillip Ezolt: 11-Jul-97
 *    Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Aug-1994
 *    Defined the macros PRINT_FORMAT1,PRINT_FORMAT2 to take care of
 *    differences in printf format on vms and osf1.
 *
 * Ravindra Kulkarni: Jun-1994
 *    Fixed the integer overflow bug which occurred while printing the total
 *    number of events. Removed the test code.
 *    Removed the events/sec column for total number of events.
 *
 * Greg Tarsa: Apr-1994
 *     Added routine to print 1 line of count info per cpu
 *    Changed code to print event name in its entirety.
 *    Removed date from the per-line output.
 *    Implemented ability to specify "-per_cpu" or "-per_event" on
 *    the command line.
 *    Added legend to the Total and Overflow output.
 *    Added column headings to the report.
 *    Removed CPU from repeating group for -per_cpu
 *
 * Greg Tarsa: Mar-1994
 *    Fixed bug in totaling code: thanks Brian!
 *
 *    Changed to properly handle end_time being a delta time and not
 *    an absolute time.
 *
 *    Fixed code to properly call counter API and display values.
 *    Values are now in "events per second" rather than "interrupts
 *    per interval."
 *
 * Greg Tarsa: Jan-1994
 *    Wrote original code based on template supplied by Brian Wibecan.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#endif
#include "stdtypes.h"
#include "iprobe_struct.h"
#include "iprobe_eventlib.h"
#include "iprobe_hdr.h"
#include "iprobe_global.h"

#ifdef vms
#include "iprobe_vms.h"
#endif

static void PrintCPUStatusInfo(time_t time_stamp,           /* time stamp in UNIX fmt */
			       time_t interval,             /* specified interval */
			       IPR_CTRMASK active_ctrs,     /* active counters */
			       uint32 active_cpumask,       /* active CPUs */
			       IPR_HISTELEM **histo_bycpu,  /* ptr to per-cpu matrix */
			       IPR_HISTELEM **histo_temp,   /* ptr to per-cpu matrix */
			       IPR_CTRNUMBER *events,       /* map of counters to event */
			       IPR_FREQ *freqs);            /* map of counters to freq */

/************************************************************************************/
/* Name: DoCounting                                                                 */
/*                                                                                  */
/* Purpose: Setup the buffers and counters for a counting run.                      */
/*                                                                                  */
/* Arguments: SelectedEvents: An array of events to measure.                        */
/*            SelectedFrequencies: The frequency to measure the events at.          */
/*            SelectionSize: The Size of the prev 2 arrays.                         */
/*            ModesToUse: The program modes to measure in.                          */
/*            StartTime: The number of seconds till start of measurement.           */
/*            Duration: The number of seconds to measure.                           */
/*            OutputInterval: The output interval.                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: If successful this will return a 1 if it is successful, a error code      */
/*        otherwise.                                                                */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS DoCounting(IPR_EVENT *SelectedEvents,           /* array of evts to measure */
		      IPR_FREQ *SelectedFrequencies,           /* freqs to measure at */
		      IPR_COUNT SelectionSize,          /* size of the prev 2 arrays */
		      IPR_MMASK ModesToUse,                /* prog modes to measure in */
		      time_t StartTime,                  /* # secs till start of meas */
		      time_t Duration,                    /* # secs to measure */
		      time_t OutputInterval)                   /* output interval */
{
  /*
     These three symbols are declared volatile because their contents
     are changed asynchronously with respect to the code flow and we don't
     want the compiler to optimize code using them away because it thinks
     the values do not change.
     */

  FLAG BufferFullFlag = 0;     /* flag for buffer full condition */
  FLAG CntrlCFlag = 0;    /* flag for CTRL-C hit by user */
  FLAG TimeOutFlag = 0;   /* flag for end time expiration */
  FLAG *StartedProcessExited;

  int32 Action;
  int32 CPUIndex;

  int CPUContext;/*This needs to be an int for VMS. */


  IPR_STATUS Status;                  /* general status return */
  IPR_CTRMASK CounterMask;                /* mask for which counters are in use*/
  IPR_CTRMASK RequestedCounterMask;             /* requested counter mask */

  /* Event to counter/freq mapping symbols */

  IPR_CTRNUMBER *WhichEvent;
  IPR_FREQ *WhichFrequency;

  IPR_HISTELEM *HistogramBuffers;      /* vector of histogram bufs */
  IPR_HISTELEM *HistogramArea[1];         /* histogram vector loc */
  IPR_HISTELEM **HistogramByCPUData;
  IPR_HISTELEM **HistogramTempData;
  IPR_HISTELEM **HistogramByCPU;
  IPR_HISTELEM **HistogramTotalByCPUData;
  IPR_HISTELEM **HistogramTotalByCPU;

  IPR_SIZE HistogramSize[1];         /* histogram vector size */
  IPR_HMASK HistogramType[1];         /* types of histograms */
  IPR_HMASK  TypeOfHistogram;          /* kind of histos to collect */


  system_info *SystemInformation;


  time_t CurrentTime;
  time_t AbsoluteEndTime;

  uint32 ActiveCPUMask = 0;   /* mask of active CPUs */
  uint32 CurrentCounter;
  uint32 CurrentCPU;
  uint32 CurrentEvent;
  uint32 *TotalNumberOfCounters;     
  uint32 *TotalNumberOfCPUs;     


  uint64 MuxValue;

  void(*PrintReport)();             /* reporting function */

  TotalNumberOfCPUs = GetTotalNumberOfCPUs();

  TotalNumberOfCounters = GetTotalNumberOfCounters();


  SystemInformation = GetSystemInformation();
  StartedProcessExited = GetStartedProcessExited();


  HistogramByCPU = (IPR_HISTELEM **) ckalloc(sizeof(IPR_HISTELEM *) * (*TotalNumberOfCPUs));

  HistogramTotalByCPU =(IPR_HISTELEM **) ckalloc(sizeof(IPR_HISTELEM *) * (*TotalNumberOfCPUs));

  HistogramByCPUData =(IPR_HISTELEM **) ckalloc(sizeof(IPR_HISTELEM *) * (*TotalNumberOfCPUs));

  for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
  {
    HistogramByCPUData[CurrentCPU] =(IPR_HISTELEM *) ckalloc(sizeof(IPR_HISTELEM) * (*TotalNumberOfCounters));
  }

  HistogramTempData =(IPR_HISTELEM **) ckalloc(sizeof(IPR_HISTELEM *) * (*TotalNumberOfCPUs));

  for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
  {
    HistogramTempData[CurrentCPU] =(IPR_HISTELEM *) ckalloc(sizeof(IPR_HISTELEM) * (*TotalNumberOfCounters));
  }

  for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
  {
    for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
    {
      HistogramTempData[CurrentCPU][CurrentCounter] = 0;
    }
  }

  HistogramTotalByCPUData =(IPR_HISTELEM **) ckalloc(sizeof(IPR_HISTELEM *) * (*TotalNumberOfCPUs));

  for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
  {
    HistogramTotalByCPUData[CurrentCPU] =(IPR_HISTELEM *) ckalloc(sizeof(IPR_HISTELEM) * (*TotalNumberOfCounters));
  }

  WhichEvent =(IPR_CTRNUMBER *)ckalloc(sizeof(IPR_CTRNUMBER) * (*TotalNumberOfCounters));
  WhichFrequency =(IPR_FREQ *)ckalloc(sizeof(IPR_FREQ) * (*TotalNumberOfCounters));
  HistogramBuffers =(IPR_HISTELEM *) ckalloc(sizeof(IPR_HISTELEM) * (*TotalNumberOfCounters));


  if((*IsDebug())){
    (void) printf("Number of counters = %d\n", (*TotalNumberOfCounters));
    (void) printf("cpu count = %d\n", (*TotalNumberOfCPUs));
  }

  /* init HistogramByCPU to look like a two dimensional array */

  for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
  {
    HistogramByCPU[CurrentCPU]     = &HistogramByCPUData[CurrentCPU][0];
    HistogramTotalByCPU[CurrentCPU] = &HistogramTotalByCPUData[CurrentCPU][0];
  }

  /* Initialize histogram mask to collect address histograms */
  TypeOfHistogram.ipr_hmask_l_bits = 0;     /* clear all bits */
  TypeOfHistogram.ipr_hmask_v_total = 1;    /* set total bit */

  if((*IsDebug()))
  {
    (void) printf("histogram mask : %x\n", TypeOfHistogram);
  }

  /* Set the reporting type */

  PrintReport = PrintCPUStatusInfo;

  /* Allocate the counters to prevent other processes' access */
  Status = IprAllocate();
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }
  Status = IprCreateHistogramBuffers(IPR_ALL_CPUS,   /* CPU mask */
				     TypeOfHistogram,                  /* kind of hist data */
				     0,      /* no beginning addr */
				     0,      /* no end addr */
				     0,      /* no granularity */
				     0);     /* unknown param */
  if(!IPR_SUCCESS(Status))
  {    
    return Status;
  }

  if((*IsDebug()))
  {
    for(CurrentEvent = 0; CurrentEvent < SelectionSize; CurrentEvent++)
      (void) printf("selected_evt[%d] = %d selected_freq[%d] = %d\n", 
		    CurrentEvent, SelectedEvents[CurrentEvent], CurrentEvent, SelectedFrequencies[CurrentEvent]);
  }

  Status = IprSpecifyOperations(IPR_ALL_CPUS,        /* cpu mask */
				SelectedEvents,      /* event array */
				SelectedFrequencies,     /* freq array */
				SelectionSize,    /* sz of both arrays */
				ModesToUse,      /* mode selection */
				&MuxValue,          /* MUX return */
				&RequestedCounterMask);       /* rtn active ctrs */
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /*
     Get a handle on which counters are in use and the
     frequencies being used.

     We assume that all CPUs are configured identically and
     so look at only one.

     Returned vector has the measured event number,
     if any, in the corresponding element.
     */

  Status = IprEventCounterMap(IPR_CURRENT_CPU,   /* CPU to check */
			      WhichEvent,  /* vector of counters */
			      (*TotalNumberOfCounters));    /* vector length */

  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  Status = IprFreqCounterMap(IPR_CURRENT_CPU,    /* CPU to check */
			     WhichFrequency, /* return vector of freqs */
			     (*TotalNumberOfCounters));    /* vector length */
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* set the active counter(s) */
  /* RequestedCounterMask gets put into sample header */

  Status = IprSetActiveCounters(IPR_ALL_CPUS, RequestedCounterMask, &CounterMask);

  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  if((*IsDebug()))
  {
    (void) printf("Requested counter mask is: 0%o\n", RequestedCounterMask);
    (void) printf("Anticipated counter mask is: 0%o\n", CounterMask);
  }

  /*
     Initialize the various arrays for a IPR_HMASK_M_TOTAL
     operation.
     */
  HistogramArea[0] = &HistogramBuffers[0]; /* histo addr */
  HistogramSize[0] = (*TotalNumberOfCounters);
  HistogramType[0].ipr_hmask_l_bits = IPR_HMASK_M_TOTAL;

  for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
  {
    for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)  /* init total tables */
    {
      HistogramByCPU[CurrentCPU][CurrentCounter] = 0;  /* iterative totals */
      HistogramTotalByCPU[CurrentCPU][CurrentCounter] = 0;  /* cumulative totals */
    }
  }


  /* If specified in cmd line, wait for start time */
  if(StartTime)
  {
    if((*IsDebug()))
    {
      (void) printf("Waiting for start time(%d seconds since program start)\n",
		    StartTime);
    }
    WaitForTime(StartTime);
  }

  if (Duration)
  {
    AbsoluteEndTime = time(0) + Duration;

    if((*IsDebug()))
    {
      (void) printf("Establishing end timer(%d seconds)\n", Duration);
    }
    /* Establish a handler for the timer signal */
    Status = OSDeclareWakeUpHandler(&TimeOutFlag);
    if(!IPR_SUCCESS(Status))
      return Status;

    /* Schedule a WakeUp signal at the ending time */
    Status = OSScheduleWakeUpTime(AbsoluteEndTime);
    if(!IPR_SUCCESS(Status))
    {
      return Status;
    }
  }
  else
  {
    if((*IsDebug()))
    {
      (void) printf("Establishing WakeUp handler\n");
    }
    /* Establish a handler for the timer signal */
    Status = OSDeclareWakeUpHandler(0);
    if(!IPR_SUCCESS(Status))
    {
      return Status;
    }
  }

  /* Establish a signal handler for user-requested abort */
  Status = OSDeclareCntrlCHandler(&CntrlCFlag);
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Establish a signal handler for the PFM interrupt */
  Status = OSDeclarePFMHandler(&BufferFullFlag);
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

	if (!(*IsNoIntermediate()))
	{
		Status = OSScheduleRepeatingWakeUp(OutputInterval);

		if(!IPR_SUCCESS(Status))
		{
			return Status;
		}
	}

  /* Issue the command if one was given */

  Status = IssueCommandLine(SystemInformation->CommandLine);

  if(IPR_SUCCESS(Status))
  {
    Status = IprStart(IPR_ALL_CPUS, CounterMask); /*Start counting on all cpus */
      if(!IPR_SUCCESS(Status))
      {
	return Status;
      }
  }

  /* Wait for signals indicating "buffer full" or "time out" */
  while(IPR_SUCCESS(Status) && !(*StartedProcessExited) && !TimeOutFlag && !CntrlCFlag)
  {

    CurrentTime = time(0);       /* time stamp */

    if((*IsDebug()))
    {
      (void) printf("Waiting for signal. . .");
      (void) fflush(stdout);
    }

    (void) OSWaitForSignal();
    if((*IsDebug()))
    {
      if(TimeOutFlag)
      {
	(void) printf("TIMER signal received\n");
      }
      if(CntrlCFlag)
      {
	(void) printf("CTRL-C signal received\n");
      }
      if(BufferFullFlag)
      {
	(void) printf("BUFFER FULL signal received\n");
      }
      if(!TimeOutFlag && !CntrlCFlag && !BufferFullFlag)
      {
	(void) printf("OTHER signal received\n");
      }
      (void) fflush(stdout);
    }

    /*
       When *any* signal occurs, read the histogram buffers and sum them.
       */
    CPUIndex = 0;
    for(CPUContext = -1; CPUContext < IPR_MAX_CPUS;)
    {
      Action = IPR_HACTION_COPY;
      /*
         Get all histograms for all CPUs
	 */
      Status = IprReadMultipleHistograms(
					 IPR_ALL_CPUS,   /* eligible CPU mask */
					 &CPUContext,    /* cpu ctx buffer */
					 HistogramArea, /* ptr to hist bufs */
					 HistogramSize, /* sizes of bufs */
					 0,      /* min value array */
					 HistogramType, /* type array */
					 0,      /* unknown arg */
					 Action, /* collect action */
					 1);         /* array size */
      if(!IPR_SUCCESS(Status))
      {
        return Status;
      }

      if((uint32)CPUContext < IPR_MAX_CPUS)
      {
	ActiveCPUMask |=(1<<CPUContext);

	/* Get the per-cpu value */

	for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
	{
	  /* record this CPU's event counts */
	  HistogramByCPU[CPUIndex][CurrentCounter] = HistogramBuffers[CurrentCounter];

	  if((*IsDebug()))
	    (void) printf("Count : %d\n", HistogramBuffers[CurrentCounter]);

	}
	CPUIndex++;
      }

    }/* end for each CPU */


	if (!(*IsNoIntermediate()))
	{
	    /* display the current values */
		(*PrintReport)(CurrentTime,	/* time stamp */
		OutputInterval,					/* sample interval(in secs) */
		RequestedCounterMask,			/* requested counters */
		ActiveCPUMask,						/* active CPUs */
		HistogramByCPU,					/* per-cpu histo matrix */
		HistogramTempData,				/* This contains previous data */
		WhichEvent,							/* ctr-evt mapping vector */
		WhichFrequency);					/* ctr-freq mapping vector */
	}

    /*  A copy of current buffer is saved   */

    for(CurrentCPU = 0; CurrentCPU < (*TotalNumberOfCPUs); CurrentCPU++)
    {
      for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
      {
	HistogramTempData[CurrentCPU][CurrentCounter] = HistogramByCPU[CurrentCPU][CurrentCounter];
      }
    }

  }/* end while(IPR_SUCCESSful and !timeout and !interrupted) */

  {
    CurrentTime = time(0);

    /* display the final totals */
    (void) printf("Total event count:\n");
    (*PrintReport)(CurrentTime,            /* time stamp */
		   0,          /* do not scale to "per sec" */
		   RequestedCounterMask,     /* requested counters */
		   ActiveCPUMask,     /* active CPUs */
		   HistogramByCPU,    /* per-cpu histo matrix */
		   HistogramTempData,    /* per-cpu histo matrix */
		   WhichEvent,      /* ctr-evt mapping vector */
		   WhichFrequency);        /* ctr-freq mapping vector */
    
  }

  /* Clear the repeating timer.  */
  Status = OSClearRepeatingWakeUp();
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Restore all signal handling to previous state */
  OSCancelPFMHandler();
  (void)OSCancelCntrlCHandler();
  (void)OSCancelWakeUpHandler();

  /* Stop the counters */
  Status = IprStop(IPR_ALL_CPUS, 0);
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Delete the buffers, now that we are done with them */
  Status = IprDeleteBuffers(IPR_ALL_CPUS);
  if(!IPR_SUCCESS(Status))
  {
    printf("failed\n");
    return Status;
  }

  /* Releae the counters for other folks */
  Status = IprDeallocate();
  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  return 1;
}/* end DoCounting() */


/************************************************************************************/
/* Name: PrintCPUStatusInfo                                                         */
/*                                                                                  */
/* Purpose: Print the specified time stamp, event name, event frequency and         */
/*          histogram value.                                                        */
/*                                                                                  */
/* Arguments: TimeStamp: time stamp in UNIX format.                                 */
/*            Interval: Specified Interval.                                         */
/*            ActiveCounters: A mask of the active counters.                        */
/*            ActiveCPUMask: A mask of the active CPUs.                             */
/*            HistogramByCPU: A pointer to per-cpu matrix.                          */
/*            TempHistogram:  A pointer to per-cpu matrix.                          */
/*            Events:  An array that maps counters to events.                       */
/*            FrequencyArray: An array that maps counters to frequencies.           */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes:  Only counters with bits set in the active counters mask will be          */
/*         printed, since other counters were not requested, the possible data      */
/*         returned, if any, is undefined.                                          */
/*                                                                                  */
/************************************************************************************/

static void PrintCPUStatusInfo(time_t TimeStamp,   /* time stamp in UNIX fmt */
				      time_t Interval,     /* specified Interval */
				      IPR_CTRMASK ActiveCounters,   /* active counters */
				      uint32 ActiveCPUMask,/* active CPUs */
				      IPR_HISTELEM **HistogramByCPU,    /* ptr to per-cpu matrix */
				      IPR_HISTELEM **TempHistogram,    /* ptr to per-cpu matrix */
				      IPR_CTRNUMBER *Events,     /* map of counters to event */
				      IPR_FREQ *FrequencyArray)       /* map of counters to freq */
{
  const char *EventName;

  char Heading[BUFSIZ];
  char Outline[BUFSIZ];

  int32 CPUIndex;
  int32 FrequencyName;

  static int32 HeaderFlag = 0;
  struct tm *CurrentTime;

  uint32 CPUNumber;
  uint32 CurrentCounter;
  uint32 TempTotalNumberOfCPUs;
  uint32 *TotalNumberOfCounters;     
  uint32 *TotalNumberOfCPUs;

  uint64 Average;
  uint64 Sum;

#ifdef nt
  static SYSTEMTIME OldTime;
  SYSTEMTIME NewTime;
  static int32 FirstFlag = 1;
  int32 TimeDifference;
#endif

  TotalNumberOfCounters = GetTotalNumberOfCounters();
  TotalNumberOfCPUs = GetTotalNumberOfCPUs();




#ifdef nt
  if(FirstFlag)
  {
    GetLocalTime(&OldTime);
  }
  else
  {
    GetLocalTime(&NewTime);
  }
  if(FirstFlag)
  {
    TimeDifference = 0;
  }
  else
  {
    if((TimeDifference = GetTimeDifference(OldTime, NewTime)) == 0)
    {
      TimeDifference = 1; /*there must be a minimum difference */
    }
  }
#else
  /* convert time stamp into tm structure */
  CurrentTime = localtime(&TimeStamp);
#endif
  CPUIndex = 0;
  for(CPUNumber = TempTotalNumberOfCPUs = 0; CPUNumber < IPR_MAX_CPUS && TempTotalNumberOfCPUs < (*TotalNumberOfCPUs); CPUNumber++)
  {
    /* skip non-active CPUs */
    if(!((1<<CPUNumber) & ActiveCPUMask))
      continue;
    if((*IsDebug()))
      (void) printf("Current cpu = %d\n", CPUNumber);

    /* print the time stamp to standard output. */
    if(HeaderFlag == 0)
    {
      (void) sprintf(Heading, "time        cpu");
    }

#ifdef nt
    if(FirstFlag)
      (void) sprintf(Outline, "%02d:%02d:%02d.%03d %2d",
		     OldTime.wHour, OldTime.wMinute, OldTime.wSecond,
		     OldTime.wMilliseconds, CPUNumber);
    else
      (void) sprintf(Outline, "%02d:%02d:%02d.%03d %2d",
		     NewTime.wHour, NewTime.wMinute, NewTime.wSecond,
		     NewTime.wMilliseconds, CPUNumber);

#else
    (void) sprintf(Outline, "%02d:%02d:%02d %2d",
		   CurrentTime->tm_hour, CurrentTime->tm_min, CurrentTime->tm_sec, CPUNumber);
#endif

    /* Now print the event information for each event */

    for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
    {
      /* skip counters we aren't measuring. */
      if((ActiveCounters &(1<<CurrentCounter)) == 0)
	continue;

      /* lookup the definitive event name in the main table */

      EventName = (char *) IprEventNumberToName(Events[CurrentCounter]);

      /* Get frequency and convert to "power of 2" */
      for(FrequencyName = 0; FrequencyName < 32 &&(1<<FrequencyName) !=(int)FrequencyArray[CurrentCounter]; 
	  FrequencyName++)
      {
	;
      }

      /* If we saw more than one bit set, assume it to be a magic number(FREQUENCY_LOWEST or FREQUENCY_HIGHEST), use -1 */

      if(FrequencyName >= 32)
      {
	FrequencyName = -1;
      }
      /* print the information to standard output. */
      if(HeaderFlag == 0)
      {
	(void) sprintf(&Heading[strlen(Heading)], " freq %*s",-(int)strlen(EventName), "event");
      }

      (void) sprintf(&Outline[strlen(Outline)], " 2^%-2d %s", FrequencyName, EventName);

      /*
	 Print a column with the total count of Events this Interval
	 and one with the events/second
	 */
      if (HeaderFlag == 0)

      {
	(void) sprintf(&Heading[strlen(Heading)]," %12s %9s", "# events", "evts/sec");
      }

      {


	if((*IsDebug()))
	{
	  (void) printf("CPUindex = %d\n", CPUIndex);
	}

	Sum = (HistogramByCPU[CPUIndex][CurrentCounter]-TempHistogram[CPUIndex][CurrentCounter]) 
	  * FrequencyArray[CurrentCounter];

	if(Interval != 0)

#ifdef nt
	  if(FirstFlag == 0)
	  {
	    Average =((HistogramByCPU[CPUIndex][CurrentCounter]-TempHistogram[CPUIndex][CurrentCounter]) 
		      *(uint64) FrequencyArray[CurrentCounter])* 1000/(uint64) TimeDifference;
	  }
	  else
	    Average =((HistogramByCPU[CPUIndex][CurrentCounter]-TempHistogram[CPUIndex][CurrentCounter]) 
		      *(uint64) FrequencyArray[CurrentCounter])/(uint64) 1;
#else
	Average =((HistogramByCPU[CPUIndex][CurrentCounter]-TempHistogram[CPUIndex][CurrentCounter])
		  *(uint64) FrequencyArray[CurrentCounter])/(uint64) Interval;
#endif
	else
	  Average =(HistogramByCPU[CPUIndex][CurrentCounter]) * (uint64) FrequencyArray[CurrentCounter] ;

	if(Interval != 0)
	{
	  (void) sprintf(&Outline[strlen(Outline)], PRINT_FORMAT1,
			 Sum, Average);
	}
	else
	{
	  (void) sprintf(&Outline[strlen(Outline)], PRINT_FORMAT2, Average);
	}
      }
    } /* end "for every counter. . ." */

    /* print the Heading, if specified */
    if (HeaderFlag == 0)
    {
      (void) printf("%s\n", Heading);
      HeaderFlag++;
    }

    (void) printf("%s\n", Outline);
    CPUIndex++;
    TempTotalNumberOfCPUs++;
  } /* end " for each CPU " */

#ifdef nt
  if(FirstFlag == 0)
  {
    OldTime = NewTime;
  }
  FirstFlag = 0;
#endif
  return;
}/* end PrintCPUStatusInfo */





