/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program: iprobe_eventlib.c
 *
 * Abstract: Module containing all event table lookup and manipulation routines.
 *     
 * Author: Greg Tarsa
 *
 * Creation Date: Dec-93
 *
 * Modification History:
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Michael Dempsey: Aug-94
 *    Modified routine ckalloc to check input argument for zero.
 *    If argument is zero, ckalloc will return will NULL instead
 *    of error
 *
 * Ravindra Kulkarni: Jun-94
 *    Fixed the frequency bug in DisplayColumnHelp. It was displaying
 *    larger number for high frequency and a smaller number for low frequency.
 *    However actually larger number represents low frequency and smaller
 *    number represents high frequency.
 *
 * Ravindra Kulkarni: Jun-94
 *    Developed the module FGetCommandOptions to read the command line arguments
 *    from a text file.
 *
 * Greg Tarsa: Apr-94
 *    Fixed bug where return statuses were not being checked for success.
 *    Replaced some static debugging statements with statements that are
 *    now under the -debug flag.
 *
 * Ravindra Kulkarni: Mar-94
 *    Added no_of_bits_set, generate_combinations,
 *    generate_columns, display_columns, display_help, get_counter_number,
 *    display_frequency, display_text_help, display_system_header
 *
 * Greg Tarsa: Mar-94
 *    Combined Ravindra's work with this module.
 *
 * Ravindra Kulkarni: Feb-94
 *    Added initiate_etable and validate_event.
 *
 * Greg Tarsa: Dec-93
 *    Wrote original code.
*/
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include <string.h>
#ifndef vms
#include <memory.h>
#endif
#include <math.h>
#ifdef vms
#include <ctype.h>
#endif
#include "stdtypes.h"
#include "iprobe_struct.h"
#include "iprobe_eventlib.h"
#include "ipr_etable.def"
#include "iprobe_hdr.h"
#include "iprobe_global.h"

#define MASK_WIDTH 32
#ifdef DEBUG
int32 memory_freed = 0;
int32 memory_used = 0;
#endif





/************************************************************************************/
/* Name: FindEventNumberByName                                                      */
/*                                                                                  */
/* Purpose: Given an EventName, this will return an event number.                   */
/*                                                                                  */
/* Arguments: EventName: The name of the event that we are to find a number for.    */
/*            PrintMessage: Whether or not we are to print out status messages.     */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: This will return the event number.                               */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

IPR_EVENT FindEventNumberByName (char *EventName, IPR_BOOLEAN PrintMessage)
{

  const char *AmbiguousList;
  char *FoundList = NULL;

  int32 NameLength;
  int32 NumberOfMatches = 1;

  IPR_EVENT FoundEvent;
  IPR_STATUS Status;

  struct eventdesc *EventDescription;

  Status = IprEventNameToNumber(EventName, &FoundEvent, &AmbiguousList);
  if (Status == IPR__AMBIGUOUS)
  {
    NumberOfMatches = 2;
    FoundEvent = 0;
  }
  else
  {
    if (!IPR_SUCCESS (Status))
    {
      FoundEvent = 0;
      NumberOfMatches = 0;
      NameLength = strlen (EventName);
      for (EventDescription = event_table; EventDescription->name; EventDescription++)
      {
	if (strcmp (EventName, EventDescription->name) == 0)
	{
	  FoundEvent = EventDescription->num;
	  NumberOfMatches = 1;
	  break;
	}

	if (strncmp (EventName, EventDescription->name, NameLength) == 0)
	{
	  NumberOfMatches++;
	  printf("multiple match\n");
	  if (FoundList)
	  {
	    FoundList = realloc (FoundList,
				 strlen (EventDescription->name) + strlen (FoundList) + 2);
	    (void) strcat (FoundList, " ");
	  }
	  else
	  {
	    FoundList = malloc (strlen (EventDescription->name) + 1);
	    FoundList [0] = '\0';
	  }

	  (void) strcat (FoundList, EventDescription->name);
	  AmbiguousList = FoundList;
	  FoundEvent = EventDescription->num;
	}
      }
    }
  }

  if (NumberOfMatches > 1)
  {
    if (PrintMessage)
    {
      (void) fprintf (stdout, "?ERROR: ambiguous event name: %s\n"
		      "\tCould be one of: %s\n", EventName, AmbiguousList);
      FoundEvent = 0;
    }
    else
    {
      if (NumberOfMatches == 0 && PrintMessage)
      {
	(void) fprintf (stdout, "?ERROR: unknown event name %s\n", EventName);
      }
    }
  }

  if (FoundList)
  {
    free(FoundList);
  }

  return FoundEvent;
}

#define COLUMN_GAP 4
#define PAGE_WIDTH 80

/************************************************************************************/
/* Name: DisplayColumnHelp                                                          */
/*                                                                                  */
/* Purpose: This is used to dynamically print out the correct events depending on   */
/*          what version of chip is present.                                        */
/*                                                                                  */
/* Arguments: None.                                                                 */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: Always returns 0.                                                         */
/*                                                                                  */
/************************************************************************************/

int32 DisplayColumnHelp(void)
{
  char PrintString[80];
  char TempString[80];

  int32 ActualGap;
  int32 ASize;
  int32 Blank;
  int32 CounterMask;
  int32 Columns;
  int32 CurrentArrayElement;
  int32 CurrentCounterNumber;
  int32 CurrentColumnFillerNumber;
  int32 CurrentColumnNumber;
  int32 CurrentEventNumber; 
  int32 DisplayIndex;
  int32 EventNumber;
  int32 EventPointerCoefficient;
  int32 *EventPointer;
  int32 *FrequencyPointer;
  int32 Length;
  int32 LineCounter;
  int32 MaximumLength;
  int32 NumberOfBlocks;
  int32 NumberOfColumns;
  int32 NumberOfGroups;
  int32 OrIndex;
  int32 TotalNumberOfEvents;

  IPR_CTRNUMBER Counter;
  IPR_EVENT Event;
  IPR_CTRMASK TemporaryMask = 0;
  IPR_FREQ FrequencyMask = 0;

  uint32 Index;
  uint32 *TotalNumberOfCounters;     

  TotalNumberOfCounters = GetTotalNumberOfCounters();



  /*    initiate_etable();*/

  TotalNumberOfEvents = IprNumberOfEvents();

  /*  Find out the maximum length of the event */

  for (Index = 1, MaximumLength = 0; Index <= TotalNumberOfEvents; Index++)
  {
    Length = strlen((char *)IprEventNumberToName(Index));
    if ( Length > MaximumLength )
    {
      MaximumLength = Length;
    }
  }

  if ((*IsDebug()))
  {
    (void) printf("Max length of an event = %d\n",MaximumLength);
  }
  /*  Now MaximumLength contains the maximum length of event_name */


  /*  Find out the number of columns that can fit in one line */

  /* How to figure out the number of columns: */

  /* NumberOfColumns * MaximumLength + (NumberOfColumns - 1) * COLUMN_GAP = PAGE_WIDTH */
  /* NumberOfColumns * MaximumLength + NumberOfColumns * COLUMN_GAP - COLUMN_GAP */
  /*       = PAGE_WIDTH */
  /* NumberOfColumns (MaximumLength + COLUMN_GAP) = PAGE_WIDTH + COLUMN_GAP */
  /* NumberOfColumns = (PAGE_WIDTH + COLUMN_GAP) / (MaximumLength + COLUMN_GAP) */

  NumberOfColumns = (PAGE_WIDTH + COLUMN_GAP) / (MaximumLength + COLUMN_GAP);

  if ((*IsDebug()))
    (void) printf("Max number of columns possible is = %d\n",NumberOfColumns);

  /*  Now NumberOfColumns contains the maximum number of columns possible */


  /*  Following section of code generates the event table */

  /*BUG: Use BZERO. */
  ASize = (*TotalNumberOfCounters) * TotalNumberOfEvents;
  EventPointer =  (int32 *)malloc(sizeof(int) * ASize);

  for ( CurrentArrayElement = 0; CurrentArrayElement < ASize; CurrentArrayElement++ )
  {
    EventPointer[CurrentArrayElement] = 0;
  }

  ASize = (*TotalNumberOfCounters) * 2;
  FrequencyPointer = (int32 *)malloc(sizeof(int) * ASize);
  for ( CurrentArrayElement = 0; CurrentArrayElement < ASize; CurrentArrayElement++ )
  {
    FrequencyPointer[CurrentArrayElement] = 0;
  }
  
  if (EventPointer == 0 && (*IsDebug()))
  {
    (void) printf("Error in allocating the memory\n");
    exit(-1);
  }

  for (Counter = 0; Counter < (*TotalNumberOfCounters); Counter++)
  {
    CounterMask = 1 << Counter;
    EventPointerCoefficient = 0;

    for (Event = 1; Event <= TotalNumberOfEvents; Event++)
    {
      (void) IprAvailableCounters (Event, &TemporaryMask);
      if (TemporaryMask & CounterMask)
      {
	FrequencyMask = 0;
	(void) PutElement(EventPointer, TotalNumberOfEvents, Counter,
			  EventPointerCoefficient++, Event);
	(void) IprAvailableFreqs (Event, &FrequencyMask);

	/*      Note: The low frequency number is represented by a larger number
		and high frequency number is represented by smaller number.
		It is the frequency of interrupts. */

	/* Note that the lowest value of both the low frequency and the */
	/* high frequency are stored.  This will allow the EV4 case of */
	/* double frequencies to be displayed easily, at the risk of */
	/* messing up some future display. */

	if (GetElement (FrequencyPointer, 2, Counter, 1) > BIprLowest (FrequencyMask)
	    || GetElement (FrequencyPointer, 2, Counter, 0) == 0)
	{
	  (void) PutElement (FrequencyPointer, 2, Counter, 1,
			     BIprLowest (FrequencyMask));
	}
	while (FrequencyMask && !BIprIs1 (FrequencyMask))
	{
	  FrequencyMask = BIprClr (FrequencyMask, BIprLowest (FrequencyMask));
	}
	if (FrequencyMask && (GetElement (FrequencyPointer, 2, Counter, 0) > FrequencyMask
			      || GetElement (FrequencyPointer, 2, Counter, 0) == 0))
	{
	  (void) PutElement (FrequencyPointer, 2, Counter, 0, FrequencyMask);
	}
      }
    }
  }

  (void) PutElement(EventPointer, TotalNumberOfEvents, Counter, EventPointerCoefficient, 0);

  /*  At this stage all the events have been stored in the array */

  /* Print the events in the array, if debugging is on. */
  if ((*IsDebug()))
  {
    for (CurrentCounterNumber = 0; CurrentCounterNumber < (*TotalNumberOfCounters); CurrentCounterNumber++)
    {
      for (CurrentEventNumber = 0; CurrentEventNumber < TotalNumberOfEvents; CurrentEventNumber++)
      {
	if (GetElement(EventPointer, TotalNumberOfEvents, CurrentCounterNumber, CurrentEventNumber) == 0)
	  break;
	(void) printf("%s\n",
		      IprEventNumberToName(GetElement(EventPointer, TotalNumberOfEvents, CurrentCounterNumber, CurrentEventNumber)));
      }
      (void) printf("\n");
    }
  }

  /*  Here the frequencies are printed, if debugging is on */

  if ((*IsDebug()))
  {
    for (CurrentCounterNumber = 0; CurrentCounterNumber < (*TotalNumberOfCounters); CurrentCounterNumber++)
    {
      (void) printf("Counter =  %d Low freq = %d High Freq = %d\n", CurrentCounterNumber,
		    GetElement(FrequencyPointer, 2, CurrentCounterNumber, 0),
		    GetElement(FrequencyPointer, 2, CurrentCounterNumber, 1));
    }
  }

  /*  At this point the data generation is complete. */

  /*  Now the data will be printed on the screen  */

  /*  If the number of columns to be displayed is lesser than the number
      of columns that can be displayed on the screen then the intercolumn
      spacing is recalculated. */

  Columns = (*TotalNumberOfCounters); /* Let the number of columns be number of counters */

  if ((*IsDebug()))
    (void) printf("The number of columns desired is = %d\n", Columns);

  /*  Now columns contains the actual no of columns to be displayed. */

  if (Columns <  NumberOfColumns )
  {
    NumberOfColumns = Columns;
  }

  /*  Block is an empty space between two columns */
  NumberOfBlocks = 1 + NumberOfColumns;
  Blank = (PAGE_WIDTH - NumberOfColumns * MaximumLength) /  NumberOfBlocks;
  ActualGap = Blank;

  if ((*IsDebug()))
  {
    (void) printf("Actual gap is = %d\n", ActualGap);
  }
  /*  ActualGap contains the gap between two columns. */

  /*  Group is an aggregate of columns that can be displayed on the
      screen at a time. If no of columns to be displayed is 6 and at a
      time only 3 can be displayed then we say there are two groups. */

  NumberOfGroups = Columns / NumberOfColumns + Columns % NumberOfColumns;

  if ((*IsDebug()))
  {
    (void) printf("Total number of groups is = %d\n",NumberOfGroups);
  }

  (void) printf("\nEvents defined on the current system -- select up to one event from each column:\n\n");
  DisplayIndex = 0;
  Index = 0;
  while (TRUE)
  {
    LineCounter = 0;
    if (Index > (*TotalNumberOfCounters) )
      break;
    while (TRUE)
    {
      /*  OrIndex is true if there is at least one element to be
	  displayed from the same group. If zero it indicates that the
	  all the events in the group are displayed and now frequencies
	  can be displayed. */
      OrIndex = 0;

      for (DisplayIndex = Index; DisplayIndex < ((int)Index + NumberOfColumns);
	   DisplayIndex++)
      {
	OrIndex |= GetElement(EventPointer, TotalNumberOfEvents, DisplayIndex,
			      LineCounter);
      }

      if (OrIndex == 0) /* event display is over */
      {
	DisplayIndex -= NumberOfColumns; /* Restore the display Index */
	(void) putchar('\n');

	/*  Display heading 'Frequencies' */
	for (CurrentColumnNumber = 0; CurrentColumnNumber < NumberOfColumns; CurrentColumnNumber++)
	{
	  for (CurrentColumnFillerNumber = 0; CurrentColumnFillerNumber < COLUMN_GAP; CurrentColumnFillerNumber++)
	    (void) putchar(' ');

	  (void) printf("%-*s", ActualGap + MaximumLength - COLUMN_GAP,
			"Frequencies");
	}
	(void) putchar('\n');

	/*  Display low frequencies */
	for (CurrentColumnNumber = 0; CurrentColumnNumber < NumberOfColumns; CurrentColumnNumber++)
	{

	  for (CurrentColumnFillerNumber = 0; CurrentColumnFillerNumber < COLUMN_GAP; CurrentColumnFillerNumber++)
	    (void) putchar(' ');

	  /*  Prepare the string for display */
	  (void) strcpy(TempString, "Low :");
	  (void) sprintf(PrintString, "%-d",
			 GetElement(FrequencyPointer, 2, CurrentColumnNumber, 0));
	  (void) strcat(TempString, PrintString);
	  (void) printf("%-*s", ActualGap + MaximumLength - COLUMN_GAP,
			TempString);
	}
	(void) putchar('\n');

	/*  Display low frequencies */
	for (CurrentColumnNumber = 0; CurrentColumnNumber < NumberOfColumns; CurrentColumnNumber++)
	{
	  for (CurrentColumnFillerNumber = 0; CurrentColumnFillerNumber < COLUMN_GAP; CurrentColumnFillerNumber++)
	    (void) putchar(' ');

	  /*  Prepare the string for display */
	  (void) strcpy(TempString, "High:");
	  (void) sprintf(PrintString, "%-d",
			 GetElement(FrequencyPointer, 2, CurrentColumnNumber, 1));
	  (void) strcat(TempString, PrintString);
	  (void) printf("%-*s", ActualGap + MaximumLength - COLUMN_GAP,
			TempString);
	}
	(void) putchar('\n');

	/*  Currently this code is displaying the Type as Only.
	    It should be modified to display as Range as needed. */
//	for (CurrentColumnNumber = 0; CurrentColumnNumber < NumberOfColumns; CurrentColumnNumber++)
//	{
//	  for (CurrentColumnFillerNumber = 0; CurrentColumnFillerNumber < COLUMN_GAP; CurrentColumnFillerNumber++)
//	    (void) putchar(' ');
//	  (void) printf("%-*s", ActualGap + MaximumLength - COLUMN_GAP,
//			"Type: Only");
//	}
	(void) putchar('\n');

	break;  /*  Out of while loop */
      }
      /*  Here the events are displayed */
      for (DisplayIndex = Index; DisplayIndex <
	   ((int)Index + NumberOfColumns); DisplayIndex++)
      {
	EventNumber = GetElement(EventPointer, TotalNumberOfEvents, DisplayIndex,
				 LineCounter);
	if (EventNumber == 0)
	  (void) printf("%*s", ActualGap + MaximumLength, "");
	else
	{
	  FrequencyMask = 0;
	  /*  Check if the frequency is same or not. If it is not same
	      then display * before the event name. */
	  (void) IprAvailableFreqs (EventNumber, &FrequencyMask);
	  FrequencyMask = BIprLowest (FrequencyMask);
	  (void) printf ("%*s%s%-*s", COLUMN_GAP - 1, "",
			 (FrequencyMask == GetElement(FrequencyPointer, 2, DisplayIndex, 1))
			 ? " " : "*",
			 ActualGap + MaximumLength - COLUMN_GAP,
			 IprEventNumberToName(EventNumber));
	}
      }
      LineCounter++;
      (void) printf("\n");
    }
    Index += NumberOfColumns;
    {
      if (Index == (*TotalNumberOfCounters))
      {
	break;
      }
      if (Index + NumberOfColumns > (*TotalNumberOfCounters))
      {
	NumberOfColumns = (*TotalNumberOfCounters) - Index;
	continue;
      }
    }
  }
  (void) printf("\n");
  (void) printf("\tAn asterisk indicates the event is measured at twice the specified frequency.\n\n");
  return 0;
}


/************************************************************************************/
/* Name: GetElement                                                                 */
/*                                                                                  */
/* Purpose: This routine computes the address of an element and stores the          */
/*             element at that address. It handles 2 dimensional arrays.            */
/*                                                                                  */
/* Arguments: Array: The array where we want to get the element.                    */
/*            LeftWidth: the "Width" of the array.                                  */
/*            Left: The row that we want.                                           */
/*            Right: The column that we want.                                       */
/*            Element: The element to be stored.                                    */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: The proper element value.                                        */
/*                                                                                  */
/* Notes: Why not just make a 2D array?                                             */
/*                                                                                  */
/************************************************************************************/

uint32 PutElement(int32 *Array, int32 LeftWidth, int32 Left, int32 Rigth, int32 Element)
{
  Array[LeftWidth * Left + Rigth] = Element;
  return 0;
}

/************************************************************************************/
/* Name: GetElement                                                                 */
/*                                                                                  */
/* Purpose: This routine computes the address of an element and returns the         */
/*             element at that address. It handles 2 dimensional arrays.            */
/*                                                                                  */
/* Arguments: Array: The array where we want to get the element.                    */
/*            LeftWidth: the "Width" of the array.                                  */
/*            Left: The row that we want.                                           */
/*            Right: The column that we want.                                       */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: The proper element value.                                        */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

uint32 GetElement(int32 *Array, int32 LeftWidth, int32 Left, int32 Right)
{
    return Array[LeftWidth * Left + Right];
}

/************************************************************************************/
/* Name: FGetCommandOptions                                                         */
/*                                                                                  */
/* Purpose: This routine reads the options from an input file and returns them      */
/*          in argv and argc format. It assumes the maximum length of an option     */
/*          to be not greater than 80. There is no limit to the number of options.  */
/*                                                                                  */
/* Arguments: FileName: The name of the file to read the options from.              */
/*            argv: A pointer to the destination vector.                            */
/*            argc: A pointer to the destination argument number.                   */
/*                                                                                  */
/* Output: argv: A vector of pointers to arguments.                                 */
/*         argc: The number of arguments in argv.                                   */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

void FGetCommandOptions(char *FileName, char ***argv, int32 *argc)
{
  char CurrentOption[80];        /*  memory to read option from file */
  char *OptionPointer;       /*  It pointes to the current option */
  char **ArgumentList;   /*  This list is created and returned to user */
  FILE *FilePointer;               /*  File pointer for reading cmd file */
  int32 NumberOfOptions;       /*  Counter for the number of options */


  /*  Open the command file for reading. */

  FilePointer = fopen(FileName, "r");
  if (!FilePointer)
  {
    (void) printf("Error in locating %s\n", FileName);
    exit(-1);
  }

  /*  Find the number of options */
  for (NumberOfOptions = 0; fscanf(FilePointer, "%s", CurrentOption) != EOF; NumberOfOptions++)
  {
    ; /*    Empty loop to count number of options   */
  }

  ArgumentList = (char **)malloc(sizeof(char *) * NumberOfOptions);

  (void) rewind(FilePointer);

  /*  Read the command file till the end of file.\n");    */

  for (NumberOfOptions = 0; fscanf(FilePointer, "%s", CurrentOption) != EOF; NumberOfOptions++)
  {
    OptionPointer = (char *)malloc(strlen(CurrentOption)+1);
    if (!OptionPointer)
    {
      (void) printf("Error in allocating memory.\n");
      exit(-1);
    }

    (void) strcpy(OptionPointer, CurrentOption);
    ArgumentList[NumberOfOptions] = OptionPointer;
  }

  *argv =  ArgumentList;     /*   return the argument list    */
  *argc = NumberOfOptions;       /*   return the number of options read */

  (void) fclose(FilePointer);

}

/************************************************************************************/
/* Name: Fatal                                                                      */
/*                                                                                  */
/* Purpose: This prints the given message and exits.                                */
/*                                                                                  */
/* Arguments: Message: The message to be printed.                                   */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

void Fatal(char *Message)
{
  (void) fprintf(stderr, "%s\n", Message);

  exit(-1);
}

/************************************************************************************/
/* Name: ckalloc                                                                    */
/*                                                                                  */
/* Purpose: This is a wrapper for malloc, that adds little benefit.                 */
/*                                                                                  */
/* Arguments: amount: The amount of memory to allocate.                             */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

void *ckalloc(int32 amount)
{
  void *p;

  if (amount == 0)
    return NULL;

  if ( (p=malloc(amount)) == 0)
  {
    Fatal("Ran out of memory.");
  }

#ifdef DEBUG
  memory_used += amount;
  (void) printf( "Total memory allocated = %d\n", memory_used );

#endif

  return p;
}

/************************************************************************************/
/* Name: ConvertStringToUpperCase                                                   */
/*                                                                                  */
/* Purpose: This converts the given string to upper case.                           */
/*                                                                                  */
/* Arguments: String: The string to be converted.                                   */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

void ConvertStringToUpperCase(char *String)
{
  while (*String)
    {
      *String = (char)toupper(*String);
      String++;
    }
}


/************************************************************************************/
/* Name: ConvertStringToLowerCase                                                   */
/*                                                                                  */
/* Purpose: This converts the given string to lower case.                           */
/*                                                                                  */
/* Arguments: String: The string to be converted.                                   */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

void ConvertStringToLowerCase(char *String)
{
  while (*String)
    {
      *String = (char)tolower(*String);
      String++;
    }
}
