/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/

 /*
  *  Program: iprobe_hdr.c
  *
  *  Abstract:  Module containing the functions for transferring the data from system
  *             structure to the sample file, platform independent data to the
  *             system structure and to display the data on the screen.
  *             This module is fully portable over VMS, OSF and NT.
  *    
  *  Author: Ravindra Kulkarni 
  *    
  *  Creation Date:  Apr-94
  *
  */
/*
 *Modification History:
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Sep-94
 *    Concatenated the sample file header data and wrote it to the sample
 *    file in one shot instead of 4 calls to IprWriteSampleBuffer.
 *    Added quad datamask in global structure and in ReadUserOptions.
 *
 * Ravindra Kulkarni: Aug-94
 *    The display of page length and page count included in case of NT.
 *
 * Michael Dempsey: Aug-94
 *    For OSF/1, the header display memory size in MB
 *    instead of memory page count and page size
 *
 * Ravindra Kulkarni: Jun-94
 *    Modified display_user_header, ReadUserOptions to take care of the new
 *    option -collect,-nocollect,-add and the specification of datamask for
 *    sampling by user.
 *
 * Greg Tarsa: May-94
 *    Added code to put mode bits into SystemInformation structure.
 *    Added code to print symbolic mode names in user-friendly header
 *    Added code to print symbolic method names.
 *
 * Greg Tarsa: Apr-94
 *    Enhance messages a bit.
 *
 * Ravindra Kulkarni: Apr-94
 *    Wrote original code.
 *
 */
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <string.h>
#include "stdtypes.h"
#include "iprobe_struct.h"
#include "iprobe_hdr.h"
#include "iprobe_eventlib.h"
#include "iprobe_global.h"

#define HEADER_BUF_SIZE 512

typedef struct header_info
{  
  char Signature[8];
  /*   Hardware specific information */

  uint64 CPUSpeed;       /* cpu speed of current system */
  uint64 TimeStamp;  /* starting time in OS specific format */
  /*   Above vms time is a mismatch in this group. Added for a better 64 bit alignment. */

  uint64 CPUType;        /* Alpha chip type */
  int32 NumberOfCounters;           /* The number of counters present on the system */
  IPR_CTRMASK CounterMask; /* counter mask on current system */
  int32 AvailibleCPUMask;     /* cpu mask for current system */

  /*   OS  related information */
  
  int32 OSType;           /* OS_VMS, OS_NT, OS_OSF1 etc */
  int32 OSVersionMajor;      /* OS major version number */
  int32 OSVersionMinor;      /* OS minor version number */

  /*   Time related information */

  time_t StartTime; /* starting time in unix format */

  /*   Other data */

  int32 DataType;          /* sampling data , histogram data etc */
  IPR_COUNT NumberOfRanges;       /* number of ranges */
  int32 NullProcessID;       /* Null process id for current system */
  IPR_DMASK DataMask;      /* Mask of the data to be collected */
  IPR_DMASK QuadDataMask;  /*    Quad data mask */

} HeaderInfoType;



/************************************************************************************/
/* Name: DisplaySystemHeader                                                        */
/*                                                                                  */
/* Purpose: This writes information about the current hardawre and the software to  */
/*          the given file descriptor.  However, this is meant to be standard out.  */
/*          It gets its information from the SystemInformation struct, which must be         */
/*          initialized ahead of time.                                              */
/*                                                                                  */
/* Arguments: FilePointer: A pointer to the output stream.                          */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/************************************************************************************/

void DisplaySystemHeader(FILE *FilePointer)
{
  system_info *SystemInformation;
  
  SystemInformation = GetSystemInformation();
 


    (void) fprintf(FilePointer, "Node name            : %s\n", SystemInformation->nodename);
    (void) fprintf(FilePointer, "OS                   : %s\n", SystemInformation->osname);
    (void) fprintf(FilePointer, "CPU count            : %d\n", SystemInformation->NumCpus);
    (void) fprintf(FilePointer, "Model                : %s\n", SystemInformation->modelname);

#if defined vms || defined nt

    (void) fprintf(FilePointer, "Page count           : %d\n", SystemInformation->memorysize);
    (void) fprintf(FilePointer, "Pagelength           : %d\n", SystemInformation->pagelength);

#endif

#ifdef unix

    (void) fprintf(FilePointer, "Memory size          : %d MB\n",
        SystemInformation->memorysize / (1024*1024));
#endif

    (void) fprintf(FilePointer, "Counter count        : %d\n", SystemInformation->num_ctrs);
}

/************************************************************************************/
/* Name: ReadUserOptions                                                            */
/*                                                                                  */
/* Purpose: This sets up the SystemInformation structure with all of the proper information  */
/*                                                                                  */
/* Arguments: Events: An array containting the events we are looking at.            */
/*	      Frequencies: An array containing the frequencies for the events to be */
/*                   observed.                                                      */
/*            Skips: An array containing the number of interrupts to skip for all   */
/*		     events.                                                        */
/*	      NumEvents: The number of events we are looking at.                    */
/*	      ModeMask: A mask describing the mode that we are observing.           */
/*		      IPR_DMASK DataMask,*/
/*		      IPR_DMASK QuadDataMask,*/
/*	      BufferCount: Total number of buffers used to hold sampling data.      */
/*	      BufferSize: Size of each individual buffer.                           */
/*            Delay: Amount of Time to wait before sampling.                        */
/*	      Duration: Total amount of time to sample.                             */
/*	      Interval: Number of seconds between printing data to the screen.      */
/*	      PCRange: PC range information.                                        */
/*	      OutputFileName: The name of the output file for sample data.          */
/*	      Method: Measuring Method                                              */
/*	      CommandLine: Command to issue                                         */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: Returns True regardless of what happens.                         */
/*                                                                                  */
/************************************************************************************/

int32 ReadUserOptions(IPR_EVENT *Events,
		      IPR_FREQ *Frequencies,
		      IPR_COUNT *Skips,
		      int32 NumEvents,
		      IPR_MMASK ModeMask,
		      IPR_DMASK DataMask,
		      IPR_DMASK QuadDataMask,
		      IPR_COUNT BufferCount,
		      IPR_COUNT BufferSize,
		      time_t Delay,
		      time_t Duration,
		      IPR_COUNT Interval,
		      struct pc_ranges *PCRange, /*  PC range information    */
		      char *OutputFileName,
		      IPR_COUNT Method,    /* measuring method */
		      char *CommandLine) /* command to issue */
{
  int32 CurrentEventNumber;
  system_info *SystemInformation;
 

  SystemInformation = GetSystemInformation();
 


  
  SystemInformation->pc_range = PCRange->next;

  SystemInformation->method = Method;

  SystemInformation->CommandLine = CommandLine;

  SystemInformation->outfile = (char *)ckalloc(strlen(OutputFileName)+1);

  (void) strcpy(SystemInformation->outfile, OutputFileName);

  SystemInformation->interval = Interval;

  SystemInformation->cur_time = time(0);

  SystemInformation->duration = Duration;

  SystemInformation->delay = Delay;

  SystemInformation->buffer_count = BufferCount;

  SystemInformation->buffer_size = BufferSize;

  SystemInformation->NumEvents = NumEvents;

  SystemInformation->sel_freqs = (IPR_FREQ *)ckalloc(sizeof (IPR_FREQ) * NumEvents);

  SystemInformation->sel_events = (IPR_EVENT *)ckalloc(sizeof (IPR_EVENT) * NumEvents);

  SystemInformation->skips = (IPR_COUNT *)ckalloc(sizeof (IPR_COUNT) * NumEvents);

  SystemInformation->modemask = ModeMask;

  SystemInformation->datamask = DataMask;

  SystemInformation->quaddatamask = QuadDataMask;

  for (CurrentEventNumber = 0; CurrentEventNumber < NumEvents; CurrentEventNumber++)
  {
    SystemInformation->sel_freqs[CurrentEventNumber] = Frequencies[CurrentEventNumber];
    SystemInformation->sel_events[CurrentEventNumber] = Events[CurrentEventNumber];
    SystemInformation->skips[CurrentEventNumber] = Skips[CurrentEventNumber];
  }
  return TRUE;

}


/************************************************************************************/
/* Name: DisplayUserOptions                                                         */
/*                                                                                  */
/* Purpose: This prints out the all of the options for iprobe.                      */
/*                                                                                  */
/* Arguments: FilePointer: The stream where all of the data will be written to.     */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: Always returns 0.                                                */
/*                                                                                  */
/************************************************************************************/

int32 DisplayUserOptions(FILE *FilePointer)
{
  char MethodBuffer[30];

  int32 CurrentEventNumber;
  int32 CurrentRangeNumber;
  int32 CurrentSampleEntry;
  int32 CurrentShiftAmount;
  int32 SamplePresent = 0;
  
  struct keywords *SampleDataTable;
  struct keywords *SampleQuadDataTable;

  system_info *SystemInformation;
  
  struct keywords *UnknownKeyword;
  struct pc_ranges *TempPCRange;
  
  UnknownKeyword = GetUnknownKeyword();
  
  SampleDataTable = OSGetSampleDataTable();
  SampleQuadDataTable = OSGetSampleQuadDataTable();

  SystemInformation = GetSystemInformation();

  /* print each event and its specication */

  for(CurrentEventNumber = 0; CurrentEventNumber < SystemInformation->NumEvents; CurrentEventNumber++)
  {
    (void) fprintf(FilePointer, "%-20s :",
		   IprEventNumberToName(SystemInformation->sel_events[CurrentEventNumber]));

    /* print frequency (and name, if applicable) */

    if (SystemInformation->sel_freqs[CurrentEventNumber] == IPR_FREQUENCY_LOWEST)
    {
      (void) fprintf(FilePointer, "%s", " Low frequency");
    }
    else if (SystemInformation->sel_freqs[CurrentEventNumber] == IPR_FREQUENCY_HIGHEST)
    {
      (void) fprintf(FilePointer, "%s", " High frequency");
    }
    else
    {
      (void) fprintf(FilePointer, " %d events/sec", SystemInformation->sel_freqs[CurrentEventNumber]);
    }

    for (CurrentShiftAmount = 1; FindMethodByNumber(1<<CurrentShiftAmount) != UnknownKeyword; CurrentShiftAmount++)
    {
      if (SystemInformation->method & (1<<CurrentShiftAmount))
      {
	if(strcmp(FindMethodByNumber(1<<CurrentShiftAmount)->name,"sample")==0)
	{

	  (void) fprintf(FilePointer," skip %d interrupts between samples",
			 SystemInformation->skips[CurrentEventNumber]);
	}
      }
    }
    (void) fprintf(FilePointer, "\n");
  }

  (void) fprintf(FilePointer, "Current time         : %s", ctime(&SystemInformation->cur_time));

  (void) fprintf(FilePointer, "Start time:          : %s",
		 (SystemInformation->delay) ? ctime(&SystemInformation->delay) : "immediate\n");

  (void) fprintf(FilePointer, "Duration             : %d%s\n", SystemInformation->duration,
		 (SystemInformation->duration) ? "" : " (until user interrupts)");

  (void) fprintf(FilePointer, "Interval             : %d \n", SystemInformation->interval);

  (void) fprintf(FilePointer,"Method               : ");

  /* There may be multiple methods specified, print them all */

  for (CurrentShiftAmount = 1; FindMethodByNumber(1<<CurrentShiftAmount) != UnknownKeyword; CurrentShiftAmount++)
  {
    if (SystemInformation->method & (1<<CurrentShiftAmount))
    {
      (void) fprintf(FilePointer, "%s ", FindMethodByNumber(1<<CurrentShiftAmount)->name);
      (void) sprintf(MethodBuffer, "%s", FindMethodByNumber(1<<CurrentShiftAmount)->name);
      if(strcmp(MethodBuffer, "sample") == 0)
	SamplePresent = 1;
    }
  }

  (void) fprintf(FilePointer, "\n");

  TempPCRange = SystemInformation->pc_range;

  CurrentRangeNumber = 1;

  while (TempPCRange)
  {
    (void) fprintf(FilePointer,
		   "Address Range %d      : 0X"ADDRFORMAT".0X"ADDRFORMAT".%d\n",
		   CurrentRangeNumber, TempPCRange->start, TempPCRange->end,
		   TempPCRange->granularity);

    TempPCRange = TempPCRange->next;
      
    CurrentRangeNumber++;
  }

  if ((*IsDebug()))
  {
    (void) fprintf(FilePointer, "Mode Bits            : 0x%x\n",
		   SystemInformation->modemask.ipr_mmask_l_bits);
  }

  (void) fprintf(FilePointer, "Measured Modes       : ");

  /* There may be multiple modes specified, print them all */

  if (SystemInformation->modemask.ipr_mmask_l_bits == 0)

    (void) fprintf(FilePointer, "all modes");

  else

    for (CurrentShiftAmount = 1; FindModeByNumber(1<<CurrentShiftAmount) != UnknownKeyword; CurrentShiftAmount++)
    {
      if (SystemInformation->modemask.ipr_mmask_l_bits & (1<< CurrentShiftAmount))
      {
	(void) fprintf(FilePointer, "%s ", FindModeByNumber(1<<CurrentShiftAmount)->name);
      }
    }

  (void) fprintf(FilePointer, "\n");

  (void) fprintf(FilePointer,"Measured Data        : ");

  if ((SystemInformation->datamask.ipr_dmask_l_bits == 0x00)&&(SystemInformation->quaddatamask.ipr_dmask_l_bits == 0x00))
  {
    (void) fprintf(FilePointer, "no data");
  }
  else
  {
    for (CurrentSampleEntry = 0; SampleDataTable[CurrentSampleEntry].num; CurrentSampleEntry++)
    {
      if (SampleDataTable[CurrentSampleEntry].num &
	  SystemInformation->datamask.ipr_dmask_l_bits)

	(void) fprintf(FilePointer, "%s ", SampleDataTable[CurrentSampleEntry].name);
    }

    for (CurrentSampleEntry = 0; SampleQuadDataTable[CurrentSampleEntry].num; CurrentSampleEntry++)
    {
      if (SampleQuadDataTable[CurrentSampleEntry].num &
	  SystemInformation->quaddatamask.ipr_dmask_l_bits)
      {

	(void) fprintf(FilePointer, "%s ", SampleQuadDataTable[CurrentSampleEntry].name);
      }
    }


  }




  (void) fprintf(FilePointer, "\n");
  if (SamplePresent)
  {
    (void) fprintf(FilePointer, "Output file          : %s \n", SystemInformation->outfile);
  }
  (void) fprintf(FilePointer, "BufferCount          : %d \n", SystemInformation->buffer_count);
  (void) fprintf(FilePointer, "Buffer_size          : %d \n", SystemInformation->buffer_size);

  return 0;
}



/************************************************************************************/
/* Name: WriteHeaderInfo                                                            */
/*                                                                                  */
/* Purpose: This takes all of the relavent header data, and writes it out to the    */
/*          sampling file. This includes: System Type, OS type, and version.        */
/*          It assumes that the structure SystemInformation is initialized.         */
/*                                                                                  */
/* Arguments: None                                                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: If there is an error, this returns the error number, if not, it  */
/*                 returns 0.                                                       */
/*                                                                                  */
/************************************************************************************/


int32 WriteHeaderInfo(void)
{
  char *HeaderBuffer;
  HeaderInfoType HeaderInfo;
  int32 BufferLength = 0;
  int32 Status;

  system_info *SystemInformation;

  SystemInformation = GetSystemInformation();
 


  HeaderInfo.DataType = SystemInformation->method;
  HeaderInfo.NumberOfRanges = SystemInformation->rangecnt;
  HeaderInfo.NullProcessID = SystemInformation->null_proc_id;
  HeaderInfo.NumberOfCounters = SystemInformation->num_ctrs;
  HeaderInfo.CPUSpeed = SystemInformation->cpu_speed;
  HeaderInfo.CounterMask = SystemInformation->counter_mask;
  HeaderInfo.AvailibleCPUMask = SystemInformation->avail_cpu_mask;
  HeaderInfo.CPUType = SystemInformation->cpu_type;
  HeaderInfo.OSType = SystemInformation->OS_type;
  HeaderInfo.OSVersionMajor = SystemInformation->OS_ver_major;
  HeaderInfo.OSVersionMinor = SystemInformation->OS_ver_minor;
  HeaderInfo.TimeStamp = SystemInformation->timestamp;
  HeaderInfo.StartTime = SystemInformation->start_time;
  HeaderInfo.DataMask = SystemInformation->datamask;
  HeaderInfo.QuadDataMask = SystemInformation->quaddatamask;
  (void) strcpy(HeaderInfo.Signature, "IPROBE");

  HeaderBuffer = (char *)ckalloc(HEADER_BUF_SIZE);

  (void) memcpy(HeaderBuffer, (char *)&HeaderInfo, sizeof(HeaderInfoType));

  BufferLength = sizeof(HeaderInfoType);

  (void) memcpy(&HeaderBuffer[BufferLength], (char *)SystemInformation->act_events, sizeof(IPR_EVENT) * SystemInformation->num_ctrs);

  BufferLength += sizeof(IPR_EVENT) * SystemInformation->num_ctrs;

  (void) memcpy(&HeaderBuffer[BufferLength], (char *)SystemInformation->act_freqs, sizeof(IPR_FREQ) * SystemInformation->num_ctrs);

  BufferLength += sizeof(IPR_FREQ) * SystemInformation->num_ctrs;

  (void) memcpy(&HeaderBuffer[BufferLength], (char *)SystemInformation->act_skips, sizeof(int32) * SystemInformation->num_ctrs);

  BufferLength += sizeof(int32) * SystemInformation->num_ctrs;

  if(SystemInformation->rangecnt)
    {
      (void) memcpy(&HeaderBuffer[BufferLength], (char *)SystemInformation->rangeptr,
		     sizeof(uint64) * 2 * SystemInformation->rangecnt * SystemInformation->num_ctrs);

      BufferLength += sizeof(uint64) * 2 * SystemInformation->rangecnt * SystemInformation->num_ctrs;
    }

  Status = IprWriteToSampleFile(HeaderBuffer, BufferLength);
  if (!IPR_SUCCESS(Status))
    {
      return Status;
    }
  return 0;
}
