/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program: ipr_histogram.c
 *
 * Abstract:  Module containing all the code to create a PC address histogram.
 *
 * Author: Ravindra Kulkarni
 *
 * Creation Date: May-94
 *
 * Modification History:
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Aug-94
 *    Defined a macro PRINT_FORMAT3 to take care of printf differences on VMS
 *    and OSF/1.
 *
 * Mike Demesey: Aug-94
 *    Removed the triple typecasts. Removed the call for pow and replaced by
 *    << operator.
 *
 * Ravindra Kulkarni: Jun-94
 *    Moved definition of ModeNames to ipr_os_vms.c. the initilisation is
 *    now done by InitializeModes which is a precondition for using ModeNames.
 *
 * Ravindra Kulkarni: May-94
 *   Wrote original code based on ipr_counting code by Greg Tarsa.
 */

#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "stdtypes.h"
#include "iprobe_struct.h"
#include "iprobe_eventlib.h"
#include "iprobe_hdr.h"
#include "iprobe_global.h"

#ifdef vms
#include "iprobe_vms.h"
#endif

#ifdef nt
#include "iprobe_nt.h"
#endif

#include "iprobe_histogram.h"

#ifdef nt
#define IPLNAME "IRQL"
#else
#define IPLNAME "IPL"
#endif


int32 DisplayNumber(int32 shift, int32 width, uint64 num);
int32 DisplayString(int32 shift, int32 width, const char *str);
uint64 ComputeHistogramElements(struct pc_ranges *pc_range);

/************************************************************************************/
/* Name: DoHistogram                                                                */
/*                                                                                  */
/* Purpose:   Setup the buffers and counters for a histogram run.                   */
/*                                                                                  */
/* Arguments: HistogramType: histogram type.                                        */
/*	      SelectedEvents: An array of events to measure.                        */
/*	      SelectedFrequency: Array of Frequencies to measure at                 */
/*	      SelectionCount:  Size of the prev 2 arrays                            */
/*	      ModesToUse: Program modes to measure.                                 */
/*	      StartTime: Number of seconds until start of measurement.              */
/*	      Duration: Number of seconds to measure.                               */
/*	      Interval: Output Interval.                                            */
/*	      RangeCount: The Number of pc ranges.                                  */
/*	      PCRange:  Pointer to the addr range list.                             */
/*	      HistogramCount: The number of histograms.                             */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS DoHistogram(IPR_HMASK HistogramType,  /* histogram type */
		       IPR_EVENT *SelectedEvents,     /* array of evts to measure */
		       IPR_FREQ *SelectedFrequency,     /* freqs to measure at */
		       IPR_COUNT SelectionCount,    /* size of the prev 2 arrays */
		       IPR_MMASK ModesToUse,     /* prog modes to measure in */
		       IPR_COUNT StartTime,     /* #secs till start of meas */
		       IPR_COUNT Duration,       /* # secs to measure */
		       IPR_COUNT Interval,       /* Output Interval */
		       uint32 RangeCount,   /*  Number of pc ranges */
		       struct pc_ranges *PCRange,/*  Pointer to the addr range list */
		       int32 HistogramCount)
{

  char BeginningAddressBuffer[32];             /*   Used to store begin pc address */

  char **ModeNames;


  /*
     These three symbols are declared volatile because their contents
     are changed asynchronously with respect to the code flow and we don't
     want the compiler to optimize code using them away because it thinks
     the values do not change.
     */

  FLAG BufferFullFlag = 0;     /* flag for buffer full condition */  
  FLAG CntrlCFlag = 0;    /* flag for CTRL-C hit by user */
  FLAG TimeOutFlag = 0;   /* flag for end time expiration */
  FLAG *StartedProcessExited;

  int32 CurrentSelectionNumber;
  int32 CurrentCounterNumber;
  int32 CurrentModeNumber;
  int32 CurrentHistogramSize;

  int32 CurrentInterruptLevel;
  int32 CurrentPriorityLevel;
  
  int32 FirstFlag = 1;
  int32 HistogramDisplayFlag = 0;           /*  This is used to indicate all the histogram elements were zeroes */
  int32 MaxEventNameLength;
  int32 MaxModeNameLength;               /*  used for saving max length of various modes like user, kernel etc. */
  int32 ModeNameLength;
  int32 NumberOfHistogramsOnLine;

  int CurrentIndex; /* VMS needs this to be an int. */
  int TempIndex;  /* VMS needs this to be an int. */
  int CPUContext;  /* VMS needs this to be an int. */


  IPR_COUNT NumberOfElements;       /* Number of elements in histogram array */
  IPR_CTRMASK CounterMask;                 /* mask for which counters are in use*/
  IPR_CTRMASK RequestedCounterMask;              /* requested counter mask */

  /* Event to counter/freq mapping symbols */
  IPR_CTRNUMBER *WhichCounter;
  IPR_EVENT *WhichEvent;
  IPR_FREQ *WhichFrequency;

  IPR_HISTELEM  **HistogramArea;
  IPR_HISTELEM  *TempHistogramPointer;

  IPR_HISTVALUE TempBeginAddress;
  IPR_HISTVALUE HighValue;
  IPR_HISTVALUE NumberValues;

  IPR_HMASK *TempHistogramType;

  IPR_STATUS Status;                   /* general status return */

  IPR_SIZE *HistogramSize;
  IPR_SIZE *HistogramIndex;


  struct event_info               /*    Used for saving event names */
  {
    const char *event_name;     /*  Event name */
    int32 event_len;              /*  Event name length   */
  };

  struct event_info *EventList;
  struct pc_ranges *PCRangePointer;
  struct pc_ranges *TempPCRange;
  struct tm *CurrentTime;

  system_info *SystemInformation;


  time_t TempCurrentTime;

  uint32 ActiveCPUMask = 0;  /* mask of active CPUs */
  uint32 BufferIndex;                  /* Used as a index while displaying histograms */
  uint32 CurrentRange;                /* general counter */
  uint32 EventIndex;
  uint32 HistogramMask;            /* Histogram mask */
  uint32 NumberOfBuffers;          /* Number of buffers requested */
  uint32 NumberOfBuffersToBeAllocated;
  uint32 *TotalNumberOfCounters;

  uint64 CurrentBufferNumber;
  uint64 *MinimumValue;
  uint64 MuxValue;

  TotalNumberOfCounters = GetTotalNumberOfCounters();

  ModeNames=OSGetModeNames();

  

  StartedProcessExited = GetStartedProcessExited();
  SystemInformation = GetSystemInformation();

  WhichEvent = (IPR_EVENT *) ckalloc(sizeof(IPR_EVENT) * (*TotalNumberOfCounters));
  WhichFrequency = (IPR_FREQ *) ckalloc(sizeof(IPR_FREQ) * (*TotalNumberOfCounters));
  WhichCounter = (IPR_CTRNUMBER *) ckalloc(sizeof(IPR_CTRNUMBER) * (*TotalNumberOfCounters));

  HistogramArea = (IPR_HISTELEM **)ckalloc(sizeof(IPR_HISTELEM) * HistogramCount);
  HistogramSize = (IPR_SIZE *) ckalloc(sizeof(IPR_SIZE) * HistogramCount);
  HistogramIndex = (IPR_SIZE *)ckalloc(sizeof(IPR_SIZE) * HistogramCount);
  TempHistogramType = (IPR_HMASK *)ckalloc(sizeof(IPR_HMASK) * HistogramCount);
  MinimumValue = (uint64 *)ckalloc(sizeof(uint64) * HistogramCount);

  InitializeModes();

  EventList = (struct event_info *)malloc(sizeof(struct event_info) * SelectionCount);

  if ((*IsDebug()) && !EventList)
  {
    (void) printf("Memory allocation error\n");
    exit(-1);
  }

  /*  Store the selected events in the event list */
  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
  {

    /* lookup the definitive event name in the main table */

    EventList[CurrentSelectionNumber].event_name = (char *) IprEventNumberToName(SelectedEvents[CurrentSelectionNumber]);
    EventList[CurrentSelectionNumber].event_len = strlen(EventList[CurrentSelectionNumber].event_name);
  }

  /* Find maximum length of selected event */

  MaxEventNameLength = EventList[0].event_len;

  for (CurrentSelectionNumber = 1; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
  {
    if (MaxEventNameLength < EventList[CurrentSelectionNumber].event_len)

      MaxEventNameLength = EventList[CurrentSelectionNumber].event_len;
  }

  HistogramMask = HistogramType.ipr_hmask_l_bits;

  /* Allocate the counters to prevent other processes' access */
  Status = IprAllocate();

  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  TempPCRange = PCRange->next;

  /* Create the histogram buffers */
  /*  Check for the pc_range being zero   */
  if(RangeCount)
  {
    Status = IprCreateHistogramBuffers(IPR_ALL_CPUS,   /* CPU mask */
				       HistogramType,                                  /* kind of hist data */
				       TempPCRange->start,                          /* beginning addr */
				       TempPCRange->end,                            /* no end addr */
				       TempPCRange->granularity,                    /* granularity */
				       0);                                         /* index = 0 */
    
  }
  else
  {
    Status = IprCreateHistogramBuffers (IPR_ALL_CPUS,   /* CPU mask */
					HistogramType,    /* kind of hist data */
					0,            /* beginning addr */
					0,            /* no end addr */
					0,            /* no granularity */
					0);           /* index = 0 */

  }
  if (!IPR_SUCCESS(Status))
  {
    if(Status == IPR__INVHBRANGE)
    {
      (void) printf("Invalid histogram buffer range specified.\n");
#ifdef vms
      (void) printf("Currently it is invalid to include 0x7fffffff and 0x80000000 in a single range\n"
                    "If it is necessary specify them in separate ranges\n");
#endif
#ifdef nt
      (void) printf("It is is invalid to include 0x7fffffff and 0x8000000 in a single range.\n"
                    "If it is necessary specify them in separate ranges\n");
#endif
    }
    if(Status == IPR__ALLOCFAIL)
    {
      (void) printf("Maximum (%d) number of histogram elements exceeded\n",
		    MAX_ELEMENTS);
    }
    return Status;
  }
  /*  Check if number of pcranges is greater than 1 and if the address
      histograms have been specified. In case the RangeCount = 1 and
      address histograms have been specified then this case is taken care
      in above IprCreateHostogramBuffers() call. In case the number of
      ranges is greater than 1 then IprCreateHostogramBuffers() is called
      repeatedly as follows. */

  if (RangeCount > 1 && (HistogramMask & IPR_HMASK_M_ADDRESS))
  {

    /*  Reset all bits in HistogramType except IPR_HMASK_M_ADDRESS */

    HistogramMask = HistogramMask & IPR_HMASK_M_ADDRESS;

    for (CurrentRange = 1; CurrentRange < RangeCount; CurrentRange++)
    {

      TempPCRange = TempPCRange->next;

      /* Create the histogram buffers */

      Status = IprAddAddressHistogramBuffers (IPR_ALL_CPUS,   /* CPU mask */
					      TempPCRange->start,              /* beginning addr */
					      TempPCRange->end,                /* end addr */
					      TempPCRange->granularity,                    /* granularity */
					      CurrentRange);       /* This parameter is used by
								      IprReadMultipleHistograms() to locate the
								      particular pc range and the output data */

      if (!IPR_SUCCESS(Status))
      {
	if(Status == IPR__INVHBRANGE)
	{
#ifdef vms
	  (void) printf("Currently it is invalid to include 0x7fffffff and 0x80000000 in a single range\n"
			"If it is necessary specify them in separate ranges\n");
#endif
#ifdef nt
	  (void) printf("It is is invalid to include 0x7fffffff and 0x8000000 in a single range.\n"
			"If it is necessary specify them in separate ranges\n");
#endif
	}

	if(Status == IPR__ALLOCFAIL)
	{
	  (void) printf("Maximum (%d) number of histogram elements exceeded\n",
			MAX_ELEMENTS);
	}
	return Status;
      }
    }

  }
  /*  At this stage all the histogram buffers have been created. */

  /*  Restore the HistogramMask */

  HistogramMask = HistogramType.ipr_hmask_l_bits;


  /*  Create the buffers for reading the histogram data. These will be supplied to IprReadMultipleHistograms(). */

  NumberOfBuffersToBeAllocated = 0;
  CurrentRange = 0;  /* This will be incremented in case of address ranges */

  if (HistogramMask & IPR_HMASK_M_TOTAL)
  {
    (void) IprHbElements ((*TotalNumberOfCounters), 1, 0, &HistogramSize[NumberOfBuffersToBeAllocated]);
    HistogramArea[NumberOfBuffersToBeAllocated]= (IPR_HISTELEM *) malloc (HistogramSize[NumberOfBuffersToBeAllocated] 
									  * sizeof (IPR_HISTELEM));
    HistogramIndex[NumberOfBuffersToBeAllocated] = CurrentRange;
    MinimumValue[NumberOfBuffersToBeAllocated] = 0;
    TempHistogramType[NumberOfBuffersToBeAllocated].ipr_hmask_l_bits  = IPR_HMASK_M_TOTAL;
    NumberOfBuffersToBeAllocated++;
  }

  if (HistogramMask & IPR_HMASK_M_IPL)
  {
    (void) IprHbElements ((*TotalNumberOfCounters), IPR_IPL_COUNT, 0, &HistogramSize[NumberOfBuffersToBeAllocated]);
    HistogramArea[NumberOfBuffersToBeAllocated]= (IPR_HISTELEM *) malloc (HistogramSize[NumberOfBuffersToBeAllocated]
									   * sizeof (IPR_HISTELEM));
    HistogramIndex[NumberOfBuffersToBeAllocated] = CurrentRange;
    MinimumValue[NumberOfBuffersToBeAllocated] = 0;
    TempHistogramType[NumberOfBuffersToBeAllocated].ipr_hmask_l_bits  = IPR_HMASK_M_IPL;
    NumberOfBuffersToBeAllocated++;
  }

  if (HistogramMask & IPR_HMASK_M_MODE)
  {
    (void) IprHbElements ((*TotalNumberOfCounters), IPR_MODE_COUNT, 0, &HistogramSize[NumberOfBuffersToBeAllocated]);
    HistogramArea[NumberOfBuffersToBeAllocated]= (IPR_HISTELEM *) malloc (HistogramSize[NumberOfBuffersToBeAllocated]
									  * sizeof (IPR_HISTELEM));
    HistogramIndex[NumberOfBuffersToBeAllocated] = CurrentRange;
    MinimumValue[NumberOfBuffersToBeAllocated] = 0;
    TempHistogramType[NumberOfBuffersToBeAllocated].ipr_hmask_l_bits  = IPR_HMASK_M_MODE;
    NumberOfBuffersToBeAllocated++;
  }

  if (HistogramMask & IPR_HMASK_M_PRIORITY)
  {
    (void) IprHbElements ((*TotalNumberOfCounters), IPR_PRIORITY_COUNT, 0, &HistogramSize[NumberOfBuffersToBeAllocated]);
    HistogramArea[NumberOfBuffersToBeAllocated] =
      (IPR_HISTELEM *) malloc (HistogramSize[NumberOfBuffersToBeAllocated] * sizeof (IPR_HISTELEM));
    HistogramIndex[NumberOfBuffersToBeAllocated] = CurrentRange;
    MinimumValue[NumberOfBuffersToBeAllocated] = 0;
    TempHistogramType[NumberOfBuffersToBeAllocated].ipr_hmask_l_bits  = IPR_HMASK_M_PRIORITY;
    NumberOfBuffersToBeAllocated++;
  }

  if (HistogramMask & IPR_HMASK_M_ADDRESS)
  {


    for (CurrentRange = 0, PCRangePointer = PCRange->next; CurrentRange < RangeCount;
	 CurrentRange++, PCRangePointer = PCRangePointer->next)
    {

      if (!IPR_SUCCESS(Status = IprHbElements ((*TotalNumberOfCounters),
					       PCRangePointer->end - PCRangePointer->start + 1,
					       PCRangePointer->granularity,
					       &HistogramSize[NumberOfBuffersToBeAllocated])))
      {
	(void) IprPrintMessage ("IprHbElements", Status, NULL);
	exit (-1);
      }
      HistogramArea[NumberOfBuffersToBeAllocated]= (IPR_HISTELEM *) malloc (HistogramSize[NumberOfBuffersToBeAllocated]
									    * sizeof (IPR_HISTELEM));
      if (!HistogramArea[NumberOfBuffersToBeAllocated])
      {
	(void) printf("Histogram memory allocation error\n");
	exit(-1);
      }
      HistogramIndex[NumberOfBuffersToBeAllocated] = CurrentRange;
      MinimumValue[NumberOfBuffersToBeAllocated] = PCRangePointer->start;
      TempHistogramType[NumberOfBuffersToBeAllocated].ipr_hmask_l_bits  = IPR_HMASK_M_ADDRESS;
      NumberOfBuffersToBeAllocated++; 
      CurrentRange++;
    }
  }

  /*  Now all the buffers have been allocated. i indicates the number of
      buffers allocated. */

  NumberOfBuffers = NumberOfBuffersToBeAllocated;

  /* Specify the Events to be recorded */

  Status = IprSpecifyOperations (IPR_ALL_CPUS,        /* cpu mask */
				 SelectedEvents,      /* event array */
				 SelectedFrequency,     /* freq array */
				 SelectionCount,    /* sz of both arrays */
				 ModesToUse,          /* mode selection */
				 &MuxValue,          /* MUX return */
				 &RequestedCounterMask);       /* rtn active ctrs */

  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }
  /*
   *  Get a handle on which counters are in use and the
   *  frequencies being used.
   *
   *  We assume that all CPUs are configured identically and
   *  so look at only one.
   *
   *  Returned vector has the measured event number,
   *  if any, in the corresponding element.
   */

  Status = IprEventCounterMap (IPR_CURRENT_CPU,   /* CPU to check */
			       WhichEvent,  /* vector of counters */
			       (*TotalNumberOfCounters));    /* vector length */

  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }
  /*  Find out which counter is being used for event 0, event 1 ... */

  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
  {
    for (CurrentCounterNumber = 0; CurrentCounterNumber < (*TotalNumberOfCounters); CurrentCounterNumber++)
    {
      if (SelectedEvents[CurrentSelectionNumber] == WhichEvent[CurrentCounterNumber])
      {
	WhichCounter[CurrentSelectionNumber] = CurrentCounterNumber;

	break;
      }
    }
  }

  Status = IprFreqCounterMap (IPR_CURRENT_CPU,    /* CPU to check */
			      WhichFrequency, /* return vector of freqs */
			      (*TotalNumberOfCounters));    /* vector length */

  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* set the active counter(s) */
  /* RequestedCounterMask gets put into sample header */

  Status = IprSetActiveCounters(IPR_ALL_CPUS, RequestedCounterMask, &CounterMask);
  if (!IPR_SUCCESS(Status))
  {
    if ((*IsDebug()))
    {
      (void) printf("Requested counter mask is: 0%o\n", RequestedCounterMask);
      (void) printf("Anticipated counter mask is: 0%o\n", CounterMask);
    }
  }

  /* If specified in cmd line, wait for start time */
  if (StartTime)
  {
    if ((*IsDebug()))
    {
      (void) printf("Waiting for start time (%d seconds since program start)\n",
		    StartTime);
    }

    WaitForTime(StartTime);
  }

  if (Duration)
  {
    time_t abs_end_time = time(0) + Duration;

    if ((*IsDebug()))
    {
      (void) printf("Establishing end timer (%d seconds)\n", Duration);
    }

    /* Establish a handler for the timer signal */
    Status = OSDeclareWakeUpHandler(&TimeOutFlag);
    if (!IPR_SUCCESS(Status))
    {
      return Status;
    }

    /* Schedule a WakeUp signal at the ending time */
    Status = OSScheduleWakeUpTime (abs_end_time);
    if (!IPR_SUCCESS(Status))
    {
      return Status;
    }
  }
  else
  {
    if ((*IsDebug()))
    {
      (void) printf("Establishing WakeUp handler\n");
    }

    /* Establish a handler for the timer signal */
    Status = OSDeclareWakeUpHandler(0);
    if (!IPR_SUCCESS(Status))
    {
      return Status;
    }
  }

  /* Establish a signal handler for user-requested abort */
  Status = OSDeclareCntrlCHandler (&CntrlCFlag);
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Establish a signal handler for the Ipr interrupt */
  Status = OSDeclarePFMHandler(&BufferFullFlag);
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  Status = OSScheduleRepeatingWakeUp(Interval);
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }
  /* Issue the command if one was given */

  Status = IssueCommandLine(SystemInformation->CommandLine);

  if (IPR_SUCCESS(Status))
  {
    Status = IprStart(IPR_ALL_CPUS, CounterMask); /*Start counting on all cpus */
    if (!IPR_SUCCESS(Status))
    {
      return Status;
    }
  }

  /* Wait for signals indicating "buffer full" or "time out" */
  while (IPR_SUCCESS(Status) && !(*StartedProcessExited) && !TimeOutFlag && !CntrlCFlag)
  {

    TempCurrentTime = time(0);       /* time stamp */

    if ((*IsDebug()))
    {
      (void) printf("Waiting for signal. . .");
      (void) fflush(stdout);
    }

    (void)OSWaitForSignal();

    if ((*IsDebug()))
    {
      if (TimeOutFlag)
      {
	(void) printf("TIMER signal received\n");
      }
      if (CntrlCFlag)
      {
	(void) printf("CTRL-C signal received\n");
      }
      if (BufferFullFlag)
      {
	(void) printf("BUFFER FULL signal received\n");
      } 
      if (!TimeOutFlag && !CntrlCFlag && !BufferFullFlag)
      {	
	(void) printf("OTHER signal received\n");
      } 
      (void) fflush(stdout);
    }

    /*
       When *any* signal occurs, read the histogram buffers and sum them.
       */

    for (CPUContext = -1 ; CPUContext < IPR_MAX_CPUS;)
    {
      /* Get all histograms for all CPUs */
      Status = IprReadMultipleHistograms (IPR_ALL_CPUS,   /* eligible CPU mask */
					  &CPUContext,    /* cpu ctx buffer */
					  HistogramArea, /* ptr to hist bufs */
					  HistogramSize, /* sizes of bufs */
					  MinimumValue,  /* min value array */
					  TempHistogramType, /* type array */
					  HistogramIndex,/* index for collecting multiple pc ranges */
					  IPR_HACTION_COPY,     /* collect action */
					  NumberOfBuffers);/* array size */

      if (!IPR_SUCCESS(Status))
      {
	return Status;
      }

      if (CPUContext >= IPR_MAX_CPUS)
      {
	break;
      }

      if ((*IsDebug()))
      {

	for (CurrentBufferNumber = 0; CurrentBufferNumber < NumberOfBuffers ; CurrentBufferNumber++)
	{
	  for (CurrentHistogramSize = 0; CurrentHistogramSize < HistogramSize[CurrentBufferNumber];
	       CurrentHistogramSize++)
	  {
	    (void) printf("Buffer:%d:%d: - %d\n", CurrentBufferNumber, CurrentHistogramSize,
			  HistogramArea[CurrentBufferNumber][CurrentHistogramSize]);
	  }
	}
      }

      TempPCRange = PCRange->next;
      TempCurrentTime = time(0);
      CurrentTime = localtime(&TempCurrentTime);

      (void) printf("Time %02d:%02d:%02d\n", CurrentTime->tm_hour, CurrentTime->tm_min, CurrentTime->tm_sec);

      for (BufferIndex = 0; BufferIndex < NumberOfBuffers; BufferIndex++)
      {
	HistogramDisplayFlag = FALSE;
	TempHistogramPointer = HistogramArea[BufferIndex];
	HistogramMask = TempHistogramType[BufferIndex].ipr_hmask_l_bits;

	if (HistogramMask == IPR_HMASK_M_TOTAL)
	{
	  if (FirstFlag)
	  {
	    (void) printf("Total Histograms\n");
	    (void) DisplayString(0, 3, "CPU");

	    for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	    {
	      (void) DisplayString(2, EventList[CurrentSelectionNumber].event_len,
				   EventList[CurrentSelectionNumber].event_name);
	    }
	    (void) printf("\n");
	  }

	  (void) DisplayNumber(0, 3, CPUContext);

	  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	  {

	    (void) DisplayNumber(2, EventList[CurrentSelectionNumber].event_len, 
				 TempHistogramPointer[WhichCounter[CurrentSelectionNumber]]);
	  }

	  (void) printf("\n");

	}  /*   End of total histogram display */

	else if (HistogramMask == IPR_HMASK_M_IPL)
	{
	  if (FirstFlag)
	  {
	    (void) printf(IPLNAME " Histograms\n");
	    (void) DisplayString(0, 3, "CPU");
	    (void) DisplayString(1, MaxEventNameLength, "Event");
	    (void) DisplayString(1, 10, IPLNAME "/ Count");
	    (void) printf("\n");
	  }

	  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	  {
	    (void) DisplayNumber(0, 2, CPUContext);
	    (void) DisplayString(2, MaxEventNameLength + 1, EventList[CurrentSelectionNumber].event_name);

	    for (CurrentInterruptLevel = 0, NumberOfHistogramsOnLine = 0; CurrentInterruptLevel < IPR_IPL_COUNT;
		 CurrentInterruptLevel++)
	    {
	      (void) IprHbIndex(CurrentInterruptLevel, WhichCounter[CurrentSelectionNumber],
				(*TotalNumberOfCounters), 0, 0, &TempIndex);
	      if (TempHistogramPointer[TempIndex])
	      {
		if (NumberOfHistogramsOnLine == 4)
		{
		  (void) printf("\n");
		  (void) DisplayNumber(5 + MaxEventNameLength, 0, 0);
		  NumberOfHistogramsOnLine = 0;
		}
		(void) DisplayNumber(1, 2, CurrentInterruptLevel);
		(void) printf("/");
		(void) DisplayNumber(1, 8, TempHistogramPointer[TempIndex]);
		NumberOfHistogramsOnLine++;
	      }
	    }
	    (void) printf("\n");
	  }
	}  /* End of IPL display */


	else if (HistogramMask == IPR_HMASK_M_PRIORITY)
	{
	  if (FirstFlag)
	  {
	    (void) printf("PRIORITY Histograms\n");
	    (void) DisplayString(0, 3, "CPU");
	    (void) DisplayString(1, MaxEventNameLength, "Event");
	    (void) DisplayString(1, 10, "PRIORITY / Count");
	    (void) printf("\n");
	  }

	  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	  {
	    (void) DisplayNumber(0, 2, CPUContext);
	    (void) DisplayString(2, MaxEventNameLength + 1, EventList[CurrentSelectionNumber].event_name);
	    for (CurrentPriorityLevel = 0, NumberOfHistogramsOnLine = 0; CurrentPriorityLevel < IPR_PRIORITY_COUNT ; 
		 CurrentPriorityLevel++)
	    {
	      (void) IprHbIndex (CurrentPriorityLevel, WhichCounter[CurrentSelectionNumber], (*TotalNumberOfCounters),
				 0, 0, &TempIndex);
	      if (TempHistogramPointer[TempIndex])
	      {
		if (NumberOfHistogramsOnLine == 4)
		{
		  (void) printf("\n");
		  (void) DisplayNumber(5 + MaxEventNameLength, 0, 0);
		  NumberOfHistogramsOnLine = 0;
		}
		(void) DisplayNumber(1, 2, CurrentPriorityLevel);
		(void) printf("/");
		(void) DisplayNumber(1, 8, TempHistogramPointer[TempIndex]);
		NumberOfHistogramsOnLine++;
	      }
	    }
	    (void) printf("\n");
	  }
	}  /* End of PRIORITY display */


	else if (HistogramMask == IPR_HMASK_M_MODE)
	{
	  if (FirstFlag)
	  {
	    (void) printf("MODE Histograms\n");
	    (void) DisplayString(0, 3, "CPU");
	    (void) DisplayString(1, MaxEventNameLength, "Event");

	    /*  Find maximum length of mode names */

	    MaxModeNameLength = strlen(ModeNames[0]);

	    for (CurrentModeNumber = 1; CurrentModeNumber < IPR_MODE_COUNT; CurrentModeNumber++)
	    {

	      ModeNameLength = strlen(ModeNames[CurrentModeNumber]);

	      if (MaxModeNameLength < ModeNameLength)
	      {
		MaxModeNameLength = ModeNameLength;
	      }
	    }

	    for (CurrentModeNumber = 0; CurrentModeNumber < IPR_MODE_COUNT; CurrentModeNumber++)
	    {
	      (void) DisplayString(2, MaxModeNameLength < 7 ? 7 : MaxModeNameLength, ModeNames[CurrentModeNumber]);
	    }
	    (void) printf("\n");
	  }

	  for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	  {
	    (void) DisplayNumber(1, 2, CPUContext);
	    (void) DisplayString(1, MaxEventNameLength, EventList[CurrentSelectionNumber].event_name);

	    for (CurrentModeNumber = 0; CurrentModeNumber < IPR_MODE_COUNT; CurrentModeNumber++)
	    {
	      ModeNameLength = strlen(ModeNames[CurrentModeNumber]);
	      (void) IprHbIndex (CurrentModeNumber, WhichCounter[CurrentSelectionNumber], 
				 (*TotalNumberOfCounters), 0, 0, &CurrentIndex);
	      (void) DisplayNumber(2, ModeNameLength < 7 ? 7 : ModeNameLength, TempHistogramPointer[CurrentIndex]);
	    }
	    (void) printf("\n");
	  }
	}  /* End of MODE display */


	else if (HistogramMask == IPR_HMASK_M_ADDRESS)
	{
	  if (FirstFlag)
	  {
	    (void) printf("ADDRESS Histograms\n");

	    (void) DisplayString(0, 3, "CPU");
	    (void) DisplayString(2, 16, "Address");

	    for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	      {
		(void) DisplayString(1, EventList[CurrentSelectionNumber].event_len > 7 ?
				     EventList[CurrentSelectionNumber].event_len :  7,
				     EventList[CurrentSelectionNumber].event_name);
	      }
	    (void) printf("\n");
	  }
	  
	  /* Start off at the beginning of the PC range */
	  
	  TempBeginAddress = TempPCRange->start;
	  NumberValues = TempPCRange->end - TempPCRange->start + 1;
	  
	  /* Use "do-while" form of loop to ensure that we go through */
	  /* the first time, even if the end of the range is at the */
	  /* maximum address */

	  do
	  {
	    for (EventIndex = 0; EventIndex < SelectionCount; EventIndex++)
	    {

	      /* See if there are any non-zero counts at the current */
	      /* histogram slot */

	      (void) IprHbIndex (TempBeginAddress, WhichCounter[EventIndex], (*TotalNumberOfCounters),
				 TempPCRange->start, TempPCRange->granularity, &CurrentIndex);

	      /* While we are checking, also calculate the max value */
	      /* for the current histogram slot to save a call to */
	      /* IprHbIndex later */

	      if (EventIndex == 0)
	      {
		(void) IprHbMaxValue (CurrentIndex, (*TotalNumberOfCounters), TempPCRange->start,
				      NumberValues, TempPCRange->granularity, &HighValue);
	      }
	      if (TempHistogramPointer[CurrentIndex])
		break;
	    }

	    /* If we broke before the end, non-zero counts exist */

	    if (EventIndex < SelectionCount)
	    {
	      HistogramDisplayFlag = TRUE;
	      (void) DisplayNumber(0, 2, CPUContext);
	      (void) sprintf(BeginningAddressBuffer, PRINT_FORMAT3, TempBeginAddress);
	      (void) DisplayString(3, 16, BeginningAddressBuffer);
 
	      for (CurrentSelectionNumber = 0; CurrentSelectionNumber < SelectionCount; CurrentSelectionNumber++)
	      {

		/* Find the histogram index for each of the counters */
		/* and display the corresponding counts */

		(void) IprHbIndex(TempBeginAddress, WhichCounter[CurrentSelectionNumber], (*TotalNumberOfCounters),
				  TempPCRange->start, TempPCRange->granularity, &CurrentIndex);
		(void) DisplayNumber(1, EventList[CurrentSelectionNumber].event_len > 7 ?
				     EventList[CurrentSelectionNumber].event_len : 7,
				     TempHistogramPointer[CurrentIndex]);
	      }
	      (void) printf("\n");
	    }

	    /* Use the previously calculated maximum value for the slot */
	    /* to determine the starting value for the next slot */

	    TempBeginAddress = HighValue + 1;

	    /* Loop continuation conditions are: */
	    /*   (1) The beginning of the slot is still inside the */
	    /*       histogram range; */
	    /*   (2) We have not overflowed the max address. */
	    /* Condition (2) is seen if the histogram goes up to the */
	    /* maximum address. */

	  } while (TempBeginAddress <= TempPCRange->end && TempBeginAddress > HighValue);

	  TempPCRange = TempPCRange->next;

	  if (!HistogramDisplayFlag)
	  {
	    (void) printf("------No non-zero elements found for the given range -----\n");
	  }
	}  /* End of ADDRESS display */

	(void) printf("\n");

      } /* End of for loop for displaying all buffers */

      FirstFlag = FALSE;

    } /* End of for loop for reading the different cpus */

  }/* end while (successful and !timeout and !interrupted) */

  /* Clear the repeating timer.  */
  Status = OSClearRepeatingWakeUp();
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Restore all signal handling to previous state */
  OSCancelPFMHandler();
  (void)OSCancelCntrlCHandler();
  (void)OSCancelWakeUpHandler();

  for (CurrentBufferNumber = 0; CurrentBufferNumber < NumberOfBuffers; CurrentBufferNumber++)
  {
    free(HistogramArea[CurrentBufferNumber]);
  }
  /* Stop the counters */
  Status = IprStop (IPR_ALL_CPUS, 0);
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Delete the buffers, now that we are done with them */
  Status = IprDeleteBuffers(IPR_ALL_CPUS);
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }
  /* Releae the counters for other folks */
  Status = IprDeallocate();
  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }
  return 1;
}

/* end DoHistogram() */

/************************************************************************************/
/* Name: DisplayNumber                                                              */
/*                                                                                  */
/* Purpose: It displays an integer on the screen. The printing is started after     */
/*          leaving shift number of spaces. If the num occupies lesser number of    */
/*          spaces than width then it is padded with spaces. In case the number     */
/*          of spaces occupied by num is greater than width then the width          */
/*          specification is ignored. If width is zero then num is not printed.     */
/*          shift number of spaces are printed on the screen.                       */
/*                                                                                  */
/* Arguments: ShiftAmount: The amount of spaces that the string is shifted to the   */
/*                         right on the screen.                                     */
/*            WidthOfScreen: The total width of the the screen.                     */
/*            NumberToBePrinted: The number to be written to the screen.            */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: This could just be a wrapper for display string!                          */
/*        (Print the int32 to a string, and then call DisplaySting! )               */
/*                                                                                  */
/************************************************************************************/

int32 DisplayNumber(int32 ShiftAmount, int32 WidthOfScreen, uint64 NumberToBePrinted)
{
  char NumberString[132];
  char SpaceString[132];

  int32 CurrentPositionInBuffer;
  int32 StringLength;

  /*  Initiate the buffer with spaces */
  for (CurrentPositionInBuffer = 0; CurrentPositionInBuffer < 132; CurrentPositionInBuffer++)
  {
    SpaceString[CurrentPositionInBuffer] = ' ';
  }
  /*  Add ShiftAmount number of spaces  */

  if (ShiftAmount > 0)
  {
    (void) fwrite(SpaceString, (uint32)ShiftAmount, 1, stdout);
  }
  if (WidthOfScreen == 0)
  {
    return 2;
  }
  /*  Move the NumberToBePrinted to the string  */
  (void) sprintf(NumberString, "%d", NumberToBePrinted);

  StringLength = strlen(NumberString);

  /*  Write the string on the screen  */
  (void) fwrite(NumberString, (uint32)StringLength, 1, stdout);

  /*  In case the WidthOfScreen is greater than length then add spaces    */
  if (WidthOfScreen - StringLength >= 0)
  {
    (void) fwrite(SpaceString, (uint32)(WidthOfScreen-StringLength), 1, stdout);
  }
  return 1;
}

/************************************************************************************/
/* Name: DisplayString                                                              */
/*                                                                                  */
/* Purpose:  It displays an string on the screen. The printing is started after     */
/*           leaving shift number of spaces. If the str occupies lesser number of   */
/*           spaces than width then it is padded with spaces. In case the number    */
/*           of spaces occupied by str is greater than width then the width         */
/*           specification is ignored. The width if zero does not write anything    */
/*           on display. It moves shift amount of spaces on screen.                 */
/*                                                                                  */
/* Arguments: ShiftAmount: The amount of spaces that the string is shifted to the   */
/*                         right on the screen.                                     */
/*            WidthOfScreen: The total width of the the screen.                     */
/*            OutputString: The string to be written to the screen.                 */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: This approach seems to be overkill.  How about, a for loop that uses      */
/*        fputc to print out the spaces.                                            */
/*                                                                                  */
/************************************************************************************/

int32 DisplayString(int32 ShiftAmount, int32 WidthOfScreen, const char *OutputString)
{
  int32 CurrentPositionInBuffer;
  int32 LengthOfGivenString;
  char SpaceString[132];

  /*  Initiate the buffer with spaces */
  for (CurrentPositionInBuffer = 0; CurrentPositionInBuffer < 132; CurrentPositionInBuffer++)
  {
    SpaceString[CurrentPositionInBuffer] = ' ';
  }

  /*  Add ShiftAmount number of spaces  */
  if (ShiftAmount > 0)
  {
    (void) fwrite(SpaceString, (uint32)ShiftAmount, 1, stdout);
  }

  if (WidthOfScreen == 0)
  {
    return 2;
  }

  LengthOfGivenString = strlen(OutputString);

  /*  Write the string on the screen  */
  (void) fwrite(OutputString, (uint32)LengthOfGivenString, 1, stdout);

  /*  In case the WidthOfScreen is greater than length then add spaces    */
  if (WidthOfScreen - LengthOfGivenString >= 0)
  {
    (void) fwrite(SpaceString, (uint32)(WidthOfScreen - LengthOfGivenString), 1, stdout);
  }

  return 1;
}

/************************************************************************************/
/* Name: ComputeHistogramElements                                                   */
/*                                                                                  */
/* Purpose: This computes the number of histogram buckets needed for this PC range. */
/*                                                                                  */
/* Arguments: PCRange                                                               */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: The number of buckets required for the PC range.                 */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

uint64 ComputeHistogramElements(struct pc_ranges *PCRange)
{
    return(((PCRange->end - PCRange->start) >> PCRange->granularity) + 1);
}
