/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program: iprobe_keylib.c
 *
 * Abstract: Functions and structures for searching and printing contents of
 *           keyword tables.
 *
 * Author: Greg Tarsa
 *
 * Creation Date: May-94
 *
 * Modification History:
 *
 * Jim Goddard:		27-Oct-1997
 *	Add new error message to indicate that a specified event name
 *	is not known.
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Aug-94
 *    Conditionally included unixio.h in case of vms and unistd.h in case of
 *    osf, and none in case of NT.
 *
 * Michael Dempsey: Aug-94
 *    Removed some unnecessary typecasts around address calculations
 *    that were causing 32 to 64 bit conversion problems for OSF/1
 *
 * Ravindra Kulkarni: Jun-94
 *    modified FindByName to take care of the exact matches.
 *
 * Greg Tarsa: May-94
 *    Added -pc_range parsing code.
 *    Wrote original code.
 */
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "stdtypes.h"

#ifdef vms
#include <unixio.h>
#endif

#ifdef unix
#include <unistd.h>
#endif

#include <errno.h>
#include "iprobe_struct.h"
#include "iprobe_eventlib.h"
#include "iprobe_keylib.h"
#include "iprobe_hdr.h"
#include "iprobe_global.h"

/*
    ParseEventString

    parse the specified argument as:
        eventname.intfreq.skip

    where:
    eventname   is the string name of the event
    intfreq     is the desired interrupt frequency
    skip        indicates that all but the skip-th interrupt should
                be ignored.

    If the interrupt frequency value is not valid for the specified counter,
    then it will be rounded to the nearest valid value. Default is "low
    frequency", for EV5 and undefined for EV6. A frequency of 0 turns a
    counter off.

    Initial value is only supported for EV6.

    If the event name is not recognized, then an appropriate message is
    printed to stdout and an error code is returned.

    Return values are placed into evtptr, freqptr, and skipptr.

    Function return value is non-zero for error, 0 for success.

    Algorithm:
    Check for invalid # of fields
    if valid, split into name/freq/initval
    validate name, default other fields.
    validate freq.
    place values into return areas
    return.
*/

/************************************************************************************/
/* Name: ParseEventString                                                           */
/*                                                                                  */
/* Purpose: This will take a string that describes the event, and extract out the   */
/*          proper information.                                                     */
/*                                                                                  */
/* Arguments: OriginialEventSpecification:  The string containing the event         */
/*                                          description.                            */
/*            EventPointer: A pointer to a variable that describes the event.       */
/*            FrequencyPointer: A pointer to a variable that describs the interrupt */
/*                              frequency. (HIGH or LOW)                            */
/*            EventPointer: A pointer to a variable that describes the skip amount. */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: 0 if no error, otherwise a number.                               */
/*                                                                                  */
/************************************************************************************/

int32 ParseEventString(char *OriginalEventSpecification, IPR_EVENT *EventPointer, IPR_FREQ *FrequencyPointer, 
		       IPR_COUNT *SkipPointer)
{
  uint32 CurrentShiftAmount;
  char *TempEventSpecification;
  char *NextInvalidCharater;

  IPR_EVENT EventNumber;

  char *EventField;
  char *FrequencyField = NULL;
  char *SkipField = NULL;
  uint64 Frequency = FREQ_DEF;
  uint64 Skip = 0;

  /* make a copy of src_spec for hacking apart */
  TempEventSpecification = (char *) ckalloc(strlen(OriginalEventSpecification) + 1);

  (void) strcpy(TempEventSpecification, OriginalEventSpecification);

  /* get start of event name */
  EventField = TempEventSpecification;

  /* isolate frequency field, if it exists */
  FrequencyField = strchr(EventField, EVT_SEP);
  if(FrequencyField)
  {
    *FrequencyField='\0';     /* null-terminate the first field */
    FrequencyField++;         /* skip separator */
  }

  /* Isolate the skip field, if there was a frequency field */
  if(FrequencyField)
  {
    SkipField = strchr(FrequencyField, EVT_SEP);
    if(SkipField)
    {
      *SkipField='\0';     /* null-terminate the third field */
      SkipField++;         /* skip separator */
    }
  }

  /* At this point if FLDx is non-zero, then it points
     to an null-terminated field value */

  /* Validate event name field */
  if(*EventField == '\0')  /* Null event name: error */
  {
    (void) fprintf(stdout, "? ERROR: Null event name specified: %s\n",
		   OriginalEventSpecification);
    free(TempEventSpecification);
    return 2;
  }

  EventNumber = FindEventNumberByName(EventField, TRUE);
  if(EventNumber == 0)
  {
    (void) fprintf(stdout, "? ERROR: Unknown event name specified: %s\n",
		   OriginalEventSpecification);
    free(TempEventSpecification);
    return 6;
  }

  /* Process frequency field, if it exists */
  if(FrequencyField)
  {
    /* check for frequency keywords */
    if(CaseInsensitiveStringCompare(FrequencyField, "default") || *FrequencyField == '\0')
    {
      Frequency = FREQ_LOW;
    }
    else if(CaseInsensitiveStringCompare(FrequencyField, "low"))
    {
      Frequency = FREQ_LOW;    /* 2^16 */
    }
    else if(CaseInsensitiveStringCompare(FrequencyField, "high"))
    {
      Frequency = FREQ_HIGH;   /* 2^8 */
    }
    else
    {
      /* process frequency values */
      Frequency = strtol(FrequencyField, &NextInvalidCharater, 10);   /* convert string to decimal */
      if(*NextInvalidCharater != '\0')
      {
	(void) fprintf(stdout, "?Error in eventspec(\"%s\")\n",OriginalEventSpecification);
	(void) fprintf(stdout,
		       "-Unrecognized characters in frequency field: %s\n",
		       NextInvalidCharater);

	free(TempEventSpecification);
	return 4;
      }
    } /* end else "process freq value" */
  }
  else
  {
    /* no frequency specified? Use the default. */
    Frequency = FREQ_LOW;
  }/* end if(FrequencyField) */

  /*
     Adjust the specified frequency to an appropriate power of two, if
     necessary.
     */

  for(CurrentShiftAmount = 1; CurrentShiftAmount < 32; CurrentShiftAmount++)
    if (Frequency == (uint32) (1<<CurrentShiftAmount) ||
	Frequency == FREQ_LOW || Frequency == FREQ_HIGH)
      
      break;
    else
    {
      /* If the freq is not exactly a power of 2, make it
	 the smallest power of 2 that it will fit into. */
      if(Frequency <(uint32)(1<<CurrentShiftAmount))
      {
	(void) fprintf(stdout, "%%Warning: rounding %s.%d to %s.%d\n",
		       IprEventNumberToName(EventNumber), Frequency,
		       IprEventNumberToName(EventNumber),(1<<CurrentShiftAmount));
	Frequency =(1<<CurrentShiftAmount);
	break;
      }
    }

  /* Process skip field, if it exists */
  if(SkipField)
  {
    /* process frequency values */
    Skip = strtol(SkipField, &NextInvalidCharater, 10);   /* convert string to decimal */
    if(*NextInvalidCharater != '\0')
    {
      (void) fprintf(stdout, "?Error in eventspec(\"%s\")\n",OriginalEventSpecification);
      (void) fprintf(stdout,"-Unrecognized characters in skip field: %s\n",
		     *NextInvalidCharater);

      free(TempEventSpecification);
      return 4;
    }
  }/* end if(SkipField) */

  /*
   *  At this point, freq is the canonical frequency value.
   */

  /* set up for return values */
  *EventPointer = EventNumber;
  *FrequencyPointer = Frequency;
  *SkipPointer = Skip;

  free(TempEventSpecification);
  return 0;
}/* end ParseEventString() */


/************************************************************************************/
/* Name: WhatBase                                                                   */
/*                                                                                  */
/* Purpose: This returns the base of the specified string.                          */
/*                                                                                  */
/* Arguments: NumberPointer: A pointer to the string to be analyzed.                */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: The base: ie, base 16 returns 16.                                */
/*                                                                                  */
/* Notes: If the first two characters are '0x', then it is assumed to be hex. If    */
/*        the first char is '0' and 2nd char not 'x', it is assumed to be octal.    */
/*        Otherwise, it is assumed to be decimal.                                   */
/*                                                                                  */
/************************************************************************************/

int32 WhatBase(char *NumberPointer)
{
  /* Is the first character 0? */
  if(*NumberPointer++ == '0')
  {
    /* Is the next character X? */
    if(*NumberPointer == 'x' || *NumberPointer == 'X')
    {
      /* Yes. We are hexadecimal. */
      return 16;
    }
    else

      /* 2nd char not X, we are octal */
      return 8;
  }
  else
    /* 1st character not '0', we are decimal */
    return 10;
  return 0;
}/* end WhatBase() */


/************************************************************************************/
/* Name: ParsePCRange                                                               */
/*                                                                                  */
/* Purpose: This takes a string of the form "%l-%l" or "%l-%l.%l", and it extracts  */
/*          out a starting and ending address, as well as a Granularity.            */
/*                                                                                  */
/* Arguments: RangeString: The string that contains the range information.          */
/*            BeginningAddress:  Pointer to the starting address of the range.      */
/*            EndingAddress:  Pointer to the ending address of the range.           */
/*            Granularity: Size of the grain.                                       */
/*                                                                                  */
/* Output: BeginningAddress:                                                        */
/*         EndingAddress:                                                           */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: Couldn't this be simply a sprintf?                                        */
/*                                                                                  */
/************************************************************************************/
/*
    ParsePCRange

    parse the specified argument as:
        begpc-endpc.Granularity

    where:
    begpc       is a starting address value
    endpc       is the ending address value
    Granularity is the size of the address increment for histograms

    If begpc is missing, then it is assumed to be 0.

    If endpc is missing, then it is assumed to be -1.

    If Granularity is missing, the it is assumed to be DEF_GRANULARITY(1).

    If a value begins with '0x' it is assumed to be a hexadecimal value,
    if it begins with a simple '0', then it is assumed to be an octal value,
    otherwise it is assumed to be decimal.

    Function return value is non-zero for error, 0 for success.
*/


int32 ParsePCRange(char *RangeString,         /* range spec */
		   ADDR *BeginningAddress,          /* ret start addr */
		   ADDR *EndingAddress,          /* ret end addr */
		   uint64 *Granularity) /* size of grain */
{

  char *CharacterPointer;   /* general character pointer */
  char *NextInvalidCharacter;
  ADDR Address;      /* general address cell */
  int32 NumberBase;       /* base of number */

#ifdef unix
  ADDR TemporaryAddress;
#endif

#if defined vms || defined nt
  int32 TemporaryAddress;
#endif

  CharacterPointer = RangeString;

  /* get the number base */
  NumberBase = WhatBase(CharacterPointer);

  /* convert the string to a long; on OSF1 this should be 64 bits */
  /*  Temporarily change the base to 16 */
  TemporaryAddress = (ADDR) strtoul(CharacterPointer, &NextInvalidCharacter, 16);
  Address = TemporaryAddress;

  /*
     strtol set NextInvalidCharacter to be the char that terminated the scan.
     If that string was "\0', then we default the ending addr and gran.
     If that string was ".", then we default the ending addr and get gran.
     If that string was "-", then we scan for an ending addr.
     If that string was anything else, we have an error.
     */

  /* Default end and Granularity */
  if (*NextInvalidCharacter == '\0')
  {
    *BeginningAddress = Address;
    *EndingAddress = DEF_END_ADDRESS;
    *Granularity = DEF_GRANULARITY;
    return 0;
  }
  else if(*NextInvalidCharacter == '.')
  {
    *BeginningAddress = Address;
    *EndingAddress = DEF_END_ADDRESS;
  }
  else if(*NextInvalidCharacter != '-')
  {
    (void) fprintf(stderr, "?Invalid char seen when looking for end addr: \"%s\"\n",
		   NextInvalidCharacter);
    return 2;
  }

  /* assign the beginning address */
  *BeginningAddress = Address;

  /* If end addr is not defaulted, then get it. */
  if(*NextInvalidCharacter != '.')
  {
    /* Evaluate end addr */
    CharacterPointer = NextInvalidCharacter + 1;

    /* get the number base */
    NumberBase = WhatBase(CharacterPointer);

    /* convert the string to a long; on OSF1 this should be 64 bits */
    TemporaryAddress = (ADDR) strtoul(CharacterPointer, &NextInvalidCharacter, 16);
    Address = TemporaryAddress;
    /*
       strtol set NextInvalidCharacter to be the char that terminated the scan.
       If that string was "\0', then we default the gran.
       If that string was ".", then set the ending addr and get gran.
       */
    /* set end addr and default Granularity */
    if(*NextInvalidCharacter == '\0')
    {
      /* end address of 0 gets forced to last address */
      *EndingAddress =(Address == 0) ? -1 : Address;

      *Granularity = DEF_GRANULARITY;
      return 0;
    }
    else
      if(*NextInvalidCharacter != '.')
      {
	(void) fprintf(stderr,
		       "?Invalid char seen when looking for Granularity: \"%s\"\n",
		       NextInvalidCharacter);
	return 4;
      }

    /* assign the end address */
    *EndingAddress = Address;
  }

  /* Evaluate the Granularity */
  CharacterPointer = NextInvalidCharacter + 1;

  /* get the number base */
  NumberBase = WhatBase(CharacterPointer);

  /* convert the string to a long; on OSF1 this should be 64 bits */
  *Granularity = (uint64) strtol(CharacterPointer, &NextInvalidCharacter, NumberBase);

  /*  Granularity is forced to be 2 in case user enters it to be less
      then 2  */

  if(*Granularity < 2)
  {
    *Granularity = 2;
  }

  /* If we are not at EOL, then we have an error */
  if(*NextInvalidCharacter != '\0')
  {
    (void) fprintf(stderr, "?Extraneous characters in pc range: \"%s\"\n",
		   NextInvalidCharacter);
    return 6;
  }
  return 0;

}/* end ParsePCRange */

/************************************************************************************/
/* Name: CaseInsensitiveStringCompare                                               */
/*                                                                                  */
/* Purpose: This will determine whether or not two strings are identical ignoring   */
/*          the case of both strings.                                               */
/*                                                                                  */
/* Arguments: String1: A pointer to the first string.                               */
/*            String2: A pointer to the second string.                              */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: If the strings are equal, this function will return a 1,         */
/*                 otherwise it will return a 0.                                    */
/*                                                                                  */
/* Notes: There are native functions to do this, however they aren't portable.      */
/*                                                                                  */
/************************************************************************************/

int32 CaseInsensitiveStringCompare(char *String1, char *String2)
{
  for(; *String1 && *String2 ; String1++, String2++)
    if((isupper(*String1) ? tolower(*String1) : *String1) !=
       (isupper(*String2) ? tolower(*String2) : *String2))
    {
      return 0;
    }
  if (*String1 || *String2)
  {
    return 0;
  }     
  return 1;
}

/************************************************************************************/
/* Name: FindByName                                                                 */
/*                                                                                  */
/* Purpose: This searches through the given table in order to find entry to match   */
/*          the given name.                                                         */
/*                                                                                  */
/* Arguments: Table: The Table in which the keyword is being look for.              */
/*            InputString: The keyword that has to be matched.                      */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: A pointer to the entry describing                                */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindByName(struct keywords *Table, char *InputString)
{
  int32 MatchesFound = 0;
  int32 ExactMatch = 0;

  struct keywords *CurrentEntryPointer;
  struct keywords *LastMatch=NULL;
  struct keywords *LastExactMatch=NULL;
  struct keywords *UnknownKeyword;

  UnknownKeyword = GetUnknownKeyword();


  /*BUG: This is some pretty nasty pointer manipulation... use CurrentEntryPointer[count] */
  for (CurrentEntryPointer = Table;  CurrentEntryPointer->name; CurrentEntryPointer++)
  {
    /* if we have a match, make note. */
    if (strncmp(InputString, CurrentEntryPointer->name, strlen(InputString))==0)
    {
      if (strlen(InputString) == strlen(CurrentEntryPointer->name))
      {
	ExactMatch++;
	LastExactMatch = CurrentEntryPointer;
      }

      MatchesFound++;
      if (MatchesFound > 1 && ExactMatch == 0)
      {
	if (MatchesFound == 2)
	{
	  (void) printf("?Multiple matches found for \"%s\":\n", InputString);
	}
	(void) printf("\t%c%s\n", FLAG_CHAR1, LastMatch->name);
      }
      LastMatch = CurrentEntryPointer;
    }

  }
  if (ExactMatch)
  {
    if (ExactMatch == 1)
    {
      return LastExactMatch;
    }
    else
    {
      (void) printf("Error in name Table.\n");
      return 0;
    }
  }

  if (MatchesFound > 1)
  {
    (void) printf("\t%c%s\n", FLAG_CHAR1, LastMatch->name);
    return UnknownKeyword;
  }

  if (MatchesFound == 0)
  {
    (void) printf("?Unrecognized keyword: \"%s\"\n",InputString);
    return UnknownKeyword;
  }

  return LastMatch;
}/* end FindByName() */

/************************************************************************************/
/* Name: FindOptionByName                                                           */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByName, it searchs the OptionTable for an     */
/*          option entry that matches the given string.                             */
/*                                                                                  */
/* Arguments: InputString: The string that is going to be looked for.               */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/
struct keywords *FindOptionByName(char *InputString)
{ 
  struct keywords *OptionTable;
  OptionTable = GetOptionTable();

  return FindByName(OptionTable, InputString); 
}

/************************************************************************************/
/* Name: FindMethodByName                                                           */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByName, it searchs the OptionTable for an     */
/*          method entry that matches the given string.                             */
/*                                                                                  */
/* Arguments: InputString: The string that is going to be looked for.               */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/
struct keywords *FindMethodByName(char *InputString)
{ 
  struct keywords *MethodTable;

  MethodTable = GetMethodTable();
  
  return FindByName(MethodTable, InputString); 


}

/************************************************************************************/
/* Name: FindModeByName                                                             */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByName, it searchs the OptionTable for an     */
/*          mode entry that matches the given string.                               */
/*                                                                                  */
/* Arguments: InputString: The string that is going to be looked for.               */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindModeByName(char *InputString)
{ 
  struct keywords *ModeTable;  
  ModeTable = OSGetModeTable();

  return FindByName(ModeTable, InputString); 
}

/************************************************************************************/
/* Name: FindByNumber                                                               */
/*                                                                                  */
/* Purpose: This returns the element matching the given number.                     */
/*                                                                                  */
/* Arguments: Table: The Table in which the element number is being looked for.     */
/*            ElementNumber: The number of the required element.                    */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: A pointer to the entry describing                                */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindByNumber(struct keywords *Table, int32 ElementNumber)
{
  struct keywords *CurrentEntryPointer;
  struct keywords *UnknownKeyword;

  UnknownKeyword = GetUnknownKeyword();

  /* BUG: I think a do while loop would be better here. */

  for (CurrentEntryPointer=Table;
       CurrentEntryPointer->name && CurrentEntryPointer->num != ElementNumber; 
       CurrentEntryPointer++)
  {
    ;
  }
  if (CurrentEntryPointer->name == 0)
  {
    return UnknownKeyword;
  }

  return CurrentEntryPointer;
}/* end FindByNumber() */

/************************************************************************************/
/* Name: FindOptionByNumber                                                         */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByNumber, it searchs the OptionTable for the   */
/*          given element number.                                                   */
/*                                                                                  */
/* Arguments: ElementNumber: The element number of the OptionTable that is needed.  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindOptionByNumber(int32 ElementNumber)
{ 
  struct keywords *OptionTable;
  OptionTable = GetOptionTable();
  return FindByNumber(OptionTable, ElementNumber); 
}

/************************************************************************************/
/* Name: FindMethodByNumber                                                         */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByNumber, it searchs the MethodTable  for the  */
/*          given element number.                                                   */
/*                                                                                  */
/* Arguments: ElementNumber: The element number of the MethodTable  that is needed. */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindMethodByNumber(int32 ElementNumber)
{
  struct keywords *MethodTable;
  MethodTable = GetMethodTable();

 return FindByNumber(MethodTable, ElementNumber); 
}

/************************************************************************************/
/* Name: FindModeByNumber                                                           */
/*                                                                                  */
/* Purpose: This is a wrapper to FindByNumber, it searchs the ModeTable for the    */
/*          given element number.                                                   */
/*                                                                                  */
/* Arguments: ElementNumber: The element number of the ModeTable that is needed.   */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: A pointer to the matching entry.                                 */
/*                                                                                  */
/************************************************************************************/

struct keywords *FindModeByNumber(int32 ElementNumber)
{ 
  struct keywords *ModeTable;
  ModeTable = OSGetModeTable();
  
  return FindByNumber(ModeTable, ElementNumber);    
}

