/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program: ipr_main
 *
 * Abstract:   Mainline module for the IPROBE utility.
 *
 * Author: Greg Tarsa
 *
 * Creation Date: Dec-93
 *
 * Modification History:
 *
 * Jim Goddard 27-Oct-1997
 *	Modify error message count after call to processarguments
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Sep-94
 *    Removed the argument granularity from the do_histogram as it is now
 *    obsolete. The granularity is being passed through pc range structure.
 *
 * Ravindra Kulkarni: Aug-94
 *    Called CreateObject in case of NT only. This routine is NT specific.
 *    Used an intermediate variable for setting the default low and default
 *    highest address ( 0xffffffff ) in case of vms only. The direct
 *    assignment is made in case of osf.
 *
 * Ravindra Kulkarni: Jul-94
 *    While collecting samples if user does not enter pc ranges then it was
 *    assumed that user is entering one range from 00000000-ffffffff. This
 *    assumption is removed and now onwards if a pc range is not entered it
 *    will be assumed that number of pc ranges is zero. It saves time since
 *    sampling code need not compare each time whether the sample belongs to
 *    the specified range or not.
 *
 * Ravindra Kulkarni: Jun-94
 *    Added collect option for entering user defined sample data.
 *
 * Ravindra Kulkarni: Jun-94
 *    Added input option for entering command file name.
 *
 * Ravindra Kulkarni: Jun-94
 *    Added buffer_size option for entering the sample buffer size.
 *
 * Greg Tarsa: May-94
 *    Added support for PC address ranges.
 *    Changed code to use new FLAGS declaration.
 *
 * Greg Tarsa: Apr-94
 *    Merged in Ravindra's enhanced help code.
 *
 * Greg Tarsa: Mar-94
 *    Added global debug flag.
 *    Added support for skip counts.
 *    Changed finish time to be a delta time.
 *
 * Greg Tarsa: Dec-93
 *    Wrote original code.
 *
 */

#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "stdtypes.h"

#include "iprobe_struct.h"
#include "iprobe_hdr.h"
#include "iprobe_eventlib.h"
#include "iprobe_histogram.h"
#include "iprobe_global.h"

#ifdef nt
#include "iprobe_nt.h"
#endif
#ifdef vms
#include "iprobe_vms.h"
#endif
#ifdef unix
#include "iprobe_osf1.h"
#endif

/* general program globals */


/************************************************************************************/
/* Name: main                                                                       */
/*                                                                                  */
/* Purpose: This is the entry point to the iprobe program.                          */
/*                                                                                  */
/* Arguments: argc: The number of arguments passed to iprobe.                       */
/*            argv: an array of pointers to those arguments.                        */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: The exit code.                                                   */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

int main(int argc, char **argv)
{
  char **CommandArgv;        /*  Pseudo argv argument    */
  char CommandLine[1024];    /*  command line to issue */
  char *OutputFileName;      /*  Output file for storing the data */

  FLAGS CommandWindowFlags;  /* command window creation/retention flags */
  FLAGS Flags;               /*  mask for the options entered by user. For every option selected a bit is made high
			      which otherwise would be zero. */
  FLAGS Method;              /*  It indicates the method (sample, count, address etc) */

  int32 CommandArgc;           /*  Pseudo argc argument    */
  int32 HistogramCount = 0;
  int32 ReturnValue;           /*  Error returned by ProcessArguments */
  int32 SkipSpecified;

  IPR_COUNT *SkipArray;      /*  contains user entered skip array */
  IPR_COUNT ArraySize;       /*  size of arrays  */
  IPR_COUNT BufferSize;            /*  Buffer size is used for creating sample buffers */
  IPR_COUNT Interval;              /*  interval for display of results - pc counting */
  IPR_COUNT TotalNumberOfBuffers;  /*  Number of buffers allocated in sampling */
  IPR_COUNT TotalNumberOfRanges = 0;    /*  Number of ranges in the linked list */


  IPR_DMASK DataMask;              /*  The mask for the specified data to be collected.(for sampling only) */
  IPR_DMASK QuadDataMask;          /*  The mask for the specified quad data to be collected.(for sampling only) */

  IPR_EVENT *EventArray;           /*  contains user entered event array */
  IPR_FREQ  *FrequencyArray;       /*  contains user entered freq array */

  IPR_MMASK ModeMask;              /*  The modemask for measurement (kernel, user etc) */
  IPR_HMASK HistogramData;         /*  contains the type/s of histogram desired   */

  IPR_STATUS Status;               /*  General ipr returned status */

  struct pc_ranges PCAddressList;  /*  linked list for pc ranges */
  struct pc_ranges *PCRangePointer;

  time_t EndTime;                  /*  End time for all operation */
  time_t StartTime;                /*  Start time for all operation */


  uint32 *TotalNumberOfCounters;     

  uint64 ElementCount;
  uint64 CurrentEventNumber;
  uint64 CurrentPCRange;  


  InitiateOS();

  TotalNumberOfCounters = GetTotalNumberOfCounters();


  if (((*TotalNumberOfCounters) = IprNumCounters()) == 0)
	{
		printf("ERROR: unable to detect processor type\n");
		exit(EXIT_FAILURE);
	}

  /* Initalise the address range */

  (void) memset(&PCAddressList, 0, sizeof (struct pc_ranges));

  EventArray = (IPR_EVENT *) ckalloc(sizeof(IPR_EVENT) * (*TotalNumberOfCounters));

  FrequencyArray = (IPR_FREQ *) ckalloc(sizeof(IPR_FREQ) * (*TotalNumberOfCounters));

  SkipArray = (IPR_COUNT *) ckalloc(sizeof(IPR_COUNT) * (*TotalNumberOfCounters));

  (void) OSGetSystemInformation();      /* populate the SystermInformation structure */

  ReturnValue = ProcessArguments(argc,           /*  command line argument argc */
				 argv,           /*  command line argument argv */
				 &Flags,         /*  It is a bit mask representing the options chosen by user */
				 &TotalNumberOfBuffers,  /*  Number of buffers returned */
				 &BufferSize,   /*  Size of sample buffer */
				 &StartTime,         /*  start time returned */
				 &EndTime,        /*  end time returned */
				 &Interval,      /*  display interval */
				 &OutputFileName,       /*  name of output file */
				 &Method,        /*  desired method sample, count etc */
				 &ModeMask,      /*  desired mode/s kernel, super, user etc */
				 &DataMask,      /*  desired data for samples    */
				 &QuadDataMask,  /*  desired quad data for samples */
				 &PCAddressList,  /*  linked list of pc address ranges */
				 EventArray,         /*  array of selected events  */
				 FrequencyArray,        /*  array of selected frequencies */
				 SkipArray,        /*  array of selected skip values */
				 &ArraySize,     /*  size of above arrays */
				 CommandLine,	/*  command line to issue */
				 sizeof(CommandLine) - 1); /* length of usable space in the command line buffer */

  if (ReturnValue)
  {
    (void) printf("Errors detected processing command line arguments: %d\n", ReturnValue);
    exit(ReturnValue);
  }

  /*  If user specifies input as an option then the arguments from the
      specified file are read as if the command line arguments by
      FGetCommandOptions and are returned as argv and argc format.
      ProcessArguments is again invoked with the new argv and argc
      arguments.  */

  if (Flags & CL_INPUT)
  {
    (void) FGetCommandOptions(OutputFileName, &CommandArgv, &CommandArgc);

    ReturnValue = ProcessArguments(CommandArgc,       /*  command line argument argc */
				   CommandArgv,       /*  command line argument argv */
				   &Flags,         /*  It is a bit mask representing the options
						       chosen by user */
				   &TotalNumberOfBuffers,  /*  Number of buffers returned */
				   &BufferSize,   /*  Size of sample buffer */
				   &StartTime,         /*  start time returned */
				   &EndTime,        /*  end time returned */
				   &Interval,      /*  display interval */
				   &OutputFileName,       /*  name of output file */
				   &Method,        /*  desired method sample, count etc */
				   &ModeMask,      /*  desired mode/s kernel, super, user etc */
				   &DataMask,      /*  desired data for sampling */
				   &QuadDataMask,  /*  desired quad data for samples */
				   &PCAddressList,  /*  linked list of pc address ranges */
				   EventArray,         /*  array of selected events  */
				   FrequencyArray,        /*  array of selected frequencies */
				   SkipArray,        /*  array of selected skip values */
				   &ArraySize,      /*  size of above arrays */
				   CommandLine,	/*  command line to issue */
				   sizeof(CommandLine) - 1); /* length of usable space in the command line buffer */
    
    if (ReturnValue)
      {
      (void) printf("Error processing command line arguments: %d\n", ReturnValue);
      exit(ReturnValue);
    }
  }
  
  
  if ((*IsDebug()))
  {

    /* Debugging code, to insure that ProcessArguments worked properly */

    (void) printf("\tTotalNumberOfBuffers = %u\n", TotalNumberOfBuffers);
    (void) printf("\tBufferSize  = %u\n", BufferSize);
    (void) printf("\tStartTime = %u (about %d)\n", StartTime, StartTime - time(0));
    (void) printf("\tend = %u\n", EndTime);
    (void) printf("\tInterval = %u\n", Interval);
    (void) printf("\tOutputFileName = %s\n", OutputFileName);
    (void) printf("\tMethod = %u\n", Method);
    (void) printf("\tcommand = %s\n", CommandLine);
    (void) printf("\tmode bits = %d\n", ModeMask.ipr_mmask_l_bits);
    (void) printf("%d events were found:\n", ArraySize);

    for (CurrentEventNumber = 0; CurrentEventNumber < ArraySize; CurrentEventNumber++)
    {
      (void) printf("\t%3d [%8d.%-1d] (%s)\n",
		    EventArray[CurrentEventNumber],
		    FrequencyArray[CurrentEventNumber],
		    SkipArray[CurrentEventNumber],
		    IprEventNumberToName (EventArray[CurrentEventNumber]));
    }

    if (PCAddressList.next)
    {
      (void) printf("The following address ranges were found:\n");

      for (PCRangePointer = PCAddressList.next; PCRangePointer; PCRangePointer = PCRangePointer->next)
      {
	(void) printf(ADDRFORMAT" - "ADDRFORMAT " %x\n",
		      PCRangePointer->start, PCRangePointer->end, PCRangePointer->granularity);
      }

    }

  }


  /*  Compute the number of pc_ranges entered by user */

  if ((Method & METH_ADDR)||(Method & METH_SAMPLE))
  {
    if (PCAddressList.next)
    {
      PCRangePointer = PCAddressList.next;

      while (PCRangePointer)
      {
	TotalNumberOfRanges++;

	PCRangePointer = PCRangePointer->next;
      }

    }
  }

  (void) ReadUserOptions(EventArray,  /*  event array                      */
			 FrequencyArray,                /*  frequency array                  */
			 SkipArray,                     /*  skip array                       */
			 ArraySize,                     /*  array size                       */
			 ModeMask,                      /*  mode mask                        */
			 DataMask,                      /*  data mask for collecting samples */
			 QuadDataMask,                  /*  quad data mask                   */
			 TotalNumberOfBuffers,          /*  buffer count                     */
			 BufferSize,                    /*  sample buffer size               */
			 StartTime,                     /*  start delay                      */
			 EndTime,                       /*  duration                         */
			 Interval,                      /*  interval value                   */
			 &PCAddressList,                /*  Ptr to addr range list           */
			 OutputFileName,                /*  output file                      */
			 Method,                        /*  iprobe method                    */
			 CommandLine);

  /* Display the "user-friendly header" */

  if (!(Flags & CL_QUIET))
  {
    (void) DisplaySystemHeader(stdout);

    (void) DisplayUserOptions(stdout);
  }


  /*  default address range if user does not specify address range while
      collecting address histograms. */

  if (!(Flags & CL_RANGE) &&  (Method & METH_ADDR))
  {
#if !defined nt
    uint64 hi;
    uint64 lo;
#endif

    PCAddressList.next = (struct pc_ranges *) malloc(sizeof(struct pc_ranges));

#ifdef vms
    lo = 0;
    hi = -1;
    PCAddressList.next->start = (uint64)(int64) lo;
    PCAddressList.next->end = (uint64)(int64) hi;
#endif
#ifdef nt
    PCAddressList.next->start = 0;
    PCAddressList.next->end =  DEF_END_ADDRESS;
#endif
#ifdef unix
    PCAddressList.next->start = (uint64)(int64) lo;
    PCAddressList.next->end = (uint64)(int64) hi;
#endif
    PCAddressList.next->granularity = (uint32)0x00000000;
    PCAddressList.next->next = 0;
    TotalNumberOfRanges = 1;

  }

  if (Method != METH_SAMPLE)
  {
    SkipSpecified = 0;
    for (CurrentEventNumber = 0; CurrentEventNumber < (*TotalNumberOfCounters); CurrentEventNumber++)
    {
      if (SkipArray[CurrentEventNumber] != 0)
	SkipSpecified = 1;
    }

    if (SkipSpecified)
    {
      (void) printf("Skip counts are not implemented for chosen method.\n");
      (void) printf("Hence they are being ignored.\n");
    }
  }

  if (Method & METH_SAMPLE)
  {
    if ((*IsDebug()))
    {
      (void) printf("\nCalling DoSampling:\n");

      (void) fflush(stdout);
    }

    DataMask.ipr_dmask_l_bits |= QuadDataMask.ipr_dmask_l_bits;

    Status = DoSampling(EventArray,                 /* events to sample */
			FrequencyArray,                /*  frequency of sampling   */
			SkipArray,                /*  skip value array    */
			ArraySize,              /*  #of events to sample   */
			ModeMask,               /*  mask of modes to sample in  */
			DataMask,               /*  data mask for collecting samples */
			QuadDataMask,           /*  data mask for collecting samples */
			StartTime,                  /*  start delay  */
			EndTime,                 /*  run time    */
			TotalNumberOfBuffers,           /*  #of buffs to allocate  */
			BufferSize,            /*  sample buffer size  */
			TotalNumberOfRanges,            /*  Number of pc ranges */
			&PCAddressList,          /*  pointer to addr range   */
			OutputFileName);               /*  Sample output file path */
  }
  else if (Method & METH_COUNT)
  {
    if ((*IsDebug()))
    {
      (void) printf("\nCalling DoCounting:\n");

      (void) fflush(stdout);
    }
    Status = DoCounting(EventArray,             /* events to sample */
			FrequencyArray,            /* frequency of sampling */
			ArraySize,          /* #of events to sample */
			ModeMask,           /* mask of modes to sample in */
			StartTime,              /* start delay */
			EndTime,             /* run time */
			Interval);           /* display interval */

  }
  else if (Method & METH_ALL)
  {
    if ((*IsDebug()))
    {
      (void) printf("\nCalling DoHistogram:\n");

      (void) fflush(stdout);
    }

    /* clear the histogram bits */

    HistogramData.ipr_hmask_l_bits = 0;

    /* now set the specified bits */

    if (Method&METH_TOTAL)
    {
      if ((*IsDebug()))

	(void) printf("Histogram type TOTAL specified\n");

      HistogramData.ipr_hmask_l_bits |= IPR_HMASK_M_TOTAL;

      HistogramCount++;
    }

    if (Method&METH_IPL)
    {
      if ((*IsDebug()))
#ifdef nt
	(void) printf("Histogram type IRQL specified\n");
#else
      (void) printf("Histogram type IPL specified\n");
#endif
      HistogramData.ipr_hmask_l_bits |= IPR_HMASK_M_IPL;

      HistogramCount++;
    }

    if (Method&METH_MODE)
    {
      if ((*IsDebug()))

	(void) printf("Histogram type MODE specified\n");

      HistogramData.ipr_hmask_l_bits |= IPR_HMASK_M_MODE;

      HistogramCount++;
    }

    if (Method&METH_PRIORITY)
    {
      if ((*IsDebug()))

	(void) printf("Histogram type PRIORITY specified\n");

      HistogramData.ipr_hmask_l_bits |= IPR_HMASK_M_PRIORITY;

      HistogramCount++;
    }

    if (Method&METH_ADDR)
    {
      if ((*IsDebug()))
      {
	(void) printf("Histogram type ADDRESS specified\n");
      }

      HistogramData.ipr_hmask_l_bits |= IPR_HMASK_M_ADDRESS;

      if (TotalNumberOfRanges)
      {
	for (PCRangePointer = PCAddressList.next, CurrentPCRange = 0;
	     PCRangePointer;
	     PCRangePointer = PCRangePointer->next, CurrentPCRange++)
	{
	  /*  Here if the number of elements requested is more than
	      MAX_ELEMENTS the granularity is adjusted so that the
	      number of elements is lesser than or equal to
	      MAX_ELEMENTS    */

	  ElementCount = ComputeHistogramElements(PCRangePointer);
	  if ((*IsDebug()))
	    (void) printf("Number of elements = %Lx\n", ElementCount);
	  while (ElementCount > MAX_ELEMENTS)
	  {
	    PCRangePointer->granularity++;
	    ElementCount = ComputeHistogramElements(PCRangePointer);
	    if ((*IsDebug()))
	      (void) printf("Number of elements = %Lx Granularity = %d\n",
			    ElementCount, PCRangePointer->granularity);

	    if(ElementCount > MAX_ELEMENTS)
	      continue;
	    else
	      (void) printf("Modified Range %d     : 0X"ADDRFORMAT"-0X"ADDRFORMAT".%d\n",
			    CurrentPCRange+1, PCRangePointer->start, PCRangePointer->end, PCRangePointer->granularity);
	  }
	}
	HistogramCount += TotalNumberOfRanges;
      }
      else
	HistogramCount++;
    }

    Status = DoHistogram(HistogramData,      /* Histogram data   */
			 EventArray,         /* events to sample */
			 FrequencyArray,        /* frequency of sampling */
			 ArraySize,      /* # of events to sample */
			 ModeMask,       /* mask of modes to sample in*/
			 StartTime,          /* start delay */
			 EndTime,         /* run time */
			 Interval,       /* Output time */
			 TotalNumberOfRanges,    /* Number of address ranges */
			 &PCAddressList,  /* Pointer to address ranges */
			 HistogramCount);    /* # of histograms entered */
    
  }
  else
  {
    (void) fprintf(stderr, "? *INVALID METHOD bits (0x%x) SEEN*\n", Method);
  }

  if (!IPR_SUCCESS(Status))
    (void) IprPrintMessage("Error ", Status, 0);
}









