/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 *  Program: iprobe_sampling.c
 *
 *  Abstract:  Module containing all the code to perform PC sampling, given the
 *             event and frequency arrays, buffer count and sampling file name.
 *
 * Modification History:
 *
 * Phillip Ezolt: 11-Jul-97
 *  Renamed variables and functions, and added comments. 
 *
 * Ravindra Kulkarni: Sep-94
 *    Added quaddatamask for create sample buffers.
 *
 * Mike Dempsey: Aug-94
 *    Added conditional compilation of ipr_os_vms.h and ipr_os_osf1.h
 *
 * Ravindra Kulkarni: Jun-94
 *    Modified the code to accept the pc sample datamask as input from user.
 *
 * Ravindra Kulkarni: Jun-94
 *    Modified the code to accept buffer size input.
 *
 * Ravindra Kulkarni: May-94
 *    Modified the code to accept the pc range input.
 *
 * Greg Tarsa: Mar-94
 *    Updated code for ctrmask argument to IprSpecifyOperations.
 *    Added Skip frequency to DoSampling.
 *    Changed code to properly handle finish_time a delta time.
 *
 * Ravindra Kulkarni: Feb-94
 *    Extensively modified to properly implement sampling.
 *
 * Greg Tarsa: Jan-94
 *    Wrote original code based on template supplied by Brian Wibecan.
 */
 
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#endif
#include <stdio.h>
#include "stdtypes.h"
#include "iprobe_struct.h"
#include "iprobe_hdr.h"
#include "iprobe_global.h"
#ifdef vms
#include "iprobe_vms.h"
#endif
#ifdef nt
#include "iprobe_nt.h"
#endif

#ifdef unix
#include "iprobe_osf1.h"
#endif
#include "iprobe_eventlib.h"


/* Globals.  Should be removed. */


/************************************************************************************/
/* Name: DoSampling                                                                */
/*                                                                                  */
/* Purpose: Setup the buffers and counters for a sampling run.                      */
/*                                                                                  */
/* Arguments: SelectedEvents: The array of events to measure.                       */
/*            SelectedFrequencies: Frequency to measure at.                         */
/*            SelectedSkips: Which interrupt to measure.                            */
/*            SelectionCount: Size of the prev 2 arrays.                            */
/*            ModesToUse: The program modes to measure in.                          */
/*            DataMask: Data mask to use for samples.                               */
/*            QuadDataMask: Data mask for quad samples.                             */
/*            StartTime: The number of secs until the start of measurement.         */
/*            Duration: The number of seconds to measure.                           */
/*            NumberOfBuffers: Number of buffers to use.                            */
/*            BufferSize: Sample buffer size.                                       */
/*            NumberOfPCRanges: Number of pc ranges.                                */
/*            PCRange: Pointer to a linked list of PC ranges.                       */
/*            SamplingFileName: Sampling file name.                                 */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: This returns IPR_Success if the function was succesful, or the error      */
/*        code of the failure.                                                      */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS DoSampling(IPR_EVENT *SelectedEvents,     /*  array of evts to measure */
		      IPR_FREQ *SelectedFrequencies, /*  freqs to measure at */
		      IPR_COUNT *SelectedSkips,      /*  which interrupt to measure */
		      IPR_COUNT SelectionCount,      /*  size of the prev 2 arrays */
		      IPR_MMASK ModesToUse,          /*  prog modes to measure in */
		      IPR_DMASK DataMask,            /*  data mask for samples */
		      IPR_DMASK QuadDataMask,        /*  data mask for quad samples */
		      time_t StartTime,              /*  #secs till start of meas */
		      time_t Duration,               /*  #secs to measure */
		      IPR_COUNT NumberOfBuffers,     /*  number of buffers to use */
		      IPR_COUNT BufferSize,          /*  sample buffer size  */
		      IPR_COUNT NumberOfPCRanges,    /*  Number of pc ranges */
		      struct pc_ranges *PCRange,     /*  Pointer to addr range */
		      char *SamplingFileName)        /*  sampling file name */
{

  /*  These three symbols are declared volatile because their contents
      are changed asynchronously with respect to the code flow and we don't
      want the compiler to optimize code using them away because it thinks
      the values do not change.  */

  FLAG CntrlCFlag = 0;               /* flag for CTRL-C hit by user */
  FLAG BufferFullFlag = 0;           /* flag for buffer full condition */
  FLAG *StartedProcessExited;
  FLAG TimeOutFlag = 0;              /* flag for end time expiration */
  

  int32 FullBufferWrites = 0;
  int32 PartialBufferWrites = 0;

  int XParts = 0;                    /* count of part buffers written */ /* This needs to be an int for VMS. */
  int XWrites = 0;                   /* count of full buffers written */ /* This needs to be an int for VMS. */

  /*  General usage symbols */
  IPR_CTRMASK RequestedCounterMask;  /* mask for which counters are requested */
  IPR_CTRMASK CounterMask;           /* mask for which counters are in use */

  IPR_CTRNUMBER *WhichEvent;         /* contains actual events  */
  IPR_FREQ *WhichFrequency;          /* contains actual frequencies */

  IPR_SIZE WhichFrequencySize;
  IPR_SIZE WhichEventSize;

  IPR_STATUS Status;                /* general status return */

  struct pc_ranges *TempPCRange;    /* temporary varaible  */
  system_info *SystemInformation;

  time_t AbsoulteEndTime;

  uint64 *RangeArray;               /* linked list of pc ranges is converted into an array of pc values. */
  uint32 CurrentRange;
  uint32 CurrentCounter;
  uint32 CurrentEvent;
  uint32 CurrentSelection;
  uint32 CurrentFrequency;
  uint32 *TotalNumberOfCounters;

  TotalNumberOfCounters = GetTotalNumberOfCounters();


  WhichFrequencySize = (*TotalNumberOfCounters);
  WhichEventSize = (*TotalNumberOfCounters);



  SystemInformation = GetSystemInformation();
  StartedProcessExited = GetStartedProcessExited();


  WhichEvent = (IPR_CTRNUMBER *)ckalloc(sizeof(IPR_CTRNUMBER) * (*TotalNumberOfCounters));
  WhichFrequency = (IPR_FREQ *)ckalloc(sizeof(IPR_FREQ) * (*TotalNumberOfCounters));



  /*  Save the data mask, range count in global data structure */
  SystemInformation->datamask = DataMask;
  SystemInformation->rangecnt = NumberOfPCRanges;


  if((*IsDebug()))
  {
    (void) printf("datamask = %x\n", DataMask.ipr_dmask_l_bits);
  }

  RangeArray = (uint64 *)ckalloc(sizeof(uint64) * NumberOfPCRanges * 2);

  TempPCRange = PCRange->next;

  for(CurrentRange = 0; CurrentRange < NumberOfPCRanges; CurrentRange++)
  {
    RangeArray[CurrentRange*2] = TempPCRange->start;
    RangeArray[CurrentRange*2+1] = TempPCRange->end;
    TempPCRange = TempPCRange->next;
  }

  SystemInformation->rangeptr = RangeArray;

  if((*IsDebug()) && NumberOfPCRanges)
  {
    (void) printf("Display of RangeArray\n");
    (void) printf("range count = %d\n", NumberOfPCRanges);

    for(CurrentRange = 0; CurrentRange < NumberOfPCRanges * 2; CurrentRange += 2)
    {
      (void) printf(ADDRFORMAT" "ADDRFORMAT "\n", RangeArray[CurrentRange], RangeArray[CurrentRange+1]);
    }
  }

  /* Allocate the counters to bar other processes' use */
  Status = IprAllocate();

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprAllocate", Status, NULL);
    return Status;
  }

  if((*IsDebug()))
  {
    (void) printf("datamask %x QuadDataMask %x\n", DataMask, QuadDataMask);
  }

  /* Create the counter sample buffers */

#if defined vms || defined unix
# define IPROBE_FULL_ACTION IPR_FULL_WAKE
#else
# define IPROBE_FULL_ACTION IPR_FULL_WRITE
#endif
  Status = IprCreateSampleBuffers(IPR_ALL_CPUS,                  /* cpu mask */
				  NumberOfBuffers,               /* # buffers/cpu */
				  BufferSize,                    /* buf size */
				  DataMask,                      /* data mask */
				  QuadDataMask,                  /* quad data mask */
				  0,                             /* no sample action */
				  IPROBE_FULL_ACTION,
				  0,                             /* Apc routine */
				  NumberOfPCRanges,              /* # of PC ranges */
				  RangeArray);                   /* PC range vector */

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprCreateSampleBuffers", Status, NULL);
    return Status;
  }

  /* Specify the Events to be recorded */

  Status = IprSpecifyOperations(IPR_ALL_CPUS,                     /* cpu mask */
				SelectedEvents,                   /* event array */
				SelectedFrequencies,              /* freq array */
				SelectionCount,                   /* sz of both arrays */
				ModesToUse,                       /* mode selection */
				0,                                /* no MUX return */
				&RequestedCounterMask);           /* which counters */

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprSpecifyOperations", Status, NULL);
    return Status;
  }

  /*  SystemInformation is a structure that holds all the specific information about
      the current system.  We collect it as it is determined. */

  SystemInformation->counter_mask = RequestedCounterMask ;

  /* Get a handle on which counters are in use */
  Status = IprEventCounterMap(IPR_CURRENT_CPU, WhichEvent, WhichEventSize);

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprEventCounterMap", Status, NULL);
    return Status;
  }

  /*  Save actual events in header structure. */
  SystemInformation->act_events =(IPR_EVENT *)malloc(sizeof(IPR_EVENT) * WhichEventSize);

  for(CurrentEvent = 0; CurrentEvent < WhichEventSize; CurrentEvent++)
  {
    SystemInformation->act_events[CurrentEvent] = WhichEvent[CurrentEvent];
  }
  Status = IprFreqCounterMap(IPR_CURRENT_CPU,                    /* CPU to check */
			     WhichFrequency,                     /* return vector of freqs */
			     WhichFrequencySize);               /* vector length */

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprFreqCounterMap", Status, NULL);
    return Status;
  }

  /*  Save actual frequencies in header structure. */
  SystemInformation->act_freqs =(IPR_FREQ *)malloc(sizeof(IPR_FREQ) * WhichFrequencySize);

  for(CurrentFrequency = 0; CurrentFrequency < WhichFrequencySize; CurrentFrequency++)
  {
    SystemInformation->act_freqs[CurrentFrequency] = WhichFrequency[CurrentFrequency];
  }

  /* set the active counter(s) */
  /* RequestedCounterMask gets put into sample header */

  Status = IprSetActiveCounters(IPR_ALL_CPUS, RequestedCounterMask, &CounterMask);

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprSetActiveCounters", Status, NULL);
    return Status;
  }

  /*  Set the sample skip rate, that is, record only the Nth interrupt
      rather than every interrupt. */

  SystemInformation->act_skips =(IPR_COUNT *) ckalloc(sizeof(IPR_COUNT) * (*TotalNumberOfCounters));

  for(CurrentCounter = 0; CurrentCounter < (*TotalNumberOfCounters); CurrentCounter++)
  {
    SystemInformation->act_skips[CurrentCounter] = 0;
  }

  for(CurrentSelection = 0; CurrentSelection < SelectionCount; CurrentSelection++)
  {
    /* Note, this doesn't correct account for duplicated events.  Sigh... */

    for(CurrentEvent = 0; WhichEvent[CurrentEvent] != SelectedEvents[CurrentSelection]; CurrentEvent++);
    SystemInformation->act_skips[CurrentEvent] = SelectedSkips[CurrentSelection];
  }

  Status = IprSetSampleSkip(IPR_ALL_CPUS,                   /*  cpu mask */
			    SystemInformation->act_skips,             /*  skip values */
			    (*TotalNumberOfCounters));                 /*  Number of skip values */

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprSetSampleSkip", Status, NULL);
    return Status;
  }

  if((*IsDebug()))
  {
    (void) printf("Requested counter mask is: 0%o\n", RequestedCounterMask);
    (void) printf("Active counter mask is: 0%o\n", CounterMask);
  }

  /* open the sample file(no default filespec at this time) */

  /*  sample file path */
  /*  default fspec */
  Status = IprOpenSampleFile(SamplingFileName,"");                             

  if(!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprOpenSampleFile", Status, NULL);
    return Status;
  }

  /* If specified in cmd line, wait for start time */

  if(StartTime)
  {
    if((*IsDebug()))
    {
      (void) printf("Waiting for start time(%d seconds since program start)\n", StartTime-time(0));
    }

    (void) printf("Waiting for %d seconds\n", StartTime - time(0));

    WaitForTime(StartTime);
  }

  /* If there is an Duration specified, then schedule a WakeUp call */

  if(Duration)
    {
      AbsoulteEndTime = time(0) + Duration;
      
      if((*IsDebug()))
	{
	  (void) printf("Establishing end timer(%d seconds)\n", Duration);
	  (void) printf("TimeOutFlag = %d\n", TimeOutFlag);
	}

    /* Establish a handler for the timer signal */

    Status = OSDeclareWakeUpHandler(&TimeOutFlag);

    if(!IPR_SUCCESS(Status))
      return Status;

    if((*IsDebug()))
      {
      (void) printf("After OSDeclareWakeUpHandler TimeOutFlag=%d\n", TimeOutFlag);

      }
    /* Schedule a WakeUp signal at the ending time */

    Status = OSScheduleWakeUpTime(AbsoulteEndTime);

    if(!IPR_SUCCESS(Status))
      return Status;
  }

  if((*IsDebug()))
  {
    (void) printf("After OSScheduleWakeUpTime TimeOutFlag=%d\n",
		  TimeOutFlag);
  }

  /* Establish a signal handler for user-requested abort */

  Status = OSDeclareCntrlCHandler(&CntrlCFlag);

  if(!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /* Establish a signal handler for the Ipr interrupt */

  Status = OSDeclarePFMHandler(&BufferFullFlag);

  if (!IPR_SUCCESS(Status))
  {
    return Status;
  }

  /*  At this time the sampling will start. So this time is recorded as
      start time for sampling.  */

  (void) UpdateTime();

  /*  Write the sample file header    */

  (void) WriteHeaderInfo();

  (void) printf("Start of sampling\n");


  /* Issue the command if one was given */

  Status = IssueCommandLine(SystemInformation->CommandLine);

  if(IPR_SUCCESS(Status))
  {
    Status = IprStart(IPR_ALL_CPUS, CounterMask); /*Start counting on all cpus*/
  }
	
  /*  Wait for signals indicating "buffer full", "time out", or "user-intr" */

  if((*IsDebug()))
  {
    (void) printf("IPR_SUCCESS(Status)=%d TimeOutFlag=%d CntrlCFlag=%d\n", IPR_SUCCESS(Status),TimeOutFlag,CntrlCFlag);
  }

  while(IPR_SUCCESS(Status) && !*StartedProcessExited && !TimeOutFlag && !CntrlCFlag)
  {
    if((*IsDebug()))
    {
      (void) printf("Waiting for signal. . .");
      (void) fflush(stdout);
    }

    (void) OSWaitForSignal();   /* wait for a signal */

    if((*IsDebug()))
    {
      if(TimeOutFlag)
      {
	(void) printf("TIMER signal received\n");
      }
      else if(CntrlCFlag)
      {             
	(void) printf("CTRL-C signal received\n");
      }
      else
	(void) printf("BUFFER FULL signal received\n");

      (void) fflush(stdout);
    }

    /*  When *any* signal occurs, write the sample buffers Loop will properly end upon timeout or user interrupt
	upon iteration. */

    Status = IprWriteSampleBuffers(IPR_ALL_CPUS,IPR_NOWRITE_PARTIAL, &XWrites, &XParts);

    if(!IPR_SUCCESS(Status))
    {
      (void) IprPrintMessage("IprWriteSampleBuffers", Status, NULL);
      return Status;
    }

    FullBufferWrites += XWrites;
    PartialBufferWrites += XParts;
  }

  /* Done: perform sampling cleanup/rundown */

  Status = IprTerminateSampling(&XWrites, &XParts);

  if(!IPR_SUCCESS(Status))
  {
    /*  Since we are in an error state, we will ignore the deallocation return Status */

    (void) IprPrintMessage("IprTerminateSampling", Status, NULL);
    (void) IprDeallocate();

    return Status;
  }

  /* Release the counters to other programs */

  Status = IprDeallocate();

  if (!IPR_SUCCESS(Status))
  {
    (void) IprPrintMessage("IprDeallocate", Status, NULL);
    return Status;
  }

  /* Restore all signal handling to previous state */

  OSCancelPFMHandler();

  (void) OSCancelCntrlCHandler();

  if (Duration)
  {
    OSCancelWakeUpHandler();
  }
  /* perform final count increments */

  FullBufferWrites += XWrites;
  PartialBufferWrites += XParts;

  (void) printf("End of sampling\n");

  (void) printf("Buffers written : %d\n", FullBufferWrites);

  (void) printf("Partial buffers written: %d\n", PartialBufferWrites);

  return IPR__SUCCESS;
}



