#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include "stdtypes.h"

#include "iprobe_struct.h"

#include "iprobe_hdr.h"
#include "iprobe_osf1.h"
#include "iprobe_global.h"

#define MAX_NUM_OF_ARGUMENTS 20


time_t completion_time;

volatile uint32 *pTimerFlag=NULL;
volatile uint32 *pBufferFullFlag=NULL;
volatile uint32 *InterruptFlag=NULL;

void (*OldAlarmHandler)();
void (*OldInterruptHandler)();
void (*OldUser1Handler)();
void (*OldChildHandler)();

/************************************************************************************/
/* Name: SignalHandler                                                              */
/*                                                                                  */
/* Purpose: This is called everytime a signal that we are handling occurs.          */
/*          The handler sets various global flags depending on which signal occured.*/
/*                                                                                  */
/* Arguments: SignalNumber: The number of the signal that we are handling.          */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

static void SignalHandler(int SignalNumber)
{

  FLAG *StartedProcessExited;
  time_t PresentTime;
  
  StartedProcessExited = GetStartedProcessExited();

  /* Increment the appropriate flag, if there is an address */
  switch (SignalNumber)
    {
    case SIGUSR1: 
      if (pBufferFullFlag!=NULL)
	{	
	  (*pBufferFullFlag)++; 
	}     
      break;
    case SIGINT:  
      if (InterruptFlag!=NULL)
	{
	  (*InterruptFlag)++;
	}
      break;
    case SIGALRM:
      {
	/* THis is need because VMS is stupid, and it doesn't remeber which signal goes to
	   which handler. */

	signal(SIGALRM, SignalHandler);
	
	/*
	 *  Only set the timeout flag if this ALRM signal
	 *  occurred and the completion time is met or exceeded.
	 *  This allows both a scheduled wait time and an delta interval
	 *  time to work.
	 */
	if (pTimerFlag!=NULL)
	  {
	    PresentTime = time(0);
	    if (PresentTime >= completion_time)
	      {
		(*pTimerFlag)++;
	      }
	  }
	break;
      }
    case SIGCHLD:
     
      (*StartedProcessExited) = 1;
 
      /* This uninstalls the child death signal handler. */
      OldChildHandler = signal(SIGCHLD, OldChildHandler);

    default:
      break;
    }
}/* End SignalHandler() */

/************************************************************************************/
/* Name: OSDeclareWakeUpHandler                                                     */
/*                                                                                  */
/* Purpose: This installs the handler that will be called when an alarm occurs.     */
/*                                                                                  */
/* Arguments: pTempTimerFlag: A pointer to a flag the is set to one if the timer    */
/*            interrupt has occured.                                                */
/*                                                                                  */
/* Output: pTimerFlag: This global variable is set to point to the variable in      */
/*                     other functions relying on it.                               */
/*                                                                                  */
/* Function Value: 1 regardless                                                     */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
IPR_STATUS OSDeclareWakeUpHandler(FLAG *pTempTimerFlag)
{
    if (pTempTimerFlag)
      {
	pTimerFlag = pTempTimerFlag;
      }

    OldAlarmHandler = signal(SIGALRM, SignalHandler);
    return 1;
}

/************************************************************************************/
/* Name: OSCancelWakeUpHandler                                                      */
/*                                                                                  */
/* Purpose: This will stop the periodic timer.                                      */
/*                                                                                  */
/* Arguments: None                                                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
void OSCancelWakeUpHandler(void)
{
    if (pTimerFlag!=NULL)
      {
	pTimerFlag = NULL;
      }

    OldAlarmHandler = signal(SIGALRM, OldAlarmHandler);
}

/************************************************************************************/
/* Name: OSDeclareCntrlCHandler                                                     */
/*                                                                                  */
/* Purpose: This sets up a handler to be called if the user presses cntrl-c.        */
/*                                                                                  */
/* Arguments: pInterruptFlag: A pointer to a flag in the collection functions, that */
/*                            is set when CntrlC has been pressed.                  */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: Returns 1 regardless.                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS OSDeclareCntrlCHandler(FLAG *pInterruptFlag)
{
    if (pInterruptFlag)
      {
	InterruptFlag = pInterruptFlag;
      }
    OldInterruptHandler = signal(SIGINT, SignalHandler);
    return 1;
}

/************************************************************************************/
/* Name: OSCancelCntrlCHandler                                                      */
/*                                                                                  */
/* Purpose: This will remove the handler for control-C.                             */
/*                                                                                  */
/* Arguments: pInterruptFlag: A pointer to a flag in the collection functions, that */
/*                            is set when CntrlC has been pressed.                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
void OSCancelCntrlCHandler(FLAG *pInterruptFlag)
{
    if (InterruptFlag!=NULL)
      {
	InterruptFlag = NULL;
      }
    OldInterruptHandler = signal(SIGINT, OldInterruptHandler);
}

/************************************************************************************/
/* Name: OSDeclarePFMHandler                                                        */
/*                                                                                  */
/* Purpose: This sets up the signal handler necessary for when a buffer overflows.  */
/*                                                                                  */
/* Arguments: pTempBufferFullFlag: This is a pointer to a variable that gets set to */
/*                                 1 if a buffer is filled.                         */
/*                                                                                  */
/* Output: pBufferFullFlag: A global pointer to the flag.                           */
/*                                                                                  */
/* Function Value: Always returns 1.                                                */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
IPR_STATUS OSDeclarePFMHandler(FLAG *pTempBufferFullFlag)
{
    if (pTempBufferFullFlag)
      {
	pBufferFullFlag = pTempBufferFullFlag;
      }
    OldUser1Handler = signal(SIGUSR1, SignalHandler);
    return 1;
}

/************************************************************************************/
/* Name: OSCancelPFMHandler                                                         */
/*                                                                                  */
/* Purpose:  This removes the handler that gets signals from a sample buffer        */
/*           overflow.                                                              */
/*                                                                                  */
/* Arguments: None                                                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
void OSCancelPFMHandler(void)
{
    if (pBufferFullFlag!=NULL)
      {
	pBufferFullFlag = NULL;
      }
    OldUser1Handler = signal(SIGUSR1, OldUser1Handler);
}

/************************************************************************************/
/* Name: OSScheduleWakeUpTime                                                       */
/*                                                                                  */
/* Purpose: This routine sets a timer which will expire at the given time.          */
/*                                                                                  */
/* Arguments: TimeOutTime: The amount of time in seconds to wait before waking up.  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: 1 regardless.                                                    */
/*                                                                                  */
/* Notes: This routine should cause the SIGALRM signal to be generated when the     */
/*        timer runs down.  The WakeUp handler should have been established by      */
/*        OSDeclareWakeUpHandler and be looking for this event.  It will set the    */
/*	  timeout flag when both a SIGALRM occurs and the current time meets or     */
/*	  exceeds the end_time.  This double check allows the ALRM signal to be used*/
/*	  for repeating delta times as well as time outs.                           */
/*                                                                                  */
/*	  It is assumed that the caller has called OS_declare_WakeUp_handler().     */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS OSScheduleWakeUpTime(time_t TimeOutTime)
{
    IPR_STATUS status;		/* general status return */
    time_t PresentTime = time(0);	/* current time */
    int64 TimeDifference;
    
    struct itimerval NewExpiration;

    /* Is there any time left before we need to stop? */
    TimeDifference = TimeOutTime - PresentTime;
    if (TimeDifference <= 0)		/* No: exit now, successfully */
      {
	(*pTimerFlag)++;	/* set timeout immediately */
	return 1;
      }

    /* init interval times to 0 */
    memset(&NewExpiration, 0, sizeof (struct itimerval));

    /* Set the timer value */
    NewExpiration.it_value.tv_sec = TimeDifference;

    /* Set the completion time */
    completion_time = TimeOutTime;

    /* Establish the interval for a one-shot interrupt */
    status = setitimer(ITIMER_REAL, &NewExpiration, NULL);
    if (status < 0)
      {
	perror("schedule_WakeUp_time: failure to set interval timer");
	return 0;
      }    

    return 1;
}/* end OSScheduleWakeUpTime */


/************************************************************************************/
/* Name: OSWaitForSignal                                                            */
/*                                                                                  */
/* Purpose: This just sits and waits for a signal to occur.                         */
/*                                                                                  */
/* Arguments: None                                                                  */
/*                                                                                  */
/* Output: None.                                                                    */
/*                                                                                  */
/* Function Value: None.                                                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
void OSWaitForSignal(void)
{
    /* wait for a handled signal */
    (void) sigpause(sigmask(0));		/* always returns -1 */

}/* end OSWaitForSignal() */


/************************************************************************************/
/* Name: OSScheduleRepeatingWakeUp                                                  */
/*                                                                                  */
/* Purpose: This schedules an event the will periodically wake up and call          */
/*          TimerHandler.                                                           */
/*                                                                                  */
/* Arguments: DeltaTime: the amount of seconds to wait between TimerHandler calls   */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: Returns 1 if successful, 0 otherwise.                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS OSScheduleRepeatingWakeUp(time_t DeltaTime)
{
    IPR_STATUS Status;		/* general status return */
    
    struct itimerval NewExpiration;

    /* if delta is negative, make it positive */
    if (DeltaTime < 0)
      {
	DeltaTime = -DeltaTime;
      }

    /* init interval times to 0 */
    memset(&NewExpiration, 0, sizeof (struct itimerval));

    /* Set the timer and the interval */
    NewExpiration.it_value.tv_sec = DeltaTime;
    NewExpiration.it_interval.tv_sec = DeltaTime;

    /* Establish the interval for a repeating interrupt */
    Status = setitimer(ITIMER_REAL, &NewExpiration, NULL);
    if (Status < 0)
      {
	perror("schedule_repeating_WakeUp: failure to set interval timer");
	return 0;
      }    

    return 1;
}/* end OSScheduleRepeatingWakeUp() */

/************************************************************************************/
/* Name: OSClearRepeatingWakeUp                                                     */
/*                                                                                  */
/* Purpose: This will stop the periodic WakeUp timer.                               */
/*                                                                                  */
/* Arguments: None                                                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: Returns 1 if successful, 0 otherwise.                            */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS OSClearRepeatingWakeUp(void)
{
    IPR_STATUS Status;		/* general status return */
    
    struct itimerval NewExpiration;

    /* init interval times to 0 */
    memset(&NewExpiration, 0, sizeof (struct itimerval));

    /* Clear any exisiting timer values */
    Status = setitimer(ITIMER_REAL, &NewExpiration, NULL);
    if (Status < 0)
      {
	perror("clear_repeating_WakeUp: failure to set interval timer");
	return 0;
      }    

    return 1;
}/* end OS_clear_repeating_WakeUp() */


/************************************************************************************/
/* Name: StopIssuedCommand							    */
/*                                                                                  */
/* Purpose: Stop the command issued by the IssueCommandLine function                */
/*                                                                                  */
/* Arguments: none                                                                  */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: none								    */
/*                                                                                  */
/* Notes: None			     				        	    */
/*                                                                                  */
/************************************************************************************/
void StopIssuedCommand(void)
{
	if ((*GetStartedProcessExited()) != 0)
	{
		kill((*GetStartedCommandPidPointer()), SIGKILL);
	}

	return;
}

/************************************************************************************/
/* Name: IssueCommandLine                                                           */
/*                                                                                  */
/* Purpose: This will spawn of the requested command.                               */
/*                                                                                  */
/* Arguments: CommandLine: the command that is to be run.                           */
/*                                                                                  */
/* Output: None                                                                     */
/*                                                                                  */
/* Function Value: IPR__SUCCESS is returned if the command was properly started.    */
/*                 IPR__FAILURE is returned if the command was not properly started.*/
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/************************************************************************************/
IPR_STATUS IssueCommandLine(char *CommandLine)
{
  char *ArgumentVector[MAX_NUM_OF_ARGUMENTS];
  FLAG *StartedProcessExited;
  int32 ChildPid;

  StartedProcessExited = GetStartedProcessExited();

  /* If we have no command line, return */
  if (CommandLine[0]=='\0')
    {
      return IPR__SUCCESS;
    }

  /* This sets up the Handler so that the parent will be called when the child dies. */

  OldChildHandler = signal(SIGCHLD, SignalHandler);

  /* Given a string with all of the command line options, this will create the argument vector. */
  CreateArgv(CommandLine, ArgumentVector);

	ChildPid=vfork();

	if (ChildPid == 0)
	{
		//
		// I am the child -- run the given command.
		//
		execv(ArgumentVector[0], ArgumentVector);
		printf("Attempted to run: %s\n", ArgumentVector[0]);
		perror("Exec Error");
		exit(1);
	}
	else 
	{
		if (ChildPid != -1)
		{
			*(GetStartedCommandPidPointer()) = ChildPid;
		}
		else
		{
			printf("IssueCommandLine: There has been a forking error.\n");
			(*StartedProcessExited) = 1;
			return IPR__FAILURE;
		}
  }

	return IPR__SUCCESS;
}


/************************************************************************************/
/* Name: CreateArgv                                                                 */
/*                                                                                  */
/* Purpose: This will take a string and break it up into a vector of pointers to    */
/*           character strings.  This is needed to turn "netscape blah balh" into   */
/*           a vector pointing to "netscape" "blah" and "balh".                     */
/*                                                                                  */
/*                                                                                  */
/* Arguments: CommandLine: A string with the complete command that is to be run.    */
/*            ArgumentVector: This is an array of pointers that point to the        */
/*                            various elements of the command line.                 */
/*                                                                                  */
/* Output: ArgumentVector: The vector that contains pointers to all of the command  */
/*                         line arguments.                                          */
/*                                                                                  */
/* Function Value: IPR__SUCCESS is returned if the command was properly started.    */
/*                 IPR__FAILURE is returned if the command was not properly started.*/
/*                                                                                  */
/* Notes: This allocates memory for the strings as needed.                          */
/*                                                                                  */
/************************************************************************************/

IPR_STATUS CreateArgv(char *CommandLine, char *ArgumentVector[])
{
  char *CurrentToken;
  char *TempCommandLine;
  int32 CurrentArgumentNumber=1;

  TempCommandLine=(char *)malloc(strlen(CommandLine));
  strcpy(TempCommandLine,CommandLine);


  /* This sets up strtok so that it knows that we are talking about "TempCommand" */

  CurrentToken=strtok(TempCommandLine," ");
  

  /* Assume we have at least one argument, the command to be run. */

  ArgumentVector[0]=(char *)malloc(strlen(TempCommandLine));
  strcpy(ArgumentVector[0],TempCommandLine);
  
  while ((CurrentToken=strtok(NULL," "))!=NULL)
    {
      ArgumentVector[CurrentArgumentNumber]=(char *)malloc(strlen(CurrentToken));
      strcpy(ArgumentVector[CurrentArgumentNumber], CurrentToken);

      (CurrentArgumentNumber)++;
    }
  /* This guarentees that the last pointer is a NULL pointer. */
  ArgumentVector[CurrentArgumentNumber]=NULL;

  return IPR__SUCCESS;
}
