/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/

/*
 * Abstract:	This file contains performance counter support routines that
 *		have no chip or operating system dependencies.  Some of these
 *		routines are for direct use by calling code; others are for
 *		the support of the chip-specific routines.
 *
 * Author:		Brian F. Wibecan
 * Creation Date:	16-Dec-1993
 *
 * Modified by:
 *
 * X-20	BFW		Brian F. Wibecan	26-Sep-1994
 *	Made modifications from code review.
 *
 * X-19	BMW		Bradley M. Waters	18-Jul-1994
 *	Modified based on kernel rationalization.
 *
 * X-18	BMW1101		Bradley M. Waters	15-Jul-1994
 *	Added api_routines_nt.h include to define api routines to ctr...
 *
 * X-17	BFW1100		Brian F. Wibecan	19-May-1994
 *	Fix type of xxxEventToNumber.
 *
 * X-16	BFW1099		Brian F. Wibecan	18-May-1994
 *	Added support for xxxSelectEventsFromArray.
 *
 * X-15	BFW1080		Brian F. Wibecan	04-May-1994
 *	Modified #pragma module to be compatible with CC on OSF/1.
 *
 * X-14	BFW1075		Brian F. Wibecan	21-Apr-1994
 *	Made event set handling routines deal with invalid event numbers.
 *
 * X-13	BFW1073		Brian F. Wibecan	06-Apr-1994
 *	Added #pragma module statement.
 *
 * X-12	BFW1063		Brian F. Wibecan	21-Mar-1994
 *	Move initialization to the OS-specific code.
 *
 * X-11	BFW1061		Brian F. Wibecan	18-Mar-1994
 *	Switch two lines in available_freqs to get around compiler bug on V1.5.
 *
 * X-10	BFW1056		Brian F. Wibecan	14-Mar-1994
 *	Fix parallel initialization bug.
 *
 * X-9	BFW1040		Brian F. Wibecan	11-Mar-1994
 *	Add FreqCounterMap.
 *
 * X-8	BFW1033		Brian F. Wibecan	04-Mar-1994
 *	Continued adding EV5 support.  Removed assumptions that an event can
 *	be counted on exactly one counter.
 *
 * X-7	BFW1028a	Brian F. Wibecan	18-Feb-1994
 *	Fixed bug in EventToNumber to return zero if the event
 *	is not in the set.
 *
 * X-6	BFW1028		Brian F. Wibecan	16-Feb-1994
 *	Added cross-checking in event assignment, to handle cases where
 *	more than one counter is capable of counting a particular event.
 *	This functionality is needed by EV5.  Also changed CtrEventToNumber
 *	so it will return the lowest event number set, to allow easy
 *	looping through event sets.  Finally, added CtrDeleteEvent to
 *	delete a single event from a set, for the same reason.
 *
 * X-5	BFW1023		Brian F. Wibecan	03-Feb-1994
 *	Put double quotes around ints.h.
 *
 * X-4	BFW1022		Brian F. Wibecan	02-Feb-1994
 *	Removed spurious reference to hwrpbdef.
 *
 * X-3	BFW1020		Brian F. Wibecan	02-Feb-1994
 *	Renamed to CTR_ROUTINES.C.
 *
 * X-2	BFW1009		Brian F. Wibecan	07-Dec-1994
 *	Added check for chip type.
 *
 * X-1	BFW1000		Brian F. Wibecan	16-Dec-1993
 *	Original version.
 *
*/

#ifdef __NT
#  include "ipr_shr_nt.h"       // includes windows.h for basic type defs for Windows NT
#  include "api_routines_nt.h"
#endif
#include "ctr_int_defs.h"
#include "bit_library.h"

/*****************************************************************************/
/* Utility routines for internal use                                         */
/*****************************************************************************/


/*****************************************************************************/
/* NULL ROUTINES                                                             */
/*                                                                           */
/* All the chip-dependent routines are referenced through pointers.  The     */
/* pointers point to these null routines initially.  They return unavailable */
/* status or zero value, depending on the semantics of the caller.  At load  */
/* time, the pointers for all chip-dependent routines that are actually      */
/* implemented will be moved to point to the real routines, and all those    */
/* that are not implemented will stay pointing to these null routines.       */
/*****************************************************************************/

IPR_STATUS NullUnavailableRoutine (void)
{
   return IPR__UNAVAILABLE;
}

IPR_COUNT NullZeroRoutine (void)
{
   return 0;
}


/*****************************************************************************/
/* Function pointers used to direct calls to the chip-specific routines      */
/*****************************************************************************/

STATUSFUNCPTR
   PtrAvailableEvents = &NullUnavailableRoutine,
   PtrAvailableFreqs = &NullUnavailableRoutine,
   PtrAvailableModes = &NullUnavailableRoutine,
   PtrAvailableFuncs = &NullUnavailableRoutine,
   PtrCounterNumber = &NullUnavailableRoutine,
   PtrEventCounterMap = &NullUnavailableRoutine,
   PtrFreqCounterMap = &NullUnavailableRoutine,
   PtrCurrentEvents = &NullUnavailableRoutine,
   PtrCurrentFreqs = &NullUnavailableRoutine,
   PtrCurrentModes = &NullUnavailableRoutine,
   PtrCurrentState = &NullUnavailableRoutine,
   PtrClearContext = &NullUnavailableRoutine,
   PtrSelectEvents = &NullUnavailableRoutine,
   PtrSelectEventsFromArray = &NullUnavailableRoutine,
   PtrSetEvents = &NullUnavailableRoutine,
   PtrSetModes = &NullUnavailableRoutine,
   PtrEnable = &NullUnavailableRoutine,
   PtrDisable = &NullUnavailableRoutine,
   PtrPause = &NullUnavailableRoutine,
   PtrResume = &NullUnavailableRoutine,
   PtrClear = &NullUnavailableRoutine,
   PtrWrite = &NullUnavailableRoutine,
   PtrRead = &NullUnavailableRoutine,
   PtrReadAll = &NullUnavailableRoutine,
   PtrAvailableCounters = &NullUnavailableRoutine,
   PtrCounterEvents = &NullUnavailableRoutine;
COUNTFUNCPTR
   PtrNumCounters = &NullZeroRoutine;


/*****************************************************************************/
/* Shell functions that call the actual functions via the pointers           */
/*****************************************************************************/

IPR_STATUS CtrAvailableEvents (IN const IPR_EVENTSET *initial_events,
                             OUT IPR_EVENTSET *more_events)
{
   return PtrAvailableEvents (initial_events, more_events);
}

IPR_STATUS CtrAvailableFreqs (IN const IPR_EVENT event, OUT IPR_FREQ *freqs)
{
   return PtrAvailableFreqs (event, freqs);
}

IPR_STATUS CtrAvailableModes (OUT IPR_MMASK *modes)
{
   return PtrAvailableModes (modes);
}

IPR_STATUS CtrAvailableFuncs (OUT IPR_FMASK *funcs)
{
   return PtrAvailableFuncs (funcs);
}

IPR_COUNT CtrNumCounters (void)
{
   return PtrNumCounters ();
}

IPR_STATUS CtrCounterNumber (IN int ctx, IN const IPR_EVENT event,
                                  OUT IPR_CTRNUMBER *counter_number)
{
   return PtrCounterNumber (ctx, event, counter_number);
}

IPR_STATUS CtrEventCounterMap (IN int ctx, OUT IPR_EVENT *event_array,
                                    IN IPR_SIZE asize)
{
   return PtrEventCounterMap (ctx, event_array, asize);
}

IPR_STATUS CtrFreqCounterMap (IN int ctx, IN IPR_FREQ *freq_array,
                                    IN IPR_SIZE asize)
{
   return PtrFreqCounterMap (ctx, freq_array, asize);
}

IPR_STATUS CtrCurrentEvents (IN int ctx, OUT IPR_EVENTSET *events)
{
   return PtrCurrentEvents (ctx, events);
}

IPR_STATUS CtrCurrentFreqs (IN int ctx, IN const IPR_EVENT event, OUT IPR_FREQ *freq)
{
   return PtrCurrentFreqs (ctx, event, freq);
}

IPR_STATUS CtrCurrentModes (IN int ctx, OUT IPR_MMASK *modes)
{
   return PtrCurrentModes (ctx, modes);
}

IPR_STATUS CtrCurrentState (IN int ctx, IN const IPR_EVENT event)
{
   return PtrCurrentState (ctx, event);
}

IPR_STATUS CtrClearContext (IN int ctx)
{
   return PtrClearContext (ctx);
}

IPR_STATUS CtrSelectEvents (IN int ctx, IN const IPR_EVENTSET *events, IN IPR_FREQ freq,
                            OUT uint64 *MUXvalue)
{
   return PtrSelectEvents (ctx, events, freq, MUXvalue);
}

IPR_STATUS CtrSelectEventsFromArray (IN int ctx, IN const IPR_EVENT *events,
                                          IN const IPR_FREQ *freqs, IN IPR_SIZE asize,
                                          OUT uint64 *MUXvalue)
{
   return PtrSelectEventsFromArray (ctx, events, freqs, asize, MUXvalue);
}

IPR_STATUS CtrSetEvents (IN int ctx, IN uint64 *MUXvalue)
{
   return PtrSetEvents (ctx, MUXvalue);
}

IPR_STATUS CtrSetModes (IN int ctx, IN IPR_MMASK modes)
{
   return PtrSetModes (ctx, modes);
}

IPR_STATUS CtrEnable (IN int ctx, IN IPR_CTRMASK ctrs)
{
   return PtrEnable (ctx, ctrs);
}

IPR_STATUS CtrDisable (IN int ctx, IN IPR_CTRMASK ctrs)
{
   return PtrDisable (ctx, ctrs);
}

IPR_STATUS CtrPause (IN int ctx, IN IPR_CTRMASK ctrs)
{
   return PtrPause (ctx, ctrs);
}

IPR_STATUS CtrResume (IN int ctx, IN IPR_CTRMASK ctrs)
{
   return PtrResume (ctx, ctrs);
}

IPR_STATUS CtrClear (IN int ctx, IN const IPR_CTRMASK ctrs)
{
   return PtrClear (ctx, ctrs);
}

IPR_STATUS CtrWrite (IN int ctx, IN const IPR_CTRMASK ctrs,
                            IN int *vals, IN IPR_SIZE siz)
{
   return PtrWrite (ctx, ctrs, vals, siz);
}

IPR_STATUS CtrRead (IN int ctx, IN const IPR_CTRMASK ctrs,
                           OUT int *vals, IN IPR_SIZE siz)
{
   return PtrRead (ctx, ctrs, vals, siz);
}

IPR_STATUS CtrReadAll (IN int ctx, OUT int *vals, IN IPR_SIZE siz)
{
   return PtrReadAll (ctx, vals, siz);
}

IPR_STATUS CtrAvailableCounters (IN IPR_EVENT evt,
                                        OUT IPR_CTRMASK *cmask)
{
   return PtrAvailableCounters (evt, cmask);
}

IPR_STATUS CtrCounterEvents (IN IPR_CTRNUMBER cnum, OUT IPR_EVENTSET *e)
{
   return PtrCounterEvents (cnum, e);
}


/*****************************************************************************/
/* FREQUENCY ADJUSTMENT ROUTINES                                             */
/*                                                                           */
/* The frequency adjustment routines below convert the specified frequency   */
/* from internal to external or from external to internal format:            */
/*                                                                           */
/*   Internal format = The actual counter delta value at the time of         */
/*                     interrupt.                                            */
/*   External format = The number of events that have occurred to cause      */
/*                     the interrupt.                                        */
/*                                                                           */
/* There are different because some counter options actually count, for      */
/* instance, double events or half events.                                   */
/*                                                                           */
/* Arguments to these functions are passed in a set of arrays, where the     */
/* same index corresponds across all the arrays:                             */
/*                                                                           */
/*   adjfreqlist [i]   The set of events that are adjusted by a particular   */
/*                     amount.                                               */
/*   shiftlist [i]     The left shift amount (could be negative) used to     */
/*                     convert the INTERNAL format to EXTERNAL format.  The  */
/*                     reverse operation will be done when converting        */
/*                     EXTERNAL to INTERNAL format.                          */
/*   listsize          The number of elements in the arrays.                 */
/*                                                                           */
/* Please note that the sense of "high" and "low" in frequency is reversed   */
/* from what might be intuitive.  If, for example, a counter can interrupt   */
/* every 256 events or every 4096 events, the higher number (4096)           */
/* represents a larger interval between interrupts, and therefore less       */
/* frequent interrupts.  Similarly, the lower number (256) represents a      */
/* shorter interval between interrupts, and therefore more frequent          */
/* interrupts.  This is because the numbers 256 and 4096 really are the      */
/* reciprocal of the frequencies.                                            */
/*                                                                           */
/* The term "frequency" probably should be "wavelength," and the terms       */
/* "higher" and "lower" should be "longer" and "shorter."  Ah, well...       */
/*                                                                           */
/*****************************************************************************/

// Convert internal to external frequency

IPR_FREQ int_to_ext_freq (
   IN const IPR_EVENT event,            // Event for which frequency is desired
   IN IPR_FREQ freq_in,                 // Original frequency
   IN const IPR_EVENTSET *adjfreqlist,  // List of adjusted frequencies
   IN const int *shiftlist,             // List of frequency shift values
   IN IPR_SIZE listsize)                // Size of frequency list
{
   IPR_SIZE
      i;

   for (i = 0; i < listsize; i++)
   {

      /* Determine if any of the event sets contain this event */

      if (CtrVerifyEventInSet (&adjfreqlist [i], event))

         /* Yes; use the corresponding shift amount to convert frequency */

         if (shiftlist [i] > 0)
            return freq_in << shiftlist [i];
         else
            return freq_in >> (-shiftlist [i]);
   }

   /* Event not in the passed sets, so no conversion necessary */

   return freq_in;
}

/* Convert external to internal frequency */

IPR_FREQ ext_to_int_freq (
   IN const IPR_EVENT event,            // Event for which frequency is desired
   IN IPR_FREQ freq_in,                 // Original frequency
   IN const IPR_EVENTSET *adjfreqlist,  // List of adjusted frequencies
   IN const int *shiftlist,             // List of frequency shift values
   IN IPR_SIZE listsize)                // Size of frequency lists
{
   IPR_SIZE
      i;

   for (i = 0; i < listsize; i++)
   {

      /* Determine if any of the event sets contain this event */

      if (CtrVerifyEventInSet (&adjfreqlist [i], event))

         /* Yes; use the corresponding shift amount to convert frequency */

         if (shiftlist [i] < 0)
            return freq_in << (-shiftlist [i]);
         else
            return freq_in >> shiftlist [i];
   }

   /* Event not in the passed sets, so no conversion necessary */

   return freq_in;
}

// Determine the frequencies available for a particular event

IPR_FREQ available_freqs (
   IN const IPR_EVENT event,           // Candidate event
   OUT IPR_FREQ *freqs,                // Returned frequency list
   IN const IPR_EVENTSET *esetarray,   // Events per counter
   IN const IPR_FREQ *freqarray,       // Frequencies per counter
   IN IPR_COUNT ctrcnt,                // Number of counters
   IN const IPR_EVENTSET *adjfreqlist, // Events that require frequency adjustment
   IN const int *shiftlist,            // Shift amounts to adjust frequencies
   IN IPR_SIZE listsize)               // Size of previous two arrays
{
   IPR_COUNT
      i;
   IPR_BOOLEAN
      freq_OK;

   //
   // Find out which counters can count this event, and add the frequency lists
   // for those counters.
   //
   *freqs = 0;
   freq_OK = FALSE;
   for (i = 0; i < ctrcnt; i++)
   {
      if (CtrVerifyEventInSet (&esetarray [i], event))
      {
         freq_OK = TRUE;
         *freqs |= int_to_ext_freq (event, freqarray [i], adjfreqlist,
                                    shiftlist, listsize);
      }
   }
   return freq_OK ? IPR__SUCCESS : IPR__INVALID;
}

// Get the current frequency setting for an event

IPR_STATUS get_current_freq (
   IN const IPR_EVENT event,            // Event for which frequency is desired
   OUT IPR_FREQ *freq,                  // Returned frequency
   IN const CTRCTXDATA *ctrdata,        // Counter context information
   IN IPR_COUNT ctrcnt,                 // Number of counters
   IN const IPR_EVENTSET *adjfreqlist,  // List of adjusted frequencies
   IN const int *shiftlist,             // List of frequency shift values
   IN IPR_SIZE listsize)                // Size of frequency lists
{
   IPR_STATUS
      status;
   IPR_CTRNUMBER
      ctr;

   status = get_counter_number (event, &ctr, ctrdata, ctrcnt);
   if (!IPR_SUCCESS (status))
      return status;
   *freq = int_to_ext_freq (event, ctrdata [ctr] . freq, adjfreqlist,
                            shiftlist, listsize);
   return IPR__SUCCESS;
}

/*****************************************************************************/
/* Determine the lowest frequency in a frequency set; that is, the frequency */
/* with the highest bit number.                                              */
/*****************************************************************************/

IPR_FREQ lowest_freq (IN IPR_FREQ fmask)
{
   IPR_FREQ
      lbit;

//
// Mask out the lowest bit.  If there are no other bits, we have it.  Otherwise,
// remove this bit and try again.
//
   lbit = BIprLowest (fmask);
   while (lbit != fmask)
   {
      fmask = BIprClr (fmask, lbit);
      lbit = BIprLowest (fmask);
   }
   return lbit;
}

/*****************************************************************************/
/* Determine the highest frequency in a frequency set; that is, the          */
/* frequency with the lowest bit number.                                     */
/*****************************************************************************/

IPR_FREQ highest_freq (IN IPR_FREQ fmask)
{
   return BIprLowest (fmask);
}

IPR_STATUS available_counters (IPR_EVENT evt,
                                      const IPR_EVENTSET *cevents,
                                      IPR_COUNT ctrcnt,
                                      IPR_CTRMASK *cmask)
{
   IPR_COUNT
      ctr;

   if (!cmask)
      return IPR__NULLPTR;
   *cmask = 0;
   for (ctr = 0; ctr < ctrcnt; ctr++)
      if (CtrVerifyEventInSet (&cevents [ctr], evt))
         *cmask = (IPR_CTRMASK) BIprSet1 (*cmask, ctr);
   return IPR__SUCCESS;
}

IPR_STATUS counter_events (IN IPR_CTRNUMBER ctr,
                                  IN const IPR_EVENTSET *cevents,
                                  IN IPR_COUNT ctrcnt,
                                  OUT IPR_EVENTSET *e)
{
   if (!e)
      return IPR__NULLPTR;
   if (ctr < 0 || ctr >= ctrcnt)
      return IPR__INVALID;
   *e = cevents [ctr];
   return IPR__SUCCESS;
}


/*****************************************************************************/
/* EVENT SET MANIPULATION ROUTINES                                           */
/*                                                                           */
/* These routines permit manipulation of event sets without requiring        */
/* knowledge of how event sets are constructed.                              */
/*****************************************************************************/

#define BITS_IN_MASK (sizeof (uint64) * 8)
#define MAX_EVENT (BITS_IN_MASK * EVENT_MASK_COUNT)
#define MIN_EVENT 1

static const int
   bits_in_mask = BITS_IN_MASK;

IPR_SIZE CtrEventSetSize (void)
{
   return sizeof (IPR_EVENTSET);
}

IPR_STATUS CtrClearEventSet (OUT IPR_EVENTSET *e)
{
   int i;

   for (i = 0; i < EVENT_MASK_COUNT; i++)
      e -> mask [i] = 0;
   return IPR__SUCCESS;
}

IPR_STATUS CtrInsertEvent (OUT IPR_EVENTSET *s, IN const IPR_EVENT e)
{
   int
      idx,
      bit;
   uint64
      msk;

   if (e)
   {
      idx = (e - 1) / bits_in_mask;
      if (idx < EVENT_MASK_COUNT)
      {
         bit = (e - 1) % bits_in_mask;
         msk = 1;
         msk <<= bit;
         s -> mask [idx] |= msk;
      }
   }
   return IPR__SUCCESS;
}

IPR_STATUS CtrDeleteEvent (OUT IPR_EVENTSET *s, IN const IPR_EVENT e)
{
   int
      idx,
      bit;
   uint64
      msk;

   if (e)
   {
      idx = (e - 1) / bits_in_mask;
      if (idx < EVENT_MASK_COUNT)
      {
         bit = (e - 1) % bits_in_mask;
         msk = 1;
         msk <<= bit;
         s -> mask [idx] &= ~msk;
      }
   }
   return IPR__SUCCESS;
}

IPR_STATUS CtrMaskEventSet (OUT IPR_EVENTSET *dst, IN const IPR_EVENTSET *msk)
{
   int
      i;

   /* Clear all event bits in the destination that are not set in the mask. */

   for (i = 0; i < EVENT_MASK_COUNT; i++)
      dst -> mask [i] &= msk -> mask [i];
   return IPR__SUCCESS;
}

IPR_STATUS CtrMergeEventSet (OUT IPR_EVENTSET *dst, IN const IPR_EVENTSET *msk)
{
   int i;

   /* Set all event bits in the destination that are set in the mask. */

   for (i = 0; i < EVENT_MASK_COUNT; i++)
      dst -> mask [i] |= msk -> mask [i];
   return IPR__SUCCESS;
}

IPR_STATUS CtrRemoveEventSet (OUT IPR_EVENTSET *dst, IN const IPR_EVENTSET *msk)
{
   int i;

   /* In the destination argument, clear all bits set in the mask argument */

   for (i = 0; i < EVENT_MASK_COUNT; i++)
      dst -> mask [i] &= ~msk -> mask [i];
   return IPR__SUCCESS;
}

IPR_BOOLEAN CtrVerifyEventInSet (IN const IPR_EVENTSET *s,
                                        IN const IPR_EVENT e)
{
   int
      idx,
      bit;
   uint64
      msk;

   if (e == 0)
      return FALSE;
   //
   // Find the corresponding bit.  Remember that event number one corresponds
   // to bit zero in the mask.
   //
   idx = (e - 1) / bits_in_mask;
   if (idx >= EVENT_MASK_COUNT)
      return FALSE;
   bit = (e - 1) % bits_in_mask;

   // Create a mask portion with the appropriate bit set

   msk = 1;
   msk <<= bit;

   // Check if the bit is set in the appropriate mask portion

   return (!!(s -> mask [idx] & msk));
}

IPR_EVENT CtrEventToNumber (IN const IPR_EVENTSET *event)
{
   int
      n,
      i;
   uint64
      m;

   n = 0;
   m = 0;

   /* Find out which portion of the event mask the bit is in */

   for (i = 0; i < EVENT_MASK_COUNT && !m; i++)
   {
      if (event -> mask [i])
         m = event -> mask [i];
      else
         n += bits_in_mask;
   }

   if (!m)
      return 0;

   // Clear out all but the lowest bit

   m &= ((~m) + 1);
   //
   // Add one for each extra bit within the event mask portion.  Note that
   // the termination condition is when m equals zero, which means that the
   // value added to n will be one more than the bit position of the set
   // set bit.  The resulting return value will thus start at one rather
   // than zero, allowing the use of zero to indicate that no value is
   // present.
   //
   for (; m; m >>= 1)
      n++;
   return n;
}

IPR_COUNT CtrNumberEventsInSet (IN const IPR_EVENTSET *x)
{
   int
      i,
      bits;

   /* Add up the number of bits in each portion of the event set */

   bits = 0;
   for (i = 0; i < EVENT_MASK_COUNT; i++)
      bits += BIprCnt (x -> mask [i]);
   return bits;
}

IPR_STATUS CtrExpandEventSet (IN const IPR_EVENTSET *s, OUT IPR_EVENT *a, IN IPR_SIZE size)
{
   IPR_SIZE
      idx;
   IPR_EVENT
      e;
   IPR_EVENTSET
      tmp;

   //
   // Scan sequentially through all possible event numbers.  Return those
   // that are in the set.
   //
   idx = 0;
   tmp = *s;
   for (e = CtrEventToNumber (&tmp); e; e = CtrEventToNumber (&tmp))
   {
      if (idx < size)
         a [idx++] = e;
      else
         return IPR__TOO_MANY_EVENTS;
      (void) CtrDeleteEvent (&tmp, e);
   }
   while (idx < size)
      a [idx++] = 0;
   return IPR__SUCCESS;
}


/********************************************/
/* Generalized core chip-specific functions */
/********************************************/

IPR_BOOLEAN valid_event_set (
   IN const IPR_EVENTSET *candidate,    // Event set to verify
   IN const IPR_EVENTSET *esetarray,    // Array of event sets for each counter
   IN IPR_COUNT ctrcnt)                 // Number of counters
{
   IPR_EVENTSET
      tmp;
   IPR_COUNT
      i;

   tmp = *candidate;

   /* Clear all valid bits in the candidate */

   for (i = 0; i < ctrcnt; i++)
      (void) CtrRemoveEventSet (&tmp, &esetarray [i]);

   // The result should be zero; indicate whether this is the case

   return (CtrNumberEventsInSet (&tmp) == 0);
}

IPR_STATUS available_events (
   IN const IPR_EVENTSET *initial_events,  // Starting group of events
   OUT IPR_EVENTSET *more_events,          // Returned additional legal events
   IN const IPR_EVENTSET *esetarray,       // Valid events for each counter
   IN IPR_COUNT ctrcnt)                    // Number of counters
{
   IPR_COUNT
      i,j;
   int
      assigned;
   IPR_BOOLEAN
      all_assigned,
      changed,
      unique;
   IPR_EVENT
      e;
   IPR_EVENTSET
      tmp,
      assigned_events;

   // Check if any bits beyond the actual capabilities are set

   assigned = 0;
   (void) CtrClearEventSet (&assigned_events);
   if (!valid_event_set (initial_events, esetarray, ctrcnt))
      return IPR__UNAVAILABLE;
   //
   // Assign an event to each counter.  If an event can only be assigned to
   // one counter, use that counter.  If an event can be assigned to more
   // than one counter, and neither counter has been spoken for, make no
   // assumptions and return the event sets for all unassigned counters.
   //
   for (i = 0; i < ctrcnt; i++)
   {
      tmp = *initial_events;
      (void) CtrMaskEventSet (&tmp, &esetarray [i]);
      for (e = CtrEventToNumber (&tmp); e; e = CtrEventToNumber (&tmp))
      {

         // An event was specified for this counter; is it unique?

         (void) CtrDeleteEvent (&tmp, e);
         unique = TRUE;
         for (j = 0; j < ctrcnt && unique; j++)
            if (j != i && CtrVerifyEventInSet (&esetarray [j], e))
               unique = FALSE;

         // If it is unique, but another unique event was found, error

         if (unique)
            if (assigned & (1 << i))
               return IPR__INVALID;
            else
            {
               assigned |= (1 << i);
               (void) CtrInsertEvent (&assigned_events, e);
            }
      }
   }
   //
   // All unique events have been assigned.  Now see if there is exactly one
   // remaining event for each unassigned counter.  Keep doing this until no
   // more changes are forthcoming.
   //
   changed = TRUE;
   while (changed)
   {
      changed = FALSE;
      for (i = 0; i < ctrcnt; i++)
      {
         if (assigned & (1 << i))
            continue;
         tmp = *initial_events;
         (void) CtrRemoveEventSet (&tmp, &assigned_events);
         (void) CtrMaskEventSet (&tmp, &esetarray [i]);
         if (CtrNumberEventsInSet (&tmp) == 1)
         {

            // Only one event for this counter, so assign it to the counter

            assigned |= (1 << i);
            (void) CtrMergeEventSet (&assigned_events, &tmp);
            changed = TRUE;
         }
      }
   }
   all_assigned = TRUE;
   (void) CtrClearEventSet (&tmp);
   for (i = 0; i < ctrcnt; i++)
      if (!(assigned & (1 << i)))
      {
         all_assigned = FALSE;
         (void) CtrMergeEventSet (&tmp, &esetarray [i]);
      }
   //
   // Remove the assigned events from the list, since those are
   // unambiguously determined.
   //
   (void) CtrRemoveEventSet (&tmp, &assigned_events);
   //
   // NOTE: It is possible that FURTHER_EVENTS may be returned even when
   // enough events have been specified to cover all counters.  This would
   // be true if several counters overlapped on more than one specified event;
   // the assignment algorithm cannot tell which counter might get assigned
   // those particular events, so it does not assign them.
   //
   if (all_assigned)
      return IPR__NO_FURTHER_EVENTS;
   if (more_events)
      *more_events = tmp;
   return IPR__FURTHER_EVENTS;
}

IPR_STATUS current_events (
   OUT IPR_EVENTSET *events,       // Returned current events
   IN const CTRCTXDATA *ctrdata,   // Counter context information
   IN IPR_COUNT ctrcnt)            // Number of counters
{
   IPR_COUNT
      i;

   // Start with an empty list

   (void) CtrClearEventSet (events);

   // Add all events that are currently active

   for (i = 0; i < ctrcnt; i++)
         (void) CtrInsertEvent (events, ctrdata [i] . event);
   return IPR__SUCCESS;
}

IPR_STATUS get_counter_number (
   IN const IPR_EVENT event,          // Event for which counter number is wanted
   OUT IPR_CTRNUMBER *counter_number, // Returned counter number
   IN const CTRCTXDATA *ctrdata,      // Counter context information
   IN IPR_COUNT ctrcnt)               // Number of counters
{
   IPR_COUNT i;

   for (i = 0; i < ctrcnt; i++)
      if (event == ctrdata [i] . event)
      {
         *counter_number = i;
         return IPR__SUCCESS;
      }
   return IPR__NOT_AMONG;
}

IPR_STATUS event_counter_map (
   IN IPR_EVENT *event_array,      // Returned array of event numbers
   IN IPR_SIZE asize,              // Size of return array
   IN const CTRCTXDATA *ctrdata,   // Counter context information
   IN IPR_COUNT ctrcnt)            // Number of counters
{
   IPR_COUNT i;

   /* Return all events that fit */

   for (i = 0; i < ctrcnt && i < asize; i++)
      event_array [i] = ctrdata [i] . event;

   /* Clear out the rest of the array, if any */

   for (i = ctrcnt; i < asize; i++)
      event_array [i] = 0;
   return (asize < ctrcnt) ? IPR__OUTRANGE : IPR__SUCCESS;
}

IPR_STATUS freq_counter_map (
   OUT IPR_FREQ *freq_array,            // Returned array of frequencies
   IN IPR_SIZE asize,                   // Size of return array
   IN const CTRCTXDATA *ctrdata,        // Counter context information
   IN IPR_COUNT ctrcnt,                 // Number of counters
   IN const IPR_EVENTSET *adjfreqlist,  // List of adjusted frequencies
   IN const int *shiftlist,             // List of frequency shift values
   IN IPR_SIZE listsize)                // Size of frequency lists
{
   IPR_COUNT i;

   // Return all frequencies that fit

   for (i = 0; i < ctrcnt && i < asize; i++)
      freq_array [i] = int_to_ext_freq (ctrdata [i] . event,
                                        ctrdata [i] . freq,
                                        adjfreqlist,
                                        shiftlist, listsize);

   // Clear out the rest of the array, if any

   for (i = ctrcnt; i < asize; i++)
      freq_array [i] = 0;
   return (asize < ctrcnt) ? IPR__OUTRANGE : IPR__SUCCESS;
}

IPR_STATUS clear_counter_data (IN CTRCTXDATA *ctrdata,
                                      IN IPR_COUNT ctrcnt)
{
   IPR_COUNT
      i;

   for (i = 0; i < ctrcnt; i++)
   {
         ctrdata [i] . event = 0;
         ctrdata [i] . freq = 0;
         ctrdata [i] . specified = FALSE;
   }
   return IPR__SUCCESS;
}

static IPR_STATUS assign_event_to_counter (
   IN const IPR_EVENTSET *req,        // Requested events
   IN IPR_CTRNUMBER ctr,              // Current (initial) counter number
   IN IPR_FREQ freq,                  // Desired interrupt frequency
   IN IPR_COUNT ctrcnt,               // Number of counters
   IN const IPR_EVENTSET *esetarray,  // Events for each counter
   IN const IPR_FREQ *farray,         // Frequencies for each counter
   IN const IPR_EVENTSET *aflist,     // Events that require shifting
   IN const int *slist,               // Amount to shift
   IN IPR_SIZE lsize,                 // Size of shift arrays
   OUT CTRCTXDATA *newctx)               // Counter context area to be filled in
{
   IPR_EVENTSET
      uset,
      rset;
   IPR_CTRNUMBER
      c;
   IPR_BOOLEAN
      freq_OK;
   IPR_FREQ
      frequencies,
      tmp_freq;
   IPR_EVENT
      e;

   rset = *req;
   for (c = ctr; c < ctrcnt; c++)
   {
      if (newctx)
      {
         newctx [c] . specified = FALSE;
         newctx [c] . event = 0;
      }

      // Extract selected events valid for the current counter

      uset = rset;
      (void) CtrMaskEventSet (&uset, &esetarray [c]);

      // Loop through all the events in the current counter event set

      for (e = CtrEventToNumber (&uset); e; e = CtrEventToNumber (&uset))
      {
         //
         // Remove the current event from the set so it will not be found
         // in a later iteration of the loop.
         //
         (void) CtrDeleteEvent (&uset, e);
         //
         // Fill in the frequency for the current event.
         // Start by finding out what frequencies are available for the
         // event.  Note that we need to convert from INTERNAL format
         // to EXTERNAL.  All comparisons with the passed value use
         // EXTERNAL format.
         //
         tmp_freq = freq;
         freq_OK = TRUE;
         if (tmp_freq != 0)      // "Zero OK" is checked by caller
         {
            frequencies = int_to_ext_freq (e, farray [c], aflist, slist,
                                           lsize);
            switch (tmp_freq)
            {
               case IPR_FREQUENCY_HIGHEST :
                  tmp_freq = highest_freq (frequencies);
                  break;
               case IPR_FREQUENCY_LOWEST :
                  tmp_freq = lowest_freq (frequencies);
                  break;
               default :
                  if ((frequencies & tmp_freq) != tmp_freq)
                     freq_OK = FALSE;
                  break;
            }
         }
         if (freq_OK)
         {
            if (newctx)
            {
               newctx [c] . specified = TRUE;
               newctx [c] . event = e;
               newctx [c] . freq = ext_to_int_freq (e, tmp_freq, aflist, slist,
                                                 lsize);
            }

            // Remove the event from further consideration for now

            (void) CtrDeleteEvent (&rset, e);

            // Attempt to assign the remaining events to counters

            if (assign_event_to_counter (&rset, c + 1, freq, ctrcnt, esetarray,
                                         farray, aflist, slist, lsize, newctx)
                == IPR__SUCCESS)
               return IPR__SUCCESS;
            //
            // Did not work, so undo the assignment and replace the event in
            // the request set.
            //
            if (newctx)
            {
               newctx [c] . specified = FALSE;
               newctx [c] . event = 0;
            }
            (void) CtrInsertEvent (&rset, e);
         }
      }
   }
   //
   // Fell through main loop, which means the counter number has been
   // exceeded.  Check for terminating conditions.
   //
   return (CtrNumberEventsInSet (req) == 0) ? IPR__SUCCESS : IPR__INVALID;
}

static IPR_STATUS assign_event_from_array (
   IN const IPR_EVENT *events,        // Requested events
   IN const IPR_FREQ *freqs,          // Requested frequencies
   IN IPR_SIZE asize,                 // Size of the event and frequency arrays
   IN IPR_SIZE idx,                   // Current (initial) array index
   IN IPR_CTRMASK assigned,           // Currently assigned counters
   IN IPR_COUNT ctrcnt,               // Number of counters
   IN const IPR_EVENTSET *esetarray,  // Events for each counter
   IN const IPR_FREQ *farray,         // Frequencies for each counter
   IN const IPR_EVENTSET *aflist,     // Events that require shifting
   IN const int *slist,               // Amount to shift
   IN IPR_SIZE lsize,                 // Size of shift arrays
   OUT CTRCTXDATA *newctx)               // Counter context area to be filled in
{
   IPR_CTRNUMBER
      ctr;
   IPR_BOOLEAN
      freq_OK;
   IPR_FREQ
      frequencies,
      tmp_freq;

   // Check if we have anything to do

   if (idx >= asize)
      return IPR__SUCCESS;

   // Look through all counters and attempt to assign the current event

   for (ctr = 0; ctr < ctrcnt; ctr++)
   {

      // Look at unassigned counters only

      if (assigned & (1 << ctr))
         continue;

      // Attempt to assign the current event to this counter

      if (CtrVerifyEventInSet (&esetarray [ctr], events [idx]))
      {
         tmp_freq = freqs [idx];
         freq_OK = TRUE;
         if (tmp_freq != 0)      // "Zero OK" is checked by caller
         {
            frequencies = int_to_ext_freq (events [idx], farray [ctr], aflist,
                                           slist, lsize);
            switch (tmp_freq)
            {
               case IPR_FREQUENCY_HIGHEST :
                  tmp_freq = highest_freq (frequencies);
                  break;
               case IPR_FREQUENCY_LOWEST :
                  tmp_freq = lowest_freq (frequencies);
                  break;
               default :
                  if ((frequencies & tmp_freq) != tmp_freq)
                     freq_OK = FALSE;
                  break;
            }
         }
         if (freq_OK)
         {
            if (newctx)
            {
               newctx [ctr] . specified = TRUE;
               newctx [ctr] . event = events [idx];
               newctx [ctr] . freq = ext_to_int_freq (events [idx], tmp_freq,
                                                   aflist, slist, lsize);
            }
            //
            // Recurse, moving on to the next event and marking the current
            // counter as assigned.
            //
            if (assign_event_from_array (events, freqs, asize, idx + 1,
                                         assigned | (1 << ctr), ctrcnt,
                                         esetarray, farray, aflist, slist,
                                         lsize, newctx)
                == IPR__SUCCESS)
               return IPR__SUCCESS;
            //
            // Did not work, so undo the assignment and replace the event in
            // the request set.
            //
            if (newctx)
            {
               newctx [ctr] . specified = FALSE;
               newctx [ctr] . event = 0;
            }
         }
      }
   }

   // Cannot make an assignment for the current event

   return IPR__INVALID;
}

IPR_STATUS fill_event_from_array (
   IN const IPR_EVENT *pEvents,             // Events to be assigned
   IN const IPR_FREQ *pFreqs,               // Desired frequencies
   IN IPR_SIZE asize,                       // Size of the arrays
   IN const IPR_EVENTSET *esetarray,        // Events for each counter
   IN const IPR_FREQ *freqarray,            // Frequencies for each counter
   IN IPR_COUNT ctrcnt,                     // Number of counters
   IN const CTRCTXDATA *old,                // Where the old assignments are
   OUT CTRCTXDATA *newctx,                     // Where to save the assignments
   IN const IPR_EVENTSET *adjfreqlist,      // Events with adjusted frequencies
   IN const int *shiftlist,                 // Shift amounts for freq adjustment
   IN IPR_SIZE listsize,                    // Number of frequency adjustments
   IN IPR_BOOLEAN zero_freq_OK)             // Flag indicates if zero freq legal
{
   IPR_SIZE
      i;
   IPR_CTRMASK
      assigned;

   // Verify set of frequencies before proceeding

   for (i = 0; i < asize; i++)
   {
      if (pFreqs [i] == 0)
      {
         if (!zero_freq_OK)
            return IPR__INVALID_FREQUENCY;
      }
      else if (pFreqs [i] != IPR_FREQUENCY_HIGHEST &&
               pFreqs [i] != IPR_FREQUENCY_LOWEST &&
               !BIprIs1 (pFreqs [i]))
         return IPR__INVALID_FREQUENCY;
   }

   // Create the assigned mask

   assigned = 0;
   for (i = 0; i < ctrcnt; i++)
      if (old && old [i] . specified)
         assigned |= (1 << i);

   // Assign the events to the counters

   return assign_event_from_array (pEvents, pFreqs, asize, 0, assigned,
                                   ctrcnt, esetarray, freqarray, adjfreqlist,
                                   shiftlist, listsize, newctx);
}


IPR_STATUS fill_event_from_set (
   IN const IPR_EVENTSET *events,           // Set of events to be assigned
   IN IPR_FREQ freq,                        // Desired frequency for this set
   IN const IPR_EVENTSET *esetarray,        // Events for each counter
   IN const IPR_FREQ *freqarray,            // Frequencies for each counter
   IN IPR_COUNT ctrcnt,                     // Number of counters
   OUT CTRCTXDATA *newctx,                     // Where to save the assignments
   IN const IPR_EVENTSET *adjfreqlist,      // Events with adjusted frequencies
   IN const int *shiftlist,                 // Shift amounts for freq adjustment
   IN IPR_SIZE listsize,                    // Number of frequency adjustments
   IN IPR_BOOLEAN zero_freq_OK)             // Flag indicates if zero freq legal
{
   if ((freq == 0 && !zero_freq_OK) ||
       (freq != IPR_FREQUENCY_HIGHEST && freq != IPR_FREQUENCY_LOWEST
        && !BIprIs1 (freq)))
      return IPR__INVALID_FREQUENCY;

   // Assign the events to the counters

   return assign_event_to_counter (events, 0, freq, ctrcnt, esetarray,
                                   freqarray, adjfreqlist, shiftlist,
                                   listsize, newctx);
}
