#ifndef LINKED_LIST_H
#define LINKED_LIST_H 1

/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/

/*
 * Abstract: The header for a suite of functions that operate on doubly
 *           linked lists, queues and stacks.
 *
 * Author: J. Goddard
 * Creation Date:  28-Aug-1997
 *
 * Modification history:
 *
 *		J. Goddard		10-Sep-1997
 *				Cleanup this file making mods as per work done to structs.c
 */

#include "stdtypes.h"

typedef uint32 nodeidentifier;

typedef struct _doublylinkedlist
{
	struct _doublylinkedlist *pNext;
	struct _doublylinkedlist *pPrevious;
	nodeidentifier Identifier;
	uint32 Size;
} doublylinkedlist, queue, stack;

typedef enum
{
	LL_SUCCESS = 1,
	LL_MALLOC_FAILED,
	LL_BAD_SIZE,
	LL_INSERTION_FAILED
} linkedlistconditioncode; 

/*

													Functions for getting/setting error codes

*/
linkedlistconditioncode GetLastListError(void);
void SetLastListError(linkedlistconditioncode Code);

/*

													Functions for allocating doubly linked list/
													            queue/stack entries

*/
doublylinkedlist *AllocateDoublyLinkedListEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier);
queue *AllocateQueueEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier);
stack *AllocateStackEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier);

/*

													Functions for initializing doubly linked list/
													            queue/stack heads

*/
doublylinkedlist *InitializeDoublyLinkedList(doublylinkedlist *pListHead);
queue *InitializeQueue(queue *pQueue);
stack *InitializeStack(stack *pStack);

/*

													Functions for reading/setting doubly linked list/
											            queue/stack entry headers

*/
void SetNodeIdentifier(doublylinkedlist *pNode, nodeidentifier *pNodeIdentifier);
void SetQueueEntryIdentifier(queue *pEntry, nodeidentifier *pIdentifier);
void SetStackEntryIdentifier(stack *pEntry, nodeidentifier *pIdentifier);
void ReadNodeIdentifier(doublylinkedlist *pNode, nodeidentifier *pNodeIdentifier);
uint32 GetNodeSize(doublylinkedlist *pNode);

/*

													Functions for deallocating doubly linked list/
  													            queue/stack entries

*/
void DeallocateNode(doublylinkedlist *pNode);
void DeallocateQueueEntry(queue *pNode);
void DeallocateStackEntry(stack *pNode);

/*

											Functions for testing a doubly linked list/queue/stack

*/
boolean IsEmptyDoublyLinkedList(doublylinkedlist *pListHead);
boolean IsEmptyQueue(queue *pQueue);
boolean IsEmptyStack(stack *pStack);
/*

											Functions for adding/removing an entry to/from a doubly linked list

*/
doublylinkedlist *AddToDoublyLinkedList(doublylinkedlist *pPrevious, doublylinkedlist *pNew);
doublylinkedlist *AddToHeadOfDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pNew);
doublylinkedlist *AddToTailOfDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pNew);
doublylinkedlist *AddToDoublyLinkedListWithAssistance(doublylinkedlist *pListHead,
doublylinkedlist *pNew,
doublylinkedlist *(*pAssistanceFunction) (doublylinkedlist *, doublylinkedlist *));
doublylinkedlist *RemoveFromDoublyLinkedList(doublylinkedlist *pNode);
doublylinkedlist *RemoveFromHeadOfDoublyLinkedList(doublylinkedlist *pListHead);
doublylinkedlist *RemoveFromTailOfDoublyLinkedList(doublylinkedlist *pListHead) ;
/*

											Functions for adding/removing an entry to/from a queue

*/
queue *AddToQueue(queue *pQueue, queue *pNewEntry);
queue *RemoveFromQueue(queue *pQueue);
/*

											Functions for adding/removing an entry to/from a stack

*/
stack *Push(stack *pStack, stack *pNewEntry);
stack *Pop(stack *pStack);
/*

										Functions for reading a doubly linked list or queue

*/
doublylinkedlist *ReadDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pLast);
queue *ReadQueue(queue *pQueue, queue *pLast);
/*

										Functions for flushing a doubly linked list/queue/stack

*/
void FlushDoublyLinkedList(doublylinkedlist *pListHead, void (*pCallBackFunction)(doublylinkedlist *));
void FlushQueue(queue *pQueue, void (*pCallBackFunction)(queue *));
void FlushStack(stack *pStack, void (*pCallBackFunction)(stack *));

/*

										Functions for walking a doublylinkedlist

*/
doublylinkedlist *Next(doublylinkedlist *pNode);
doublylinkedlist *Previous(doublylinkedlist *pNode);


/* Functions for freeing a doublylinkedlist  */

void FreeLinkedList(doublylinkedlist *pNode);

#endif  /* LINKED_LIST_H */

