/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Program: REP (Read Entry Points) For All Platforms.
 *
 * Abstract: This module contains the routines for managing REP output files
 *
 * Author: Michael Dempsey
 * Creation Date:  15-July-1994
 *
 * Modification history:
 *
 * J. Goddard		24-Aug-1998
 *			Problem: Not all symbols on NT get processed
 *			Solution: The problem is that sometimes Visual C uses whitespace in symbol names.
 *						 This caused the parsing code in AppendRepFile to prematurely stop.
 *			Problem: REP is too slow.
 *			Solution: Reworked symbol accumulation so that instead of creating an ordered list
 *						 during symbol intake we now dump the symbols in a list, after all symbols
 *						 are gotten all nodes in the list are dump into an array and qsorted. The
 *						 sorted contents of the array is then used to form a new linked list.
 * J. Goddard		5-Mar-1998
 *			Problem: rep too slow for images with many symbols
 *			Solution: remove function/module name lists and store
 *						 the function/module names in the symbol node
 *
 * J. Goddard		17-Feb-1998
 *			Problem: REP processes images with many symbols too slowly
 *			Solution: 
 * J. Goddard		10-Sep-1997
 *			Problem: REPing images with large numbers of symbols
 *							 often exceeds the statically allocated symbol
 *							 array. Its a major pain to get problem reports
 *							 every month saying that REP ran out of space.
 *			Solution: Modify REP to use a dynamically allocated symbol
 *								list....the way it should have been done in the
 *								first place. :^( This method will be implemented
 *								using doubly linked lists.
 *
 * Phil Ezolt: 18-Jun-97 
 *     Problem:  VMS Wouldn't write out 64 bit numger.
 *     Solution: Put in proper REPLINEFORMAT
 * 
 * Phil Ezolt: 18-Jun-97 
 *     Problem:  Not Platform independent.
 *     Solution: Changed from NT file descriptors to standard UNIX file descriptors. 
 *               Made all addresses uint64 bit, and changed file specification.
 *
 * Phil Ezolt: 27-May-97 
 *     Problem:  Code hard to read and some functions not commented. 
 *     Solution: Renamed varibles and added block comments before functions, moved 
 *               some constants to #defines.
 *
 */
#ifdef nt
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "stdtypes.h"
#include "structs.h"
#include "rep_common.h"

static char *FindModuleNameField(char *pRecord);
static uint32 IncrementSymbolCount(void);
static uint32 *GetSymbolCountPointer(void);

/************************************************************************************/
/* Name: GetFileHandlePointer                                                       */
/*                                                                                  */ 
/* Purpose: This function has static storage space to store the FileHandle.         */
/*                                                                                  */ 
/* Arguments: None.                                                                 */ 
/*                                                                                  */ 
/* Output: None                                                                     */
/*                                                                                  */ 
/* Function Value:  A pointer to the file handle.                                   */
/*                                                                                  */ 
/************************************************************************************/
FILE **GetFileHandlePointer (void)
{
  static FILE *FileDescriptor = NULL;		// Global Handle...

  return &FileDescriptor;
}

/************************************************************************************/
/* Name: GetRepFileNamePointer                                                      */
/*                                                                                  */ 
/* Purpose: This function has static storage space to store the output file name.   */
/*                                                                                  */ 
/* Arguments: None.                                                                 */ 
/*                                                                                  */ 
/* Output: None                                                                     */
/*                                                                                  */ 
/* Function Value:  A pointer to the output file name.                              */
/*                                                                                  */ 
/************************************************************************************/
char *GetRepFileNamePointer(void)
{
  static char RepFileName[PATHNAMESIZE];

  return RepFileName;
}

/************************************************************************************/
/* Name: OpenRepFile                                                                */
/*                                                                                  */ 
/* Purpose: This function opens the resolved addresses output file                  */
/*                                                                                  */ 
/* Arguments: FileName: This is the filename of the output file.                    */ 
/*            CommandTable: A pointer to the table of command line flags.           */
/*                                                                                  */ 
/* Output: None                                                                     */
/*                                                                                  */ 
/* Function Value:  A boolean that represents the success and/or failure of the     */
/*                  function.                                                       */ 
/*                                                                                  */ 
/************************************************************************************/
boolean OpenRepFile(char *FileName, CMD_TABLE *CommandTable)
{
	boolean FunctionValue = TRUE;
  char *RepFileName;
  FILE **FileDescriptor = GetFileHandlePointer();
  RepFileName = GetRepFileNamePointer();

	//
	// Keep a local copy of the resolved addresses file name
	//
  strcpy(RepFileName, FileName);    

	//
	// Create the file
	//
  if ((*FileDescriptor = fopen(FileName, "wb")) == NULL && !CommandTable->Quiet)
  {
		fprintf(stderr, "Error opening file \"%s\", error code: %d\n", FileName, errno);
    FunctionValue = FALSE;
  }

  return FunctionValue;
}

/************************************************************************************/
/* Name: CloseRepFile                                                               */ 
/*                                                                                  */ 
/* Purpose: This closes the REP resolved addresses output file.                     */  
/*                                                                                  */ 
/* Arguments: None                                                                  */ 
/*                                                                                  */ 
/* Output: None                                                                     */ 
/*                                                                                  */ 
/* Notes: None.                                                                     */ 
/*                                                                                  */ 
/************************************************************************************/
void CloseRepFile(void)
{
  FILE **FileDescriptor = GetFileHandlePointer();

  if (*FileDescriptor != NULL)
  {
    fclose(*FileDescriptor);
  }

	return;
}

/************************************************************************************/
/* Name: GetSymbolListHeadPointer																	*/ 
/*                                                                                  */ 
/* Purpose: Get the pointer to the symbol listhead												*/
/*                                                                                  */
/* Arguments: none																						*/
/*                                                                                  */ 
/* Output: a pointer to the symbol list listhead												*/
/*                                                                                  */ 
/************************************************************************************/
doublylinkedlist *GetSymbolListHeadPointer(void)
{
	static doublylinkedlist SymbolList = {&SymbolList, &SymbolList};

	return &SymbolList;
}

/************************************************************************************/
/* Name: AddressCompare																					*/ 
/*                                                                                  */ 
/* Purpose: Compare two address ranges																*/
/*                                                                                  */
/* Arguments: pAddressA: a pointer to a pointer of the first symbol node				*/
/*				  pAddressB: a pointer to a pointer of the second symbol node				*/
/*                                                                                  */ 
/* Output: none																							*/
/*																												*/
/* Function value: A_EQUALS_B if the address ranges are identical							*/
/*						 A_LESS_THAN_B if the first address range is less than the			*/
/*											second															*/
/*						 A_GREATER_THAN_B if the first address range is greater than the	*/
/*												second														*/
/*																												*/
/* Notes: none																								*/
/*                                                                                  */ 
/************************************************************************************/
static int AddressCompare(const void *Thing1, const void *Thing2)
{
	int32 FunctionValue;
	symbolnode **pSymbolNodeA = (symbolnode **) Thing1, **pSymbolNodeB = (symbolnode **) Thing2;

  if ((*pSymbolNodeA)->BeginningAddress < (*pSymbolNodeB)->BeginningAddress)
  {
    FunctionValue = A_LESS_THAN_B;
  }
  else
	{
		if ((*pSymbolNodeA)->BeginningAddress > (*pSymbolNodeB)->BeginningAddress)
		{
	    FunctionValue = A_GREATER_THAN_B;
	  }
	  else
		{
			if ((*pSymbolNodeA)->EndingAddress < (*pSymbolNodeB)->EndingAddress)
		  {
		    FunctionValue = A_LESS_THAN_B;
		  }
			else
			{
				if ((*pSymbolNodeA)->EndingAddress > (*pSymbolNodeB)->EndingAddress)
				{
			    FunctionValue = A_GREATER_THAN_B;
			  }
			  else
			  {
			    FunctionValue = A_EQUALS_B;
			  }
			}
		}
	}

	return FunctionValue;
}

/************************************************************************************/
/* Name: AddSymbolToList																				*/ 
/*                                                                                  */ 
/* Purpose: Add a new symbol to the symbol list													*/
/*                                                                                  */
/* Arguments:  pSymbolName: The Current Symbol													*/
/*             pDllName: The name of the DLL that it comes from.                    */
/*             BeginningAddress: The start address of the function                  */  
/*             EndAddress: The end address of the function                          */
/*             CommandTable: A pointer to the table containing the command flags    */
/*                                                                                  */ 
/* Output: The function returns TRUE if it was successful.                          */ 
/*                                                                                  */ 
/************************************************************************************/
boolean AddSymbolToList(char *pSymbolName, char *pDllName, uint64 BeginningAddress, uint64 EndingAddress)
{
	boolean FunctionValue = TRUE;
	string *pString;
	symbolnode *pSymbolNode;

	//
	// Allocate symbol node.
	//
	//		 A symbol node contains pertinent information about a function.
	//		 (i.e. its name, name of constituent module and address range)
	//
	//		Note: Some libraries (and probably other things as well) on Digital
	//					Unix have the symbols stripped out of them. Instead of getting
	//					no symbols you get a raft of symbols named '(stripped)'. :^/
	//
#ifndef unix
	if ((pSymbolNode = AllocateSymbolNode(strlen(pSymbolName) + strlen(pDllName))) != NULL)
#else
	if (strcmp(pSymbolName, "(stripped)") != 0 &&
	(pSymbolNode = AllocateSymbolNode(strlen(pSymbolName) + strlen(pDllName))) != NULL)
#endif
	{
		//
		// Populate the new symbol node...
		//
		//		Notice that the function and moduel names are packed together and separated
		//		by a NUL. The PointToFunctionName/PointToModuleName functions will locate
		//		the desired field.
		//
		pSymbolNode->BeginningAddress = BeginningAddress;
		pSymbolNode->EndingAddress = EndingAddress != BeginningAddress ? EndingAddress : 0xffffffffffffffffL;
		memmove(PointToFunctionName(pSymbolNode), pSymbolName, strlen(pSymbolName) + sizeof(NUL));
		memmove(PointToModuleName(pSymbolNode), pDllName, strlen(pDllName) + sizeof(NUL));

		IncrementSymbolCount();
		AddToHeadOfDoublyLinkedList(GetSymbolListHeadPointer(), (doublylinkedlist *) pSymbolNode);
	}
	else
	{
		FunctionValue = FALSE;
	}

	return FunctionValue;
}

/************************************************************************************/
/* Name: WriteRep                                                                   */ 
/*                                                                                  */ 
/* Purpose:  Writes an entry to the resolved addresses file                         */ 
/*                                                                                  */
/* Arguments:  SymbolName: The Current Symbol                                       */
/*             DllName: The name of the DLL that it comes from.                     */
/*             BeginningAddress: The start address of the function                  */  
/*             EndAddress: The end address of the function                          */
/*             CommandTable: A pointer to the table containing the command flags    */
/*                                                                                  */ 
/* Output: The function returns TRUE if it was successful.                          */ 
/*                                                                                  */ 
/* Note: o The pointer CommandTable may be NULL.												*/
/*                                                                                  */ 
/************************************************************************************/
boolean WriteRep(CMD_TABLE *CommandTable)
{
	boolean FunctionValue = TRUE;
	char TempSymbolAndDllString[TEMP_STRING_SIZE], SymbolAndDllString[TEMP_STRING_SIZE];
  FILE **FileDescriptor = GetFileHandlePointer();
	symbolnode *pLast = NULL;
	uint32 SymbolAndDllStringLength;

	while ((pLast =
	(symbolnode *) ReadDoublyLinkedList(GetSymbolListHeadPointer(), (doublylinkedlist *) pLast)) != NULL)
	{
		(void) sprintf(TempSymbolAndDllString, "%s %s", PointToFunctionName(pLast), PointToModuleName(pLast)); 
    
	  SymbolAndDllStringLength = sprintf(SymbolAndDllString, REPLINEFORMAT, TempSymbolAndDllString,
		pLast->BeginningAddress, pLast->EndingAddress);
    
	  if (SymbolAndDllStringLength != (fwrite(SymbolAndDllString,1,SymbolAndDllStringLength,*FileDescriptor)))
	  {
			FunctionValue = FALSE;

	    if (CommandTable != NULL && !CommandTable->Quiet)
			{
				fprintf(stderr, "Error writing \"%s\", error code: %d\n", GetRepFileNamePointer(), errno);
	    }

			break;
		}
	}

  return FunctionValue;
}

/************************************************************************************/
/* Name: AppendRepFile                                                              */
/*                                                                                  */
/* Purpose: Opens the given resolved addresses file for appending                   */
/*                                                                                  */
/* Arguments: FileName: The FileName of the file that the resolved addresses        */
/*                      will be appended to.                                        */
/*            CommandTable: A pointer to the table containing the command flags     */
/*                                                                                  */
/* Function Value: The function returns TRUE if it was successful.                  */ 
/*                                                                                  */
/************************************************************************************/
boolean AppendRepFile(char *FileName, CMD_TABLE *CommandFlags)
{
	boolean FunctionValue = TRUE, ScanfFailed = FALSE;
#ifdef nt
	char Function[TEMP_STRING_SIZE], Module[TEMP_STRING_SIZE], InputBuffer[550], *pModuleNameField;
#else
	char Function[TEMP_STRING_SIZE], Module[TEMP_STRING_SIZE];
#endif
	FILE *TempFileHandle;
	uint32 RecordNumber = 1;
	uint64 BeginningAddress, EndingAddress;

	if ((TempFileHandle = fopen(FileName, "r")) != NULL) 
	{
#ifdef nt
		while (fgets(InputBuffer, sizeof(InputBuffer) - 1, TempFileHandle) != NULL)
		{
			if (sscanf(InputBuffer, SCAN_INPUT, Function, Module, &BeginningAddress, &EndingAddress) != 4)
			{
				if ((pModuleNameField = FindModuleNameField(InputBuffer)) != NULL)
				{
					*(pModuleNameField - 1) = NUL;
					strcpy(Function, InputBuffer);

					if (sscanf(pModuleNameField, SCAN_SHORT_INPUT, Module, &BeginningAddress, &EndingAddress) != 3)
					{
						ScanfFailed = TRUE;
						break;
					}	
				}
				else
				{
					ScanfFailed = TRUE;
					break;
				}
			}

			if (!AddSymbolToList(Function, Module, BeginningAddress, EndingAddress))
			{
				FunctionValue = FALSE;
				break;
			}

			RecordNumber++;
		}
#else
		while (ScanfFailed = (fscanf(TempFileHandle, SCAN_INPUT, Function, Module, &BeginningAddress, &EndingAddress) == 4) &&
		(FunctionValue = AddSymbolToList(Function, Module, BeginningAddress, EndingAddress)))
		{
			RecordNumber++;
		}
#endif

		if (ScanfFailed)
		{
			FunctionValue = FALSE;
			fprintf(stderr, "Error: bad record format on record %d\n", RecordNumber);
		}

		if (ferror(TempFileHandle))
		{
			FunctionValue = FALSE;
			fprintf(stderr, "Error: I/O error while reading %s\n"
			"\t%s\n", FileName, strerror(errno));
			FunctionValue = FALSE;
		}

		fclose(TempFileHandle);
	}
	else 
	{
		fprintf(stderr, "Error: unable to open input file %s\n\t%s\n", FileName, strerror(errno));
		FunctionValue = FALSE;
	}

	return FunctionValue;
}

/************************************************************************************/
/* Name: FixAddresses                                                               */ 
/*                                                                                  */ 
/* Purpose: This sign extends a 32 bit value living in a 64 bit number.             */ 
/*                                                                                  */ 
/* Arguments: Value: The value to be extended.                                      */ 
/*                                                                                  */ 
/* Output: None                                                                     */ 
/*                                                                                  */ 
/* Function Value: the Sign extended value.                                         */ 
/*                                                                                  */ 
/************************************************************************************/
uint64 FixAddresses(uint64 Value)
{
	//
	// This function sign-extends the 32-bit version of the passed value on
	// 32-bit operating systems, and leaves it alone on 64-bit systems.  The
	// result is returned as an unsigned 64-bit value.
	//
  if (Value >= 0x80000000L && Value <= 0xffffffffL)
  {
    Value |= 0xffffffff00000000L;
  }

  return Value;
}

/************************************************************************************/
/* Name: PointToFunctionName																			*/ 
/*                                                                                  */ 
/* Purpose: Return a pointer to the function name field in a symbol node				*/ 
/*                                                                                  */ 
/* Arguments: pSymbolNode: pointer to a symbol node											*/ 
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: pointer to the function name field											*/ 
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
char *PointToFunctionName(symbolnode *pSymbolNode)
{
	return &pSymbolNode->FunctionAndModuleNames;
}

/************************************************************************************/
/* Name: PointToModuleName																				*/ 
/*                                                                                  */ 
/* Purpose: Return a pointer to the moduel name field in a symbol node					*/ 
/*                                                                                  */ 
/* Arguments: pSymbolNode: pointer to a symbol node											*/ 
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: pointer to the module name field											*/ 
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
char *PointToModuleName(symbolnode *pSymbolNode)
{
	return strchr(&pSymbolNode->FunctionAndModuleNames, NUL) + 1;
}

/************************************************************************************/
/* Name: AllocateSymbolNode                                                         */ 
/*                                                                                  */ 
/* Purpose: Allocate and initialize a new symbol node                               */ 
/*                                                                                  */ 
/* Arguments: LengthOfNames: the lengths of the module and function names less the  */ 
/*			     trailing NULs			   		    */
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: if success then a pointer to the new symbol node otherwise			*/ 
/*						 NULL																					*/
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
symbolnode *AllocateSymbolNode(uint32 LengthOfNames)
{
	symbolnode *FunctionValue;

	if ((FunctionValue = malloc(sizeof(symbolnode) - sizeof(FunctionValue->FunctionAndModuleNames) + LengthOfNames +
	sizeof(NUL) + sizeof(NUL))) != NULL)
	{
		memset((void *) &FunctionValue->FunctionAndModuleNames, NUL, sizeof(NUL) + sizeof(NUL));
	}

	return FunctionValue;
}

/************************************************************************************/
/* Name: OrderSymbolList																				*/ 
/*                                                                                  */ 
/* Purpose: Orders the raw symbol list by ascending address									*/ 
/*				record.																						*/
/*                                                                                  */ 
/* Arguments: none																						*/ 
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: boolean indicating success (TRUE or failure (FALSE)					*/ 
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
boolean OrderSymbolList(void)
{
	boolean FunctionValue = FALSE;
	symbolnode **pSymbolArray;
	uint32 NumberOfSymbols = *(GetSymbolCountPointer()), Index;

	if ((pSymbolArray = (symbolnode **) malloc(NumberOfSymbols * sizeof(symbolnode *))) != NULL)
	{
		for (Index = 0; Index < NumberOfSymbols; Index++)
		{
			pSymbolArray[Index] = (symbolnode *) RemoveFromHeadOfDoublyLinkedList(GetSymbolListHeadPointer());
		}

		qsort((void *) pSymbolArray, NumberOfSymbols, sizeof(symbolnode *), AddressCompare);

		for (Index = 0; Index < NumberOfSymbols; Index++)
		{
			AddToTailOfDoublyLinkedList(GetSymbolListHeadPointer(), (doublylinkedlist *) pSymbolArray[Index]);
		}

		free(pSymbolArray);

		FunctionValue = TRUE;
	}
	else
	{
		fprintf(stderr,"Error: unable to allocate symbol array\n");
	}

	return FunctionValue;
}

/************************************************************************************/
/* Name: IncrementSymbolCount																			*/ 
/*                                                                                  */ 
/* Purpose: Adds one to the number of symbols collected										*/
/*                                                                                  */ 
/* Arguments: none																						*/ 
/*                                                                                  */ 
/* Output: symbol count is increased by one														*/ 
/*                                                                                  */ 
/* Function Value: the new (incremented) symbol count											*/ 
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
static uint32 IncrementSymbolCount(void)
{
	return ++(*(GetSymbolCountPointer()));
}

/************************************************************************************/
/* Name: GetSymbolCountPointer																		*/ 
/*                                                                                  */ 
/* Purpose: Return a pointer to the number of symbols collected							*/
/*                                                                                  */ 
/* Arguments: none																						*/ 
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: a pointer to the number of symbols collected							*/
/*                                                                                  */
/* Notes: none																								*/
/*																												*/ 
/************************************************************************************/
static uint32 *GetSymbolCountPointer(void)
{
	static uint32 SymbolCount = 0;

	return &SymbolCount;
}
#ifdef nt
/************************************************************************************/
/* Name: FindModuleField																				*/ 
/*                                                                                  */ 
/* Purpose: Calculates the starting address of the module field given an unresolved	*/ 
/*				record.																						*/
/*                                                                                  */ 
/* Arguments: pRecord: pointer to the unresolved record										*/ 
/*                                                                                  */ 
/* Output: None																							*/ 
/*                                                                                  */ 
/* Function Value: pointer to the module field													*/ 
/*                                                                                  */
/* Notes: o This function returns NULL if a third field is not present.					*/
/*			 o This function expects a record of the form:										*/
/*					function_name module_name beginning_address ending_address				*/
/*				where function_name is a function name which may contain embedded			*/
/*										  whitespace														*/
/*						module_name is the name of the module (image file) which				*/
/*										contains the function											*/
/*						beginning_address/ending_address are the hexadecimal beginning		*/
/*																	and ending addresses of the		*/
/*																	named function							*/
/*			o The fields descripbed above may be separated by 1 or more whitespace		*/
/*			  characters.																					*/
/*			o The algorithm used to find the moduel field assumes that the					*/
/*			  module_name, beginning_address and ending_address fields do not contain	*/
/*			  embedded whitespace.																		*/
/*																												*/ 
/************************************************************************************/
static char *FindModuleNameField(char *pRecord)
{
	char *pNext = pRecord + strlen(pRecord) - 2;
	uint32 FieldsFound = 0;

	//
	// main loop -- do until the module field is located or we run out of record
	//
	while (FieldsFound != 3 && pNext != NULL)
	{
		//
		// field loop -- look backwards until the field separator is found or
		//					  we run out of record
		//
		while (!isspace(*pNext) && pNext != pRecord)
		{
			pNext--;
		}

		//
		// whitespace loop -- gobble up extra whitespace between fields
		//
		while (pNext != pRecord && isspace(*pNext))
		{
			pNext--;
		}

		//
		// if all is kosher then chaulk up another field found otherwise
		// jump out of the main loop now
		//
		if (pNext != pRecord)
		{
			FieldsFound++;
		}
		else
		{
			pNext = NULL;
			break;
		}
	}

	if (pNext != NULL)
	{
		pNext += 2;
	}
	
	return pNext;
}
#endif

