/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*

 * Abstract: This program converts a list of beginning and ending addresses 
 * 	     into a complete list, in sorted order, with no overlapping address 
 * 	     ranges and no holes in the address space.
 * 			
 * Author: Brian F. Wibecan
 * Date: 18-Nov-1991
 *
 * MODIFICATION HISTORY:
 *
 * J. Goddard		10-Sep-1997
 *			Problem: REPing images with large numbers of symbols
 *							 often exceeds the statically allocated symbol
 *							 array. Its a major pain to get problem reports
 *							 every month saying that REP ran out of space.
 *			Solution: Modify REP to use a dynamically allocated symbol
 *								list....the way it should have been done in the
 *								first place. :^( This method will be implemented
 *								using doubly linked lists.
 *
 * Phillip G. Ezolt 20-Jun-1997
 *    Problem: Extra NOT_FOUNDS were appearing at the end of appended files.
 *    Solution: Made sure that the 64 bit aligning in GetInputLine didn't cause 
 *              the current address to overflow.  Also, changed program so that it
 *              would not subtract one if the address we were going to write out was
 *              the last address.
 *
 * Phillip G. Ezolt 20-Jun-1997
 *    Problem:
 *    Solution: Removed All Global Variables, moved structure definitions to "repcommon.h"
 *
 * Phillip G. Ezolt 20-Jun-1997
 *    Problem:
 *    Solution: Removed several VMS dependancies. 
 *
 * Phillip G. Ezolt 18-Jun-1997
 *    Problem: 
 *    Solution: Made code more platform independent. All numbers read in/written out are 64-bit numbers.
 *
 * Phillip G. Ezolt 25-May-1997
 *    Problem: 
 *    Solution: Commented various functions, and renamed differnt variable names, 
 *           in order to make code more readable.  
 *
 * Brian F. Wibecan 16-Aug-1994
 *    Problem:
 *    Solution: Revised to enhance performance.  Made program do all transformations
 *	        in a single pass over the data.
 *
 * Satyamahesh Avala 24-Jun-1994
 *    Problem:
 *    Solution:  Modifications are done to link this file to VMS version of REP.
 *               Also modifications carried out such that same file can be linked
 *               to both OSF/1 and Open VMS versions of REP.
 *
 * Satyamahesh Avala 09-Jun-1994
 *    Problem:
 *    Solution: Carried out modifications to facilitate merging of REP with this
 *              for OSF/1.
 *
 * Satyamahesh Avala 10-May-1994
 *    Problem:
 *    Solution: Ported the program on to OSF1.
 *
 * Brian F. Wibecan 02-Oct-1992
 *    Problem:
 *    Solution: Added program header (!) and module identification.  Changed handling
 *	        of NOT_FOUND to check only the beginning of the string, to allow for
 *	        filling in of base addresses for holes in the address space.
 *	        Expanded size of symbol array.  (Lots of resident installed images,
 *	        symbol table files, etc.)
 *
 * Brian F. Wibecan ??-???-????
 *   Problem:
 *   Solution: Fixed up handling of file name.
 *
 * Brian F. Wibecan 18-Nov-1991
 *   Problem: 
 *   Solution: Original program.
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "stdtypes.h"
#include "rep_common.h"

/************************************************************************************/
/* Name: ResolveAddress                                                             */
/*                                                                                  */
/* Purpose: Given a list of symbols sorted in ascending order of address range		*/
/*					fill in address range holes and resolve address range overlaps.		*/
/*                                                                                  */
/* Arguments: OutputFileName: The name of the sorted destination file.              */
/*            Logging: A boolean variable descriping whether or not logging is on.  */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/* Outputs: If successful then the symbol list is written to a file.						*/
/*                                                                                  */
/* Notes: none																								*/
/*                                                                                  */
/************************************************************************************/

void ResolveAddress(boolean Logging, char *OutputFileName)
{
	boolean Error = FALSE;
	FILE *OutputFile;
	symbolnode *pCurrent, *pNode;
	uint32 Added = 0, Overlaps = 0, Underlaps = 0, Zeroes = 0, ImpreciseDiscarded = 0, TotalSymbols = 0,
			 ImpreciseConverted = 0;
	uint64 LastAddress = -1;

	pCurrent = (symbolnode *) GetSymbolListHeadPointer()->pNext;

	//
	// The resolution process is split into two passes. The first pass, this one,
	// discards imprecise symbols (see definition below) and, when possible, converts
	// imprecise symbols into precise.
	//
	if (Logging)
	{
		printf("********\n"
				 "*\n"
				 "*\tPass 1\n"
				 "*\n");
	}

	while (!IsEmptyDoublyLinkedList(GetSymbolListHeadPointer()) &&
	(doublylinkedlist *) pCurrent != GetSymbolListHeadPointer())
	{
		//
		// If the beginning address of the current and next entries are the same and the
		// next entry ends at the high limit of the virtual address range then toss the next
		// entry. The tossed entry is called an imprecise symbol. That is, when we collected
		// it its size was zero. Instead of tossing it then we assigned it an ending address
		// of all f's in the hope that it might 'resolve' based on its neighbors. (Note that
		// when the symbol list is initially constructed it's in random order and doesn't get
		// sorted by address until immediately before calling this function.) Well in this
		// case it didn't so purge it from the list.
		//
		if (pCurrent->BeginningAddress == ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress &&
		((symbolnode *) pCurrent->Links.pNext)->EndingAddress == 0xffffffffffffffffL)
		{
			ImpreciseDiscarded++;
			if (Logging)
			{
				printf("* WARNING: discarding duplicate\n"
						 "* " SCAN_OUTPUT,
 						 PointToFunctionName((symbolnode *) pCurrent->Links.pNext),
 						 PointToModuleName((symbolnode *) pCurrent->Links.pNext),
						 ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress,
						 ((symbolnode *) pCurrent->Links.pNext)->EndingAddress);
			}
			RemoveFromDoublyLinkedList(pCurrent->Links.pNext);
			continue;
		}

		if (pCurrent->EndingAddress == 0xffffffffffffffffL)
		{
			ImpreciseConverted++;
			pCurrent->EndingAddress = ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress;
		}

		TotalSymbols++;
		pCurrent = (symbolnode *) pCurrent->Links.pNext;
	}

	if (Logging)
	{
		printf("* Imprecise symbols converted: %d\n"
				 "* Imprecise symbols discarded: %d\n"
				 "* Symbols into pass 2: %d\n"
				 "*\n"
				 "********\n",
				ImpreciseDiscarded, ImpreciseConverted, TotalSymbols);
	}

	//
	// The resolution process is split into two passes. The second pass, this one,
	// resolves precise symbols (see definition below) and fills in the holes.
	//
	if (Logging)
	{
		printf("********\n"
				 "*\n"
				 "*\tPass 2\n"
				 "*\n");
	}

	pCurrent = (symbolnode *) GetSymbolListHeadPointer()->pNext;
	TotalSymbols = 0;

	while (!IsEmptyDoublyLinkedList(GetSymbolListHeadPointer()) &&
	(doublylinkedlist *) pCurrent != GetSymbolListHeadPointer())
	{
		//
		// If the next list entry is not the listhead (i.e. the current entry is at
		// the tail of the list) and the beginning address of the current entry is
		// greater than the beginning address of the next entry then its an underlap.
		//
		if (pCurrent->Links.pNext != GetSymbolListHeadPointer() &&
		pCurrent->BeginningAddress > ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress)
		{
			//
			// An underlap was detected so count it and then swap the positions
			// of the current and next node so they'll be properly ordered again. Note
			// that its not necessary to check the ending address as code down stream
			// takes care of that complication as an overlap.
			//
			Underlaps++;

			pNode = (symbolnode *) pCurrent->Links.pNext;						// address the successor node
			RemoveFromDoublyLinkedList((doublylinkedlist *) pCurrent);		// remove the current node
			AddToDoublyLinkedList((doublylinkedlist *) pNode,					// insert it after the successor node
			(doublylinkedlist *) pCurrent);
			pCurrent = pNode;																// make the successor node the current node

			continue;
		}

		//
		// Overlap occurs if the ending address for the current node follows the
		// beginning address of the successor node. The various senerios are:
		//
		//	Senerio					Convert this:												To this:
		//
		//     1                |----|                           dropped
		//                      |----|               						 |----|
		//
		//     2                |----|                           |----|
		//                      |------|                              |-|
		//
		//     3                |----|                           |-|
		//                        |----|                           |----|
		//
		//     4                |------|                         |-|
		//                        |----|                           |----|
		//
		//     5                |--------|                       |-|    |-|
		//                        |----|                           |----|
		//
		//
		// Note that cases 2 and 5 involve changing the starting point of
		// the final range in the group being examined.  It is possible that
		// the range after this group will have a starting address that precedes
		// the starting address of the new range in the group.  For example:
		//
		//                |-----|                 |-----|
		//                |--------|                    |--|
		//                |-----------|           |-----------|
		//
		//                |-----|                 |-----|
		//                |--------|                    |--|
		//                  |-----------|            |-----------|
		//
		//                |----------|            |--|    |--|
		//                  |----|                   |----|
		//                  |------|                 |------|
		//
		//                |----------|            |--|    |--|
		//                   |----|                  |----|
		//                     |------|                |-------|
		//
		//
		// This leaves open the possibility that we have to back track, which
		//is done by the underlap code (above) in the next loop iteration.
		//
		//
		// If the next list entry is not the listhead (i.e. the current entry is at
		// the tail of the list) and the ending address of the current entry is
		// greater than the beginning address of the next entry then...
		//
		if (pCurrent->Links.pNext != GetSymbolListHeadPointer() &&
		pCurrent->EndingAddress > ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress)
    {
      Overlaps++;

      if (pCurrent->EndingAddress == ((symbolnode *) pCurrent->Links.pNext)->EndingAddress ||		// Cases 1, 4
			(pCurrent->EndingAddress < ((symbolnode *) pCurrent->Links.pNext)->EndingAddress &&			// Case 3
			pCurrent->BeginningAddress < ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress))
      {
				//
				// How it works for...
				//
				//		case 1: The current entry becomes a zero length function when
				//				  the its ending address is set to the beginning address
				//				  of the next entry. Code down stream will squeeze out
				//				  the new zero length entry.
				//
				//		cases 3 and 4: The address range of the current entry shrinks
				//							by an amount equal to the absolute value of the
				//							difference between the ending address of the
				//							current entry and the beginning address of the
				//							next entry.
				//							There is an exception to this rule in the event that
				//							the address rnage of the next symbol is imprecise.
				//							In this case the next symbol gets converted from imprecise
				//							to precise by setting its ending address to the beginning
				//							address of the next symbols successor.
				//
				pCurrent->EndingAddress = ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress;
      }
      else
      {
				if (pCurrent->EndingAddress < ((symbolnode *) pCurrent->Links.pNext)->EndingAddress)	// Case 2
				{
				//
				// How it works for case 2...
				//
				//		The address range of the next entry shrinks by an amount equal
				//		to the absolute value of the difference between the ending
				//		address of the current entry and the beginning address of the
				//		next entry.
				//
					((symbolnode *) pCurrent->Links.pNext)->BeginningAddress = pCurrent->EndingAddress;
				}
				else							// Case 5
				{
					//
					// How it works for case 5...
					//
					//		First...
					//
					//		Create a new list entry, populate it with the contents
					//		of the current entry and insert it after the current entry's
					//		successor. (The order is current, next and new.)
					//
					if (AddSymbolToList(PointToFunctionName(pCurrent), PointToModuleName(pCurrent),
					((symbolnode *) pCurrent->Links.pNext)->EndingAddress, pCurrent->EndingAddress))
					{
						pCurrent->EndingAddress = ((symbolnode *) pCurrent->Links.pNext)->BeginningAddress;
					}
					else
					{
						fprintf(stderr, "Error: (case 5) allocation failure\n");
						Error = TRUE;
						break;
					}
				}
      }
    }
		//
		// If the current entry has an address range length of zero squeeze it out.
		// (This can happen during under/overlap processing.) If the current node has
		// a predecessor (its not at the head of the list) then that will 
		// become the new current entry. Otherwise if the current node has a
		// successor (its not at the tail of the list) then that will become
		// the new current node. If there is neither a successor or predecessor
		// then the list is (or will be) empty...so stop.
		//
    if (pCurrent->BeginningAddress == pCurrent->EndingAddress)
    {
	      Zeroes++;

			if (pCurrent->Links.pPrevious != GetSymbolListHeadPointer())
			{
				pNode = (symbolnode *) pCurrent->Links.pPrevious;
			}
			else
			{
				pNode = (symbolnode *) pCurrent->Links.pNext;
			}

			RemoveFromDoublyLinkedList((doublylinkedlist *) pCurrent);
			DeallocateNode((doublylinkedlist *) pCurrent);
			pCurrent = pNode;
    }
    else
    {
			//
			// If the address range from the beginning of the predecessor entry
			// to the end of the current entry is not contiguous (there's a hole
			// between them) then create an entry to fill in the hole.
			//
			if (pCurrent->Links.pPrevious != GetSymbolListHeadPointer() &&
			pCurrent->BeginningAddress != ((symbolnode *) pCurrent->Links.pPrevious)->EndingAddress)
			{
				Added++;

				if ((pNode = AllocateSymbolNode(strlen(NOT_FOUND) + strlen(NOT_FOUND))) != NULL)
				{
					pNode->BeginningAddress = ((symbolnode *) pCurrent->Links.pPrevious)->EndingAddress;
					pNode->EndingAddress = pCurrent->BeginningAddress;
					memmove(PointToFunctionName(pNode), NOT_FOUND, sizeof(NOT_FOUND));
					memmove(PointToModuleName(pNode), NOT_FOUND,	sizeof(NOT_FOUND));

					AddToDoublyLinkedList(pCurrent->Links.pPrevious, (doublylinkedlist *) pNode);
				}
				else
				{
					fprintf(stderr, "Error: (Overlap) allocation failure\n");
					Error = TRUE;
					break;
				}
		
      }
    }
      
		pCurrent = (symbolnode *) pCurrent->Links.pNext;
  }
	
	//
	// If no errors occured while cruising thru the symbol list then...
	//
	if (!Error)
	{
		pCurrent = (symbolnode *) GetSymbolListHeadPointer()->pPrevious;

		//
		// If there's a hole at the end of the list (i.e. the ending address of the last
		// entry isn't 0xfff...fff) then fill it in.
		//
		if (pCurrent->EndingAddress != LastAddress)
		{ 
			Added++;

			if ((pNode = AllocateSymbolNode(strlen(NOT_FOUND) + strlen(NOT_FOUND))) != NULL)
			{
				memmove(PointToFunctionName(pNode), NOT_FOUND, sizeof(NOT_FOUND));
				memmove(PointToModuleName(pNode), NOT_FOUND, sizeof(NOT_FOUND));
				pNode->BeginningAddress = !IsEmptyDoublyLinkedList(GetSymbolListHeadPointer()) ?
				pCurrent->EndingAddress : 0;
				pNode->EndingAddress = LastAddress;

				AddToDoublyLinkedList((doublylinkedlist *) pCurrent, (doublylinkedlist *) pNode);
			}
			else
			{
				fprintf(stderr, "Error: (Other) allocation failure\n");
				Error = TRUE;
			}
		}

		if (!Error)
		{
			if (Logging)
			{
				printf("* Overlaps: %d\n"
						 "* Underlaps: %d\n"
						 "* Zeroes length symbols: %d\n"
						 "* Symbols added: %d\n",
						 Overlaps, Underlaps, Zeroes, Added);
			}

			if ((OutputFile = fopen(OutputFileName, "w")) != NULL)
			{
				pCurrent = NULL;

				while ((pCurrent =
				(symbolnode *) ReadDoublyLinkedList(GetSymbolListHeadPointer(), (doublylinkedlist *) pCurrent))
				!= NULL)
				{
					TotalSymbols++;
					(void) fprintf(OutputFile, SCAN_OUTPUT, PointToFunctionName(pCurrent), PointToModuleName(pCurrent),
					pCurrent->BeginningAddress, pCurrent->EndingAddress); 
				}

			  (void) fclose(OutputFile);

				if (Logging)
				{
					printf("* Symbols out of pass 2: %d\n"
							 "*\n"
							 "********\n",
							 TotalSymbols);
				}
			}
			else
			{
				(void) fprintf(stderr, "Error: Can't open file %s\n", OutputFileName);
			}
	  }
	}

	return;
}
