/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
 * Abstract: A suite of functions that operate on doubly linked lists, queues
 *           and stacks.
 *
 * Author: J. Goddard
 * Creation Date:  28-Aug-1997
 *
 * Modification history:
 *
 *		J. Goddard		10-Sep-1997
 *			Cleaned up/added documentation and code for several functions --
 *			just minor cosmetic stuff.
 *
 * General guide lines, etc. for using this suite of functions:
 *
 *	1) All list heads should be initialized prior to first use by
 *     calling the Initialize{type} functions or statically by setting
 *     the links the point back to the list head. This applies to queues
 *		 and stacks as well.
 *
 *	2) Queues and stacks are implemented using doubly linked lists. So, most
 *     all instructions concerning doubly linked lists apply to queues and stacks
 *     as well.
 */

#include <stdlib.h>
#include "stdtypes.h"
#include "structs.h"

/************************************************************************************************/
/*			    				      				        */
/* Name: GetLastListErrorPointer					       		      	*/
/*							    	  			    	*/
/* Purpose: Gets the pointer of the 'last error' cell.					    	*/
/*  				       								*/
/* Inputs: none																																									*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: a pointer to the 'last error' cell																						*/
/*																																															*/
/* Notes: o This function is local to this source module																				*/
/*	  		o This function is not included in structs.h therefore it must occur before the first	*/
/*	    		reference in this file.																															*/
/*																																															*/
/************************************************************************************************/
static linkedlistconditioncode *GetLastListErrorPointer(void)
{
	static linkedlistconditioncode LastListError = LL_SUCCESS;

	return &LastListError;
}

/************************************************************************************************/
/*		    			   	       					 	*/
/* Name: GetLastListError		   				       	 		*/
/*	    	      		        				    			*/
/* Purpose: Returns the contents of the 'last error' cell                                       */
/*																																															*/
/* Inputs: none																																									*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: the contents of the 'last error' cell																				*/
/*																																															*/
/* Notes: o This function returns values enumerated in type linkedlistconditioncode (in					*/
/*	    		structs.h).																																					*/
/*																																															*/
/************************************************************************************************/
linkedlistconditioncode GetLastListError(void)
{
	return *(GetLastListErrorPointer());
}

/************************************************************************************************/
/*																																															*/
/* Name: SetLastListError																																				*/
/*																																															*/
/* Purpose: Sets the contents of the 'last error' cell to a new value														*/
/*																																															*/
/* Inputs: Code: the new value																																	*/
/*																																															*/
/* Outputs: the contents of the 'last error' cell is changed to the value supplied by Code			*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: o The input value for this function must come from the enumerated type								*/
/*	    		linkedlistconditioncode (in structs.h).																							*/
/*																																															*/
/************************************************************************************************/
void SetLastListError(linkedlistconditioncode Code)
{
	*(GetLastListErrorPointer()) = Code;

	return;
}

/************************************************************************************************/
/*  					 			    			     	*/
/* Name: InitializeDoublyLinkedList																															*/ 
/*																																															*/
/* Purpose: Initialize a doubly linked list node's Next and Previous pointers										*/
/*																																															*/
/* Inputs: pListHead: a pointer to a doubly linked list node																		*/
/*																																															*/
/* Outputs: The links in pListHead are initialized so they point at pListHead.									*/
/*																																															*/
/* Function value: Same as input																																*/
/*																																															*/
/* Notes: o This function is primarily used to initialize list heads; however it can also				*/
/*	    		be used to initialize individual nodes as well.																			*/
/*				o This function assumes that pListHead is not NULL.																		*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *InitializeDoublyLinkedList(doublylinkedlist *pListHead)
{
	pListHead->pNext = pListHead;
	pListHead->pPrevious = pListHead;

	return pListHead;
}

/************************************************************************************************/
/*																																															*/
/* Name: InitializeQueue and InitializeStack																										*/
/*																																															*/
/* Notes: o see InitializeDoublyLinkedList																											*/
/*																																															*/
/************************************************************************************************/
queue *InitializeQueue(queue *pQueue)
{
	return InitializeDoublyLinkedList(pQueue);
}

stack *InitializeStack(stack *pStack)
{
	return InitializeDoublyLinkedList(pStack);
}

/************************************************************************************************/
/*																																															*/
/* Name: AllocateDoublyLinkedListEntry																													*/
/*																																															*/
/* Purpose: Allocate a new doubly linked list entry																							*/
/*																																															*/
/* Inputs: AllocationSize: Total size (including links) of the node to allocate									*/
/*         pNodeIdentifer: pointer to the identifier for this node															*/ 
/*																																															*/
/* Outputs: If successful then the Next/Previous links of the node are initialized else					*/
/*	    		the 'last error' cell is set to the reason for failure															*/
/*																																															*/
/* Function value: If successful then a pointer to the allocated node else NULL									*/
/*																																															*/
/* Notes: o AllocationSize must be atleast the doublylinkedlist bytes long (the size of one			*/
/*					set of links).																																			*/
/*				o This function does not insert the newly created node into a list. Other functions		*/
/*					included in this suite contain that functionality.																	*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *AllocateDoublyLinkedListEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier)
{
	doublylinkedlist *pNew = NULL;

	if (AllocationSize >= sizeof(doublylinkedlist))
  {
		if ((pNew = (doublylinkedlist *) malloc(AllocationSize)) != NULL)
	  {
			SetNodeIdentifier(pNew, pNodeIdentifier);
			pNew->Size = AllocationSize;
			InitializeDoublyLinkedList(pNew);
	  }
		else
	  {
			SetLastListError(LL_MALLOC_FAILED);
	  }
  }
	else
  {
		SetLastListError(LL_BAD_SIZE);
  }

	return pNew;
}

/************************************************************************************************/
/*																																															*/
/* Name: AllocateQueueEntry and AllocateStackEntry																							*/
/*																																															*/
/* Notes: o see AllocateDoublyLinkedListEntry																										*/
/*																																															*/
/************************************************************************************************/
queue *AllocateQueueEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier)
{
	return AllocateDoublyLinkedListEntry(AllocationSize, pNodeIdentifier);
}

stack *AllocateStackEntry(uint32 AllocationSize, nodeidentifier *pNodeIdentifier)
{
	return AllocateDoublyLinkedListEntry(AllocationSize, pNodeIdentifier);
}

/************************************************************************************************/
/*																																															*/
/* Name: SetNodeIdentifier																																			*/
/*																																															*/
/* Purpose: Set a node's identifier field																												*/
/*																																															*/
/* Inputs: pNode: pointer to a node 																														*/
/*				 pNodeIdentifier: pointer to the identifier																						*/
/*																																															*/
/* Outputs: The node's identifier field is set to the contents of pNodeIdentifier.							*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: This function assumes that pNode and pNodeIdentifier are not NULL.										*/
/*																																															*/
/************************************************************************************************/
void SetNodeIdentifier(doublylinkedlist *pNode, nodeidentifier *pNodeIdentifier)
{
	pNode->Identifier = *pNodeIdentifier;

	return;
}

/************************************************************************************************/
/*																																															*/
/* Name: SetQueueEntryIdentifier and SetStackEntryIdentifier																		*/
/*																																															*/
/* Notes: o see SetNodeIdentifier																																*/
/*																																															*/
/************************************************************************************************/
void SetQueueEntryIdentifier(queue *pEntry, nodeidentifier *pIdentifier)
{
	SetNodeIdentifier(pEntry, pIdentifier);

	return;
}

void SetStackEntryIdentifier(stack *pEntry, nodeidentifier *pIdentifier)
{
	SetNodeIdentifier(pEntry, pIdentifier);

	return;
}

/************************************************************************************************/
/*																																															*/
/* Name: ReadNodeIdentifier																																			*/
/*																																															*/
/* Purpose: Read a node's identifier field																											*/
/*																																															*/
/* Inputs: pNode: pointer to a node 																														*/
/*				 pNodeIdentifier: pointer to the identifier																						*/
/*																																															*/
/* Outputs: The identifier value of pNode is moved to the field pointed at by pNodeIdentifier.	*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: This function assumes that pNode and pNodeIdentifier are not NULL.										*/
/*																																															*/
/************************************************************************************************/
void ReadNodeIdentifier(doublylinkedlist *pNode, nodeidentifier *pNodeIdentifier)
{
	*pNodeIdentifier = pNode->Identifier;
}

/************************************************************************************************/
/*																																															*/
/* Name: GetNodeSize																																						*/
/*																																															*/
/* Purpose: Determine the size of a node																												*/
/*																																															*/
/* Inputs: pNode: pointer to a node 																														*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: the node's size																															*/
/*																																															*/
/* Notes: This function assumes that pNode is not NULL.																					*/
/*																																															*/
/************************************************************************************************/
uint32 GetNodeSize(doublylinkedlist *pNode)
{
	return pNode->Size;
}

/************************************************************************************************/
/*				      							       	*/
/* Name: DeallocateNode			       							*/
/*						   						*/
/* Purpose: Deallocate an entry from a doubly linked list																				*/
/*																																															*/
/* Inputs: pNode: pointer to the node to be deallocated																					*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: o This function does not remove the node from a list. Other functions included in			*/
/*					this suite contain that functionality. Therefore it is necessary to remove the			*/
/*					node prior to deallocating it.																											*/
/*				o This function assumes that pOld is not NULL.																				*/
/*																																															*/
/************************************************************************************************/
void DeallocateNode(doublylinkedlist *pNode)
{
	free((void *) pNode);

	return;
}


/************************************************************************************************/
/*											        */
/* Name: FreeLinkedList         							        */
/*								    	                        */
/* Purpose: This frees hether a doubly linked list.                                             */
/*                                                                                              */
/* Inputs: pListHead: pointer to the list head of the list to check                             */
/*                                                                                              */
/* Outputs: none                                                                                */
/*                                                                                              */
/* Function value: None.                                     			                */ 
/*                                                                                              */
/* Notes: This function doesn't work.  Think of it as a starting point.  		    	*/
/*                                                                                              */
/************************************************************************************************/

void FreeLinkedList(doublylinkedlist *pNode)
{
  doublylinkedlist *pNextNode;

  /* FIXME: This is the WRONG thing to do, however, in our implementation, the head is declared
     static while the rest are dynamically allocated.  */
  /* However, it is works don't fix it (Especially this close to release.. */
  /* We'll just start by freeing the next one. */ 

  pNode=pNode->pNext;

  while (pNode!=NULL)
    {
      pNextNode=pNode->pNext; 
      DeallocateNode(pNode);
      pNode=pNextNode;
    }

}





/************************************************************************************************/
/*											        */
/* Name: IsEmptyDoublyLinkedList							        */
/*								    	                        */
/* Purpose: Determine whether a doubly linked list is empty				        */
/*                                                                                              */
/* Inputs: pListHead: pointer to the list head of the list to check	                   	*/
/*												*/
/* Outputs: none				     	  		                        */
/*				                 					        */
/* Function value: TRUE if the list is empty otherwise FALSE			                */ 
/*					    				                        */
/* Notes: o This function assumes that pListhead is not NULL.				    	*/
/*							    		                        */
/************************************************************************************************/
boolean IsEmptyDoublyLinkedList(doublylinkedlist *pListHead)
{
	return pListHead->pNext == pListHead && pListHead->pPrevious == pListHead ? TRUE : FALSE;
}

/************************************************************************************************/
/*																																															*/
/* Name: IsEmptyQueue and IsEmptyStack																													*/
/*																																															*/
/* Notes: o see IsEmptyDoublyLinkedList																													*/
/*																																															*/
/************************************************************************************************/
boolean IsEmptyQueue(queue *pQueue)
{
	return IsEmptyDoublyLinkedList(pQueue);
}

boolean IsEmptyStack(stack *pStack)
{
	return IsEmptyDoublyLinkedList(pStack);
}

/************************************************************************************************/
/*																																															*/
/* Name: AddToDoublyLinkedList																																	*/
/*																																															*/
/* Purpose: Insert a node into a doubly linked list																							*/
/*																																															*/
/* Inputs: pNode: pointer to the node previous to the insertion point														*/
/*				 pNode: pointer to the node to insert																									*/
/*																																															*/
/* Outputs: The links of pPrevious and pNew are changed																					*/ 
/*																																															*/
/* Function value: pNew																																					*/
/*																																															*/
/* Notes: o This function assumes that pPrevious and pNew are not NULL.													*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *AddToDoublyLinkedList(doublylinkedlist *pNode, doublylinkedlist *pNew)
{
	pNew->pNext = pNode->pNext;
	pNew->pPrevious = pNode;

	pNode->pNext->pPrevious = pNew;
	pNode->pNext = pNew;

	return pNew;
}

/************************************************************************************************/
/*																																															*/
/* Name: AddToHeadOfDoublyLinkedList																														*/
/*																																															*/
/* Purpose: Add a node to the front of a doubly linked list																			*/ 
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*				 pNew: pointer to the node to insert																									*/
/*																																															*/
/* Outputs: The links of pListHead and pNew are changed																					*/ 
/*																																															*/
/* Function value: value of AddToDoublyLinkedList																								*/
/*																																															*/
/* Notes: o This function assumes that pListHead and pNew are not NULL.													*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *AddToHeadOfDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pNew)
{
	return AddToDoublyLinkedList(pListHead, pNew);
}

/************************************************************************************************/
/*																																															*/
/* Name: AddToTailOfDoublyLinkedList																														*/
/*																																															*/
/* Purpose: Add a node to the end of a doubly linked list.																			*/
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*				 pNew: pointer to the node to insert																									*/
/*																																															*/
/* Outputs: The links of the node previous to pListHead and pNew are changed										*/ 
/*																																															*/
/* Function value: value of AddToDoublyLinkedList																								*/
/*																																															*/
/* Notes: o This function assumes that pListHead and pNew are not NULL.													*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *AddToTailOfDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pNew)
{
	return AddToDoublyLinkedList(pListHead->pPrevious, pNew);
}

/************************************************************************************************/
/*																																															*/
/* Name: AddToDoublyLinkedListWithAssistance																										*/
/*																																															*/
/* Purpose: Programatically add a node to a doubly linked list.																	*/
/*																																															*/
/* Inputs: pListhead: pointer to the list's list head																						*/
/*				 pNew: pointer to the node being inserted																							*/
/*				 pAssitanceFunction: pointer to the function that will determine the insertion point	*/
/*																																															*/
/* Outputs: as noted else where																																	*/
/*																																															*/
/* Function value: NULL if the assitance function could not determine the insertion point or		*/
/*								 the value of AddToDoublyLinkedList																						*/
/*																																															*/
/* Notes: o This function assumes that pListHead, pNew and pAssistanceFunction are not NULL.		*/
/*				o The user supplied function, pAssistanceFunction, takes two inputs:									*/
/*						1) a pointer to the list's list head																							*/
/*						2) a pointer to the node to insert																								*/
/*					and has a value of the a pointer to the node previous to the insertion point.				*/
/*					If the assistance function can not determine the insertion point it should					*/
/*					return NULL. If NULL is returned no insertion will take place and the last					*/
/*					error code will be set to LL_INSERTION_FAILED. Clearly a non-NULL value should			*/
/*					define a valid insertion point in the list defined by pListHead.										*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *AddToDoublyLinkedListWithAssistance(doublylinkedlist *pListHead,
doublylinkedlist *pNew,
doublylinkedlist *(*pAssistanceFunction) (doublylinkedlist *, doublylinkedlist *))
{
	doublylinkedlist *pNode;

	SetLastListError(LL_SUCCESS);

	if ((pNode = (pAssistanceFunction)(pListHead, pNew)) != NULL)
	{
		pNode = AddToDoublyLinkedList(pNode, pNew);
	}
	else
	{
		SetLastListError(LL_INSERTION_FAILED);
	}

	return pNode;
}

/************************************************************************************************/
/*																																															*/
/* Name: RemoveFromDoublyLinkedList																															*/
/*																																															*/
/* Purpose: Remove a node form a doubly linked list																							*/
/*																																															*/
/* Inputs: pNode: pointer to the node to remove																									*/
/*																																															*/
/* Outputs: links in the previous/next nodes are changed																				*/
/*																																															*/
/* Function value: pointer to the removed node or NULL if the list is empty											*/
/*																																															*/
/* Notes: o This function assumes that pOld is not NULL.																				*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *RemoveFromDoublyLinkedList(doublylinkedlist *pNode)
{
	boolean EmptyList = IsEmptyDoublyLinkedList(pNode);

	if (!EmptyList)
  {
		pNode->pPrevious->pNext = pNode->pNext;
		pNode->pNext->pPrevious = pNode->pPrevious;
  }

	return EmptyList ? NULL : pNode;
}

/************************************************************************************************/
/*																																															*/
/* Name: RemoveFromHeadOfDoublyLinkedList																												*/
/*																																															*/
/* Purpose: Remove a node from the front of a doubly linked list																*/
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*																																															*/
/* Outputs: The links of the list head and node after the first are changed											*/
/*																																															*/
/* Function value: the value of RemoveFromDoublyLinkedList																			*/
/*																																															*/
/* Notes: o This function assumes that pListHead is not NULL.																		*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *RemoveFromHeadOfDoublyLinkedList(doublylinkedlist *pListHead)
{
	return RemoveFromDoublyLinkedList(pListHead->pNext);
}

/************************************************************************************************/
/*																																															*/
/* Name: RemoveFromTailOfDoublyLinkedList																												*/
/*																																															*/
/* Purpose: Remove a node from the end of a doubly linked list																	*/
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*																																															*/
/* Outputs: The links of the list head and node prior to the list head 													*/
/*																																															*/
/* Function value: the value of RemoveFromDoublyLinkedList																			*/
/*																																															*/
/* Notes: o This function assumes that pListHead is not NULL.																		*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *RemoveFromTailOfDoublyLinkedList(doublylinkedlist *pListHead)
{
	return RemoveFromDoublyLinkedList(pListHead->pPrevious);
}

/************************************************************************************************/
/*																																															*/
/* Name: ReaDoublyLinkedList																																		*/
/*																																															*/
/* Purpose: Read the contents of a doubly linked list from head to tail.												*/
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*				 pLast: pointer to the last node read																									*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: pointer the the next node in the list or NULL if all nodes have been read		*/
/*																																															*/
/* Notes: o This function assumes that the pointer pListHead is not NULL.												*/
/*				o If pLast is NULL then reading begins at the head of the list. For each successive		*/
/*					call the value of pLast must be the function value of the previous call.						*/
/*					For example:																																				*/
/*								doublylinkedlist ListHead, *pNode = NULL;																			*/
/*								                    .																													*/
/*								                    .																													*/
/*								                    .																													*/
/*								InitializeDoublyLinkedList(&ListHead);																				*/
/*								                    .																													*/
/*								                    .																													*/
/*								                    .																													*/
/*								while ((pNode = ReadDoublyLinkedList(&ListHead, pNode)) != NULL)							*/
/*								{																																							*/
/*								                    .																													*/
/*								                    .																													*/
/*								                    .																													*/
/*								}																																							*/
/*																																															*/
/************************************************************************************************/
doublylinkedlist *ReadDoublyLinkedList(doublylinkedlist *pListHead, doublylinkedlist *pLast)
{
	doublylinkedlist *pElement = pLast == NULL ? pListHead : pLast;

	return pElement->pNext != pListHead ? pElement->pNext : NULL;
}





/************************************************************************************************/
/*																																															*/
/* Name: ReadQueue																																							*/
/*																																															*/
/* Notes: o see ReadDoublyLinkedList																														*/
/*																																															*/
/************************************************************************************************/
queue *ReadQueue(queue *pQueue, queue *pLast)
{
	return ReadDoublyLinkedList(pQueue, pLast);
}

/************************************************************************************************/
/*																																															*/
/* Name: FlushDoublyLinkedList																																	*/
/*																																															*/
/* Purpose: Remove and deallocate the nodes of a list calling a user defined function						*/
/*					(if defined) between removal and deallocation.																			*/
/*																																															*/
/* Inputs: pListHead: pointer to the list's list head																						*/
/*				 pCallBackFunction: pointer to a user defined function call back function...this			*/
/*														function does auxillary processing prior to node deallocation			*/
/*																																															*/
/* Outputs: A node is removed from the list																											*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: o This function assumes that pListHead is not NULL.																		*/
/*				o The call back function address may be NULL indicating that no extra processing			*/
/*					is required between node removal and deallocation.																	*/
/*																																															*/
/************************************************************************************************/
void FlushDoublyLinkedList(doublylinkedlist *pListHead, void (*pCallBackFunction)(doublylinkedlist *))
{
	doublylinkedlist *pOld = NULL;

	while ((pOld = RemoveFromHeadOfDoublyLinkedList(pListHead)) != NULL)
  {
		if (pCallBackFunction != NULL)
	  {
			(pCallBackFunction)(pOld);
	  }

		DeallocateNode(pOld);
  }

	return;
}

/************************************************************************************************/
/*																																															*/
/* Name: AddToQueue																																							*/
/*																																															*/
/* Purpose: Add an entry to a queue																															*/
/*																																															*/
/* Inputs: pQueue: pointer to the queue to which the addition is to be made											*/
/*				 pNewEntry: pointer to the entry to be added																					*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: see AddToTailOfDOublyLinkedList																							*/
/*																																															*/
/* Notes: o This function assumes that pointers pQueue and pNewEntry are not NULL.							*/ 
/*																																															*/
/************************************************************************************************/
queue *AddToQueue(queue *pQueue, queue *pNewEntry)
{
	return AddToTailOfDoublyLinkedList(pQueue, pNewEntry);
}

/************************************************************************************************/
/*																																															*/
/* Name: RemoveFromQueue																																				*/
/*																																															*/
/* Purpose: Remove an entry from a queue																												*/ 
/*																																															*/
/* Inputs: pQueue: pointer to the queue from which to remove an entry														*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: see RemoveFromHeadOfDoublyLinkedList																					*/
/*																																															*/
/* Notes: o This function assumes that the pointer pQueue is not NULL														*/
/*																																															*/
/************************************************************************************************/
queue *RemoveFromQueue(queue *pQueue)
{
	return RemoveFromHeadOfDoublyLinkedList(pQueue);
}

/************************************************************************************************/
/*																																															*/
/* Name: FlushQueue																																							*/
/*																																															*/
/* Purpose: Remove and deallocate the entries of a queue calling a user defined function				*/
/*					(if defined) between removal and deallocation.																			*/
/*																																															*/
/* Inputs: pQueue: pointer to the queue																													*/
/*				 pCallBackFunction: pointer to a user defined function call back function...this			*/
/*														function does auxillary processing prior to entry deallocation		*/
/*																																															*/
/* Outputs: A node is removed from the list																											*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: o This function assumes that pQueue is not NULL.																			*/
/*				o The call back function address may be NULL indicating that no extra processing			*/
/*					is required between node removal and deallocation.																	*/
/*																																															*/
/************************************************************************************************/
void FlushQueue(queue *pQueue, void (*pCallBackFunction)(queue *))
{
	queue *pOld;

	while ((pOld = RemoveFromQueue(pQueue)) != NULL)
  {
		if (pCallBackFunction != NULL)
	  {
			(pCallBackFunction)(pOld);
	  }

		DeallocateNode(pOld);
  }

	return;
}

/************************************************************************************************/
/*																																															*/
/* Name: Push																																										*/
/*																																															*/
/* Purpose: Push an entry on a stack																														*/
/*																																															*/
/* Inputs: pStack: pointer of the stack to add to																								*/
/*         pNewEntry: pointer to the entry to push																							*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: see AddToHeadOfDoublyLinkedList																							*/
/*																																															*/
/* Notes: o This function assumes the pointers pStack and pNewEntry are not NULL.								*/
/*																																															*/
/************************************************************************************************/
stack *Push(stack *pStack, stack *pNewEntry)
{
	return AddToHeadOfDoublyLinkedList(pStack, pNewEntry);
}

/************************************************************************************************/
/*																																															*/
/* Name: Pop																																										*/
/*																																															*/
/* Purpose: Pop an entry off a stack																														*/
/*																																															*/
/* Inputs: pStack: pointer to the stack to pop																									*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: see RemoveFromHeadOfDoublyLinkedList																					*/
/*																																															*/
/* Notes: o This function assumes that pStack is not NULL.																			*/ 
/*																																															*/
/************************************************************************************************/
stack *Pop(stack *pStack)
{
	return RemoveFromHeadOfDoublyLinkedList(pStack);
}

/************************************************************************************************/
/*																																															*/
/* Name: FlushStack																																							*/
/*																																															*/
/* Purpose: Remove and deallocate the entries of a stack calling a user defined function				*/
/*					(if defined) between removal and deallocation.																			*/
/*																																															*/
/* Inputs: pStack: pointer to the stack																													*/
/*				 pCallBackFunction: pointer to a user defined function call back function...this			*/
/*														function does auxillary processing prior to entry deallocation		*/
/*																																															*/
/* Outputs: none																																								*/
/*																																															*/
/* Function value: none																																					*/
/*																																															*/
/* Notes: o This function assumes that pStack is not NULL.																			*/
/*				o The call back function address may be NULL indicating that no extra processing			*/
/*					is required between node removal and deallocation.																	*/
/*																																															*/
/************************************************************************************************/
void FlushStack(stack *pStack, void (*pCallBackFunction)(stack *))
{
	stack *pOld;

	while ((pOld = Pop(pStack)) != NULL)
  {
		if (pCallBackFunction != NULL)
	  {
			(pCallBackFunction)(pOld);
	  }

		DeallocateNode(pOld);
  }

	return;
 }
